package de.tud.kom.p2psim.impl.analyzer;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.GZIPOutputStream;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.analyzer.Analyzer.KBROverlayAnalyzer;
import de.tud.kom.p2psim.api.analyzer.Analyzer.NetAnalyzer;
import de.tud.kom.p2psim.api.analyzer.Analyzer.OperationAnalyzer;
import de.tud.kom.p2psim.api.common.Host;
import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.network.NetID;
import de.tud.kom.p2psim.api.network.NetMessage;
import de.tud.kom.p2psim.api.overlay.KBR;
import de.tud.kom.p2psim.api.overlay.OverlayContact;
import de.tud.kom.p2psim.api.overlay.OverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.ForwardMsg;
import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

/**
 * This Analyzer is used to write all data generated by using 
 * the KBR layer of DHTs into a log file. This log file can 
 * then be processed by the class <code>KBROverlayLogParser</code>
 * 
 * @author Julius Rueckert
 *
 */
public class KbrOverlayAnalyzer implements KBROverlayAnalyzer, OperationAnalyzer, NetAnalyzer {

	private final static Logger log = SimLogger.getLogger(KbrOverlayAnalyzer.class);
	
	private OutputStream outStream;
	
	private String outputPath;
	
	private boolean compressOutput;
	
	private boolean outStreamClosed = false;
	
	private HashMap<Message, Long> openQueries = new HashMap<Message, Long>();
	private HashMap<Message, Host> senderOfOpenQueries = new HashMap<Message, Host>();
	
	Map<NetID, Host> hostsForNetID = new HashMap<NetID, Host>();
	
	public KbrOverlayAnalyzer(){
		setOutputPath("kbrAnalyzerOutput/kbrOut.txt");
		setCompressOutput(false);
	}
	
	@Override
	public void messageDelivered(OverlayContact contact, Message msg, int hops) {
		writeToOutput("app_m_deliver "+contact.getOverlayID()+" "+msg.getPayload().getClass().getSimpleName()+" "+msg.getSize());
		
		if(openQueries.containsKey(msg.getPayload())){
			writeToOutput("q_succeeded "+contact.getOverlayID()+" "+msg.getPayload().getClass().getSimpleName()+" "
					      +((Simulator.getCurrentTime()-openQueries.get(msg.getPayload()))/Simulator.MILLISECOND_UNIT)+" "+hops);
		
			/**
			 * Clean up
			 */
			openQueries.remove(msg.getPayload());
			senderOfOpenQueries.remove(msg.getPayload());
		}
	}

	@Override
	public void messageForwarded(OverlayContact sender, OverlayContact receiver, Message msg, int hops) {
		if(hops==1)
			writeToOutput("app_m_init "+sender.getOverlayID()+" "+receiver.getOverlayID()+" "+msg.getPayload().getClass().getSimpleName()+" "+msg.getSize());
		else
			writeToOutput("app_m_forward "+sender.getOverlayID()+" "+receiver.getOverlayID()+" "+msg.getPayload().getClass().getSimpleName()+" "+msg.getSize());
	}
	
	@Override
	public void queryStarted(OverlayContact contact, Message appMsg){
		openQueries.put(appMsg, Simulator.getCurrentTime());
		senderOfOpenQueries.put(appMsg, hostsForNetID.get(contact.getTransInfo().getNetId()));
		writeToOutput("q_init "+contact.getOverlayID()+" "+appMsg.getClass().getSimpleName());
	}
	
	@Override
	public void queryFailed(OverlayContact failedHop, Message appMsg) {
		
		OverlayID initiator = senderOfOpenQueries.get(appMsg).getOverlay(KBR.class).getOverlayID();
		OverlayID failedHopID = failedHop.getOverlayID();
		
		writeToOutput("q_failed "+failedHopID+" "+initiator);
		
		/**
		 * Clean up
		 */
		openQueries.remove(appMsg);
		senderOfOpenQueries.remove(appMsg);
	}

	@Override
	public void start() {
		log.debug("KBROverlayAnalyzer started");

	}

	@Override
	public void stop(Writer output) {
		try {
			//Write net message statistic
			writeToOutput("m_stat "+netMsgCount);
			
			this.outStream.flush();
			outStreamClosed = true;
			this.outStream.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	
	}

	public void setOutputPath(String path){
		outputPath = path;
	}
	
	public void setCompressOutput(boolean compress){
		compressOutput = compress;
	}
	
	private void setupFile(){
		try {
			File outFile = new File(outputPath);
			if(!outFile.exists()){
				outFile.getParentFile().mkdirs();
			}
			
			if(compressOutput)
				outStream = new BufferedOutputStream(new GZIPOutputStream(new FileOutputStream(outFile)));
			else
				outStream = new BufferedOutputStream(new FileOutputStream(outFile));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void writeToOutput(String str){
		if(!outStreamClosed){
			if(outStream==null)
				setupFile();
			try {
				outStream.write((str+"\n").getBytes());
			} catch (IOException e) {
				e.printStackTrace();
			}
		} else {
			System.out.println("Stream already closed!");
		}
	}

	@Override
	public void operationFinished(Operation<?> op) {
		
		Host host = op.getComponent().getHost();
		hostsForNetID.put(host.getNetLayer().getNetID(), host);
	}

	@Override
	public void operationInitiated(Operation<?> op) {
		
		Host host = op.getComponent().getHost();
		hostsForNetID.put(host.getNetLayer().getNetID(), host);
	}

	Map<Message, NetID> netSender = new HashMap<Message, NetID>();
	Map<Message, Long> sentTimes = new HashMap<Message, Long>();
	double netMsgCount = 0;
	
	@Override
	public void netMsgReceive(NetMessage msg, NetID id) {
		OverlayID senderID = hostsForNetID.get(netSender.get(msg)).getOverlay(KBR.class).getOverlayID(); 
		OverlayID receiverID = hostsForNetID.get(id).getOverlay(KBR.class).getOverlayID();
		Message olMessage = msg.getPayload().getPayload();
		String messageType = olMessage.getClass().getSimpleName();
		
		long transmissionTime = Simulator.getCurrentTime()-sentTimes.get(msg)/Simulator.MILLISECOND_UNIT;
		writeToOutput("ol_m_transmitted "+senderID+" "+receiverID+" "+messageType+" "+olMessage.getSize()+" "+transmissionTime);

		/**
		 * Clean up
		 */
		sentTimes.remove(msg);
		netSender.remove(msg);
	}

	@Override
	public void netMsgSend(NetMessage msg, NetID id) {
		netSender.put(msg, id);
		sentTimes.put(msg, Simulator.getCurrentTime());
		netMsgCount++;
	}

	@Override
	public void netMsgDrop(NetMessage msg, NetID id) {
		Host host = hostsForNetID.get(id);
		Message olMessage = msg.getPayload().getPayload();
		
		if(host.getNetLayer().isOffline()){
			writeToOutput("m_lost stale_receiver "+host.getOverlay(KBR.class).getOverlayID()+" "+ olMessage.getClass().getSimpleName());
		} else {
			writeToOutput("m_lost unkonwn "+host.getOverlay(KBR.class).getOverlayID()+" "+ olMessage.getClass().getSimpleName());
		}
		
		if(olMessage instanceof ForwardMsg){
			ForwardMsg fm = (ForwardMsg) olMessage;
			
			if(fm.getKey()!=null){
				/*
				 * This was a forward message that was supposed to be routed, 
				 * that means it was a query that failed 
				 */
				OverlayID initiator = senderOfOpenQueries.get(fm.getPayload()).getOverlay(KBR.class).getOverlayID();
				OverlayID failedHop = hostsForNetID.get(id).getOverlay(KBR.class).getOverlayID();
				
				writeToOutput("q_failed "+failedHop+" "+initiator);
				
				/**
				 * Clean up
				 */
				openQueries.remove(fm.getPayload());
				senderOfOpenQueries.remove(fm.getPayload());
			}
		}
		
		/**
		 * Clean up
		 */
		sentTimes.remove(msg);
		netSender.remove(msg);
	}

}
