package de.tud.kom.p2psim.impl.network.gnp.topology;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Set;

import de.tud.kom.p2psim.impl.network.IPv4NetID;
import de.tud.kom.p2psim.impl.network.gnp.geoip.IspLookupService;

/**
 * Host objects holds and manage information about their location on earth and
 * gnp space and about measured rtts to monitors
 * 
 * @author Gerald Klunker
 * @version 0.1, 09.01.2008
 * 
 */
public class Host implements Serializable {

	private static final long serialVersionUID = 5135831087205329917L;

	public static int MONITOR = 1;

	public static int HOST = 2;

	private Long ipAddress = 0L;

	private int type;

	private String countryCode;

	private String region;

	private String city;

	private String isp;

	private String area;

	private boolean locatable = false;

	private double latitude = Double.NaN;

	private double longitude = Double.NaN;

	private HashMap<Host, Double> rttToMonitors = new HashMap<Host, Double>();

	private HostMap mapRef;

	public GnpPosition coordinateRef;

	/**
	 * 
	 * @param ipAddress
	 *            of the host
	 * @param mapRef
	 *            reference to the related HostMap Object
	 */
	public Host(Long ipAddress, HostMap mapRef) {
		super();
		this.ipAddress = ipAddress;
		this.mapRef = mapRef;
	}

	/**
	 * adds a measured rtt to monitor. if there is still a rtt, it will be
	 * replaced, if the new one is smaller
	 * 
	 * @param host
	 * @param rtt
	 *            measured rtt
	 */
	public void addRtt(Host monitor, double rtt) {
		if (rttToMonitors.containsKey(monitor)) {
			double currentRtt = rttToMonitors.get(monitor);
			if (currentRtt < rtt)
				return;
		}
		rttToMonitors.put(monitor, rtt);
	}

	/**
	 * set a measured rtt to monitor.
	 * 
	 * @param host
	 * @param rtt
	 *            measured rtt
	 */
	public void setRtt(Host monitor, double rtt) {
		rttToMonitors.put(monitor, rtt);
	}

	/**
	 * 
	 * @param monitor
	 * @return measured rtt from host to monitor
	 */
	public double getRtt(Host monitor) {
		if (rttToMonitors.containsKey(monitor))
			return rttToMonitors.get(monitor);
		else
			return Double.NaN;
	}

	/**
	 * remove them measured rtt from host to monitor
	 * 
	 * @param monitor
	 */
	protected void removeRTT(Host monitor) {
		rttToMonitors.remove(monitor);
	}

	/**
	 * 
	 * @return set of monitors with a measured rtt to host
	 */
	protected Set<Host> getMeasuredMonitors() {
		return rttToMonitors.keySet();
	}

	/**
	 * 
	 * @return latitude
	 */
	public double getLatitude() {
		return latitude;
	}

	/**
	 * 
	 * @return longitude
	 */
	public double getLongitude() {
		return longitude;
	}

	/**
	 * 
	 * @return longitude
	 */
	public String getArea() {
		return area;
	}

	/**
	 * 
	 * @return 2-digits country code
	 */
	public String getCountryCode() {
		return countryCode;
	}

	/**
	 * @return region name
	 */
	public String getRegion() {
		return region;
	}

	/**
	 * @return city name
	 */
	public String getCity() {
		return city;
	}

	/**
	 * @return isp name
	 */
	public String getISP() {
		return isp;
	}

	/**
	 * 
	 * @return ip-address as 32bit value
	 */
	public Long getIpAddress() {
		return ipAddress;
	}

	/**
	 * 
	 * @param peerType
	 *            (1:monitor, 2:host)
	 */
	protected void setHostType(int peerType) {
		this.type = peerType;
	}

	/**
	 * 
	 * @return type of host (1:monitor, 2:host)
	 */
	public int getHostType() {
		return type;
	}

	/**
	 * 
	 * @param coordinateRef
	 *            reference to the related gnp position
	 */
	protected void setPositionReference(GnpPosition coordinateRef) {
		this.coordinateRef = coordinateRef;
	}

	/**
	 * 
	 * @return reference to the related gnp position, null, if no position was
	 *         calculated
	 */
	protected GnpPosition getGnpPositionReference() {
		return coordinateRef;
	}

	/**
	 * Try to locate the host on earth using the geolocator
	 * 
	 * @param locator
	 */
	public void setLocation(Geolocator locator, IspLookupService service) {
		this.locatable = false;
//		if (locator.search(ipAddress) && locator.getCountryCode() != null && locator.getCityName() != null && locator.getRegionName() != null && service.getISP(ipAddress) != null) {
			 if (locator.search(ipAddress) && locator.getCountryCode() != null && locator.getCityName() != null && locator.getRegionName() != null) {
			this.countryCode = locator.getCountryCode();
			this.region = locator.getRegionName();
			this.latitude = locator.getLatitude();
			this.longitude = locator.getLongitude();
			this.city = locator.getCityName();
			if (service != null)
				this.isp = service.getISP(ipAddress);
			this.mapRef.getCountryLookup().addCountryFromGeoIP(countryCode, locator.getCountryName());
			this.area = this.mapRef.getCountryLookup().getPingErRegionName(countryCode);
			if (this.area != null)
				this.locatable = true;
		} else {
			this.latitude = Double.NaN;
			this.longitude = Double.NaN;
		}
	}

	/**
	 * 
	 * @return true, if host was located on a longitude/latitude
	 */
	public boolean isLocatable() {
		return locatable;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		String returnString = "";
		returnString += IPv4NetID.ipToString(ipAddress);
		returnString += "\t(" + this.getCountryCode() + ")";
		return returnString;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object host) {
		return (this.ipAddress == ((Host) host).getIpAddress());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return ipAddress.hashCode();
	}
}
