package de.tud.kom.p2psim.impl.network.gnpPrio;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import org.apache.log4j.Logger;
import org.dom4j.Document;
import org.dom4j.DocumentException;
import org.dom4j.Element;
import org.dom4j.io.SAXReader;

import de.tud.kom.p2psim.api.common.ComponentFactory;
import de.tud.kom.p2psim.api.common.Host;
import de.tud.kom.p2psim.impl.network.IPv4NetID;
import de.tud.kom.p2psim.impl.network.gnp.topology.CountryLookup;
import de.tud.kom.p2psim.impl.network.gnp.topology.GnpPosition;
import de.tud.kom.p2psim.impl.network.gnp.topology.PingErLookup;
import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

public class GnpNetLayerFactory implements ComponentFactory {

	private static Logger log = SimLogger.getLogger(GnpNetLayerFactory.class);

	private final GnpSubnet subnet;

	private final static double DEFAULT_DOWN_BANDWIDTH = 1000l;

	private final static double DEFAULT_UP_BANDWIDTH = 1000l;

	private double downBandwidth;

	private double upBandwidth;

	private HashMap<IPv4NetID, GnpHostInfo> hostPool;

	private HashMap<String, ArrayList<IPv4NetID>> namedGroups;

	private PingErLookup pingErLookup;

	private CountryLookup countryLookup;

	public GnpNetLayerFactory() {
		subnet = new GnpSubnet();
		this.downBandwidth = DEFAULT_DOWN_BANDWIDTH;
		this.upBandwidth = DEFAULT_UP_BANDWIDTH;
	}

	public GnpNetLayer createComponent(Host host) {
		GnpNetLayer netLayer = newNetLayer(host.getProperties().getGroupID());
		netLayer.setHost(host);
		return netLayer;
	}

	/**
	 * random node form group
	 * 
	 * @param id
	 * @return
	 */
	public GnpNetLayer newNetLayer(String id) {
		if (this.namedGroups.containsKey(id) && !this.namedGroups.get(id).isEmpty()) {
			int size = namedGroups.get(id).size();
			IPv4NetID netId = namedGroups.get(id).get(Simulator.getRandom().nextInt(size));
			namedGroups.get(id).remove(netId);
			return newNetLayer(netId);
		} else {
			throw new IllegalStateException("No (more) Hosts are assigned to \"" + id + "\"");
		}
	}

	private GnpNetLayer newNetLayer(IPv4NetID netID) {
		GnpPosition gnpPos = this.hostPool.get(netID).gnpPosition;
		GeoLocation geoLoc = this.hostPool.get(netID).geoLoc;
		GnpNetLayer nw = new GnpNetLayer(this.subnet, netID, gnpPos, geoLoc, this.downBandwidth, this.upBandwidth);
		hostPool.remove(netID);
		return nw;
	}

	public void setGnpFile(String gnpFileName) {

		File gnpFile = new File(gnpFileName);

		hostPool = new HashMap<IPv4NetID, GnpHostInfo>();
		namedGroups = new HashMap<String, ArrayList<IPv4NetID>>();

		log.info("Read hosts from file " + gnpFile);
		SAXReader reader = new SAXReader(false);

		Document configuration = null;
		try {
			configuration = reader.read(gnpFile);
		} catch (DocumentException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		Element root = configuration.getRootElement();
		assert root.getName().equals("gnp");

		for (Object obj : root.elements()) {
			Element elem = (Element) obj;
			if (elem.getName().equals("GroupLookup")) {
				for (Iterator iter = elem.elementIterator("Group"); iter.hasNext();) {
					Element variable = (Element) iter.next();
					String id = variable.attributeValue("id");

					ArrayList<IPv4NetID> group = new ArrayList<IPv4NetID>();
					for (Iterator ipIter = variable.elementIterator("IPs"); ipIter.hasNext();) {
						Element ipElement = (Element) ipIter.next();
						String[] ips = ipElement.attributeValue("value").split(",");
						for (int c = 0; c < ips.length; c++)
							group.add(new IPv4NetID(Long.parseLong(ips[c])));
					}
					if (namedGroups.containsKey(id)) {
						throw new IllegalStateException("Multiple Group Definition in " + gnpFileName + " ( Group: " + id + " )");
					} else {
						namedGroups.put(id, group);
					}
				}
			} else if (elem.getName().equals("Hosts")) {
				for (Iterator iter = elem.elementIterator("Host"); iter.hasNext();) {
					Element variable = (Element) iter.next();

					// IP-Address
					IPv4NetID hostID = new IPv4NetID(Long.parseLong(variable.attributeValue("ip")));

					// GNP-Coordinates
					String[] coordinatesS = variable.attributeValue("coordinates").split(",");
					double[] coordinatesD = new double[coordinatesS.length];
					for (int c = 0; c < coordinatesD.length; c++)
						coordinatesD[c] = Double.parseDouble(coordinatesS[c]);
					GnpPosition gnpPos = new GnpPosition(coordinatesD);

					// GeoLocation
					String continentalArea = variable.attributeValue("continentalArea");
					String countryCode = variable.attributeValue("countryCode");
					String region = variable.attributeValue("region");
					String city = variable.attributeValue("city");
					String isp = variable.attributeValue("isp");
					double longitude = Double.parseDouble(variable.attributeValue("longitude"));
					double latitude = Double.parseDouble(variable.attributeValue("latitude"));
					GeoLocation geoLoc = new GeoLocation(continentalArea, countryCode, region, city, isp, latitude, longitude);

					GnpHostInfo hostInfo = new GnpHostInfo(geoLoc, gnpPos);
					hostPool.put(hostID, hostInfo);
				}
			} else if (elem.getName().equals("PingErLookup")) {
				pingErLookup = new PingErLookup();
				pingErLookup.loadFromXML(elem);

			} else if (elem.getName().equals("CountryLookup")) {
				countryLookup = new CountryLookup();
				countryLookup.importFromXML(elem);
			}
		}
	}

	public void setDownBandwidth(double downBandwidth) {
		this.downBandwidth = downBandwidth;
	}

	public void setUpBandwidth(double upBandwidth) {
		this.upBandwidth = upBandwidth;
	}

	public void setLatencyModel(GnpLatencyModel model) {
		model.init(pingErLookup, countryLookup);
		subnet.setLatencyModel(model);
	}

	public void setBandwidthManager(AbstractGnpNetBandwidthManager bm) {
		subnet.setBandwidthManager(bm);
	}

	public void setPbaPeriod(double seconds) {
		subnet.setPbaPeriod(Math.round(seconds * Simulator.SECOND_UNIT));
	}

	private class GnpHostInfo {

		private GnpPosition gnpPosition;

		private GeoLocation geoLoc;

		public GnpHostInfo(GeoLocation geoLoc, GnpPosition gnpPos) {
			this.gnpPosition = gnpPos;
			this.geoLoc = geoLoc;
		}

	}
	
	/**
	 *  Set size of the buffer in bytes for messages that wait for delivering
	 * 
	 * @param size
	 */
	public void setBufferSize(double size){
		subnet.setBufferSize(size);
	}
	
	public void setBandwidthMode(String mode){
		subnet.setBandwidthMode(mode);
	}

}
