package de.tud.kom.p2psim.impl.overlay.dht.chord;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.Map;

import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.api.overlay.OverlayKey;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransMessageListener;
import de.tud.kom.p2psim.api.transport.TransMsgEvent;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.overlay.AbstractOverlayNode.PeerStatus;
import de.tud.kom.p2psim.impl.overlay.dht.ForwardMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.IndexTransferMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.JoinMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.LeaveMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.LookupReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.LookupRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.NotifyMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.PingMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.PongMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.StabilizeMsg;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.StoreReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.StoreRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.LookupOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.StabilizeOperation;

/**
 * This class handles all incoming messages of a chord node
 * 
 * @author Yue Sheng
 */
public class ChordMessageHandler implements TransMessageListener, OperationCallback<Object> {

	private ChordNode master;

	/**
	 * @param node
	 */
	public ChordMessageHandler(ChordNode node) {
		master = node;
	}

	public void messageArrived(TransMsgEvent receivingEvent) {
		if(!master.getPeerStatus().equals(PeerStatus.PRESENT)){
			return;
		}
		Message msg = receivingEvent.getPayload();
		if (msg instanceof PingMsg) { // handle ping
			PingMsg ping = (PingMsg) msg;
			PongMsg pong = new PongMsg(master.getChordID(), ping.getSender());
			master.getTransLayer().sendReply(pong, receivingEvent, master.getPort(), TransProtocol.UDP);
		} else if (msg instanceof LookupRequest) { // handle Lookup request
			LookupRequest req = (LookupRequest) msg;
			req.incHop();
			ChordContact succ = master.nextActiveSucc();
			if(succ!=null){
				ChordID keyID = new ChordID(req.getKey().getKeyID());
				ChordID nId = master.getChordID(), succId = succ.getOverlayID();
				if (keyID.between(nId, succId) || keyID.equals(succId)) {
					// get result
					Object lookupResult = succ;
					// send lookup reply to the originator
					LookupReply lookupRpl = new LookupReply(req, lookupResult);
					TransInfo starterAdr = req.getStarter().getTransInfo();
					master.getTransLayer().send(lookupRpl, starterAdr, master.getPort(), TransProtocol.UDP);
				} else {
					ChordContact nextHop = master.closestPre(keyID);
					if(nextHop.getOverlayID().equals(master.getChordID())){
						System.err.println("potential deadlock"); //test
					}
					LookupRequest lookupReq = new LookupRequest(master.getChordID(), nextHop.getOverlayID(), req.getKey(), req.forValue());
					lookupReq.setStarter(req.getStarter());
					lookupReq.setOpID(req.getOpID());
					lookupReq.setHop(req.getHopCount());
					master.getTransLayer().send(lookupReq, nextHop.getTransInfo(), master.getPort(), TransProtocol.UDP);
				}
			}

		} else if (msg instanceof LookupReply) { // handle lookup reply
			LookupReply lookupRpl = (LookupReply) msg;
			LookupOperation op = master.lookups.get(lookupRpl.getOpID());
			if (op != null) {
				op.found(lookupRpl);
			}
		} else if (msg instanceof NotifyMsg) { // handle notify command
			NotifyMsg notifyMsg = (NotifyMsg) msg;
			master.notify(notifyMsg.getContent());
		} else if(msg instanceof StabilizeMsg){ // handle stabilization command
			new StabilizeOperation(master, this).scheduleImmediately();
		} else if (msg instanceof GetInfoRequest) { // handle getInfo request
			GetInfoRequest getInfoReq = (GetInfoRequest) msg;
			GetInfoReply reply = null;
			String type = getInfoReq.getType();
			if (type.equalsIgnoreCase("pre")) {
				ChordContact pre = master.getPre();
				if (pre != null)
					pre = pre.clone();
				reply = new GetInfoReply(getInfoReq, pre);
			} else if (type.equalsIgnoreCase("succ")) {
				reply = new GetInfoReply(getInfoReq, master.getSucc().clone());
			} else if (type.equalsIgnoreCase("dhtObject")) {
				BigInteger key = (BigInteger) getInfoReq.getKey().getUniqueValue();
				reply = new GetInfoReply(getInfoReq, master.getIndex().get(key));
			}
			if (reply != null) {
				master.getTransLayer().sendReply(reply, receivingEvent, master.getPort(), TransProtocol.UDP);
			}
		} else if (msg instanceof StoreRequest) { // handle store operation
			StoreRequest req = (StoreRequest) msg;
			master.getIndex().put((BigInteger) req.getKey().getUniqueValue(), req.getValue());
			StoreReply reply = new StoreReply(req);
			master.getTransLayer().sendReply(reply, receivingEvent, master.getPort(), TransProtocol.UDP);
		} else if (msg instanceof IndexTransferMsg) { // receive key-value pairs
			IndexTransferMsg indexTransfer = (IndexTransferMsg) msg;
			master.getIndex().putAll(indexTransfer.getContent());
		} else if (msg instanceof JoinMsg) { // handle node joining
			JoinMsg join = (JoinMsg) msg;
			ChordContact joiningNode = join.getJoiningNode();
			Map<BigInteger, DHTObject> map = new HashMap<BigInteger, DHTObject>();
			ChordID n = master.getChordID(), newID = joiningNode.getOverlayID();
			for (BigInteger key : master.getIndex().keySet()) {
				ChordID keyId = new ChordID(key);
				if (!keyId.equals(master.getChordID()) && !keyId.between(newID, n)) {
					map.put(key, master.getIndex().remove(key));
				}
			} // move this part of keys to the new node
			if (!map.isEmpty()) {
				IndexTransferMsg indexTransfer = new IndexTransferMsg(master.getChordID(), joiningNode.getOverlayID(), map);
				master.getTransLayer().send(indexTransfer, joiningNode.getTransInfo(), master.getPort(), TransProtocol.UDP);
			}
			joiningNode.setAlive(true);	// del
			if(master.getPre()!=null){
				StabilizeMsg m = new StabilizeMsg(master.getChordID(),master.getPre().getOverlayID());
				master.getTransLayer().send(m, master.getPre().getTransInfo(), master.getPort(), TransProtocol.UDP);
			}
			master.setPre(joiningNode);
			
			ChordContact nextSucc = master.nextActiveSucc();
			if (nextSucc!=null && nextSucc.getOverlayID().equals(master.getChordID())) {
				master.getSucc().set(0, joiningNode);
				new StabilizeOperation(master, this).scheduleImmediately();
			}
		} else if (msg instanceof LeaveMsg) { // handle node departure
			LeaveMsg leave = (LeaveMsg) msg;
			ChordContact leavingNode = leave.getLeavingNode();
			if (leavingNode.equals(master.getPre())) {
				master.setPre(leave.getPre());
			} else if (leavingNode.equals(master.nextActiveSucc())) {
				master.nextActiveSucc().setAlive(false);
			}
		} 
	}

	public void calledOperationFailed(Operation<Object> op) {
		// TODO Auto-generated method stub
		
	}

	public void calledOperationSucceeded(Operation<Object> op) {
		// TODO Auto-generated method stub
		
	}
}
