package de.tud.kom.p2psim.impl.overlay.dht.chord;

import java.math.BigInteger;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.common.ConnectivityEvent;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.overlay.DHTNode;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.api.overlay.OverlayKey;
import de.tud.kom.p2psim.api.transport.TransLayer;
import de.tud.kom.p2psim.impl.overlay.AbstractOverlayNode;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.CheckPreOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.FixFingerOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.JoinOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.LeaveOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.LookupOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.NotifyOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.StabilizeOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.StoreOperation;
import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

/**
 * @author Yue Sheng
 * 
 */
public class ChordNode extends AbstractOverlayNode implements DHTNode {

	private static Logger log = SimLogger.getLogger(ChordNode.class);

	/** the length of successor list */
	public static final short r = 5;

	private ChordBootstrapManager bootstrap = null;

	private FingerTable ft;

	private ChordContact pre = null, localContact = null;

	private LinkedList<ChordContact> succ = new LinkedList<ChordContact>();

	private Map<BigInteger, DHTObject> index = new HashMap<BigInteger, DHTObject>();

	private TransLayer transLayer;

	private ChordMessageHandler msgHandler;

	private OperationListener opListener;

	Map<Integer, LookupOperation> lookups = new HashMap<Integer, LookupOperation>();

	/**
	 * @param transLayer
	 * @param port
	 */
	public ChordNode(TransLayer transLayer, short port) {
		super(new ChordID(transLayer.getLocalTransInfo(port)), port);
		this.transLayer = transLayer;
		this.ft = new FingerTable(this);
		localContact = new ChordContact(getChordID(), transLayer
				.getLocalTransInfo(port));
		// assign the operation listener to this
		opListener = new OperationListener(this);

		// create message handler
		msgHandler = new ChordMessageHandler(this);
		// add message handler to the listeners list
		this.getTransLayer().addTransMsgListener(this.msgHandler,
				this.getPort());

	}

	@Override
	public TransLayer getTransLayer() {
		return this.transLayer;
	}

	public void connectivityChanged(ConnectivityEvent ce) {
		if (ce.isOffline()) {
			setPeerStatus(PeerStatus.ABSENT);
			localContact.setAlive(false);
		} else if (ce.isOnline()) {
			setPeerStatus(PeerStatus.PRESENT);
			localContact.setAlive(true);
			enableBackgroundOps();
		}
	}

	public int join(OperationCallback callback) {
		// get the first active node in bootstrap, to join in the overlay
		JoinOperation join = new JoinOperation(this, callback);
		join.scheduleImmediately();

		return join.getOperationID();
	}

	/* start the periodically called tasks */
	public void enableBackgroundOps() {
		FixFingerOperation fixFingers = new FixFingerOperation(this, opListener);
		StabilizeOperation stabilize = new StabilizeOperation(this, opListener);
		CheckPreOperation checkPre = new CheckPreOperation(this, opListener);

		stabilize.scheduleImmediately();
		checkPre.scheduleWithDelay(50 * Simulator.MILLISECOND_UNIT);
		fixFingers.scheduleWithDelay(100 * Simulator.MILLISECOND_UNIT);

	}

	public int leave(OperationCallback callback) {
		LeaveOperation leave = new LeaveOperation(this, callback);
		leave.scheduleImmediately();
		return leave.getOperationID();
	}

	public int valueLookup(OverlayKey key, OperationCallback callback) {
		ChordKey k = (ChordKey) key;
		LookupOperation op = new LookupOperation(this, k, callback, true);
		op.scheduleImmediately();
		// lookups.put(op.getOperationID(), op);
		return op.getOperationID();
	}

	public int nodeLookup(OverlayKey key, OperationCallback callback) {
		ChordKey k = (ChordKey) key;
		LookupOperation op = new LookupOperation(this, k, callback, false);
		op.scheduleImmediately();
		// lookups.put(op.getOperationID(), op);
		return op.getOperationID();
	}

	public int store(OverlayKey key, DHTObject obj, OperationCallback callback) {
		StoreOperation op = new StoreOperation(this, key, obj, callback);
		op.scheduleImmediately();
		return op.getOperationID();
	}

	/**
	 * notifies this node about its potential predecessor
	 */
	void notify(ChordContact pre) {
		NotifyOperation notify = new NotifyOperation(this, pre, opListener);
		notify.scheduleImmediately();
	}

	/* handle node failures */
	public ChordContact nextActiveSucc() {
		for (int i = 0; i < succ.size() && i < r; i++) {
			ChordContact succ_i = succ.get(i);
			if (succ_i != null && succ_i.isAlive())
				return succ_i;
		}
		log.warn(Simulator.getSimulatedRealtime()+" - "+getOverlayID()+" - All successors of the node failed!!");
		return null;
	}

	public ChordContact closestPre(ChordID id) {
		ChordContact c = ft.getClosestPre(id);
		ChordID succ_j, closestFinger = c.getOverlayID();
		// search successor list
		for (int j = succ.size() - 1; j >= 0; j--) {
			if (!succ.get(j).isAlive())
				continue;
			succ_j = succ.get(j).getOverlayID();
			if (succ_j.between(closestFinger, id)) {
				return succ.get(j);
			}
		}
		return c;
	}

	public boolean isRootOf(OverlayKey key) {
		ChordID keyId = new ChordID((BigInteger) key.getUniqueValue());
		if (pre == null) {
			ChordContact c = nextActiveSucc();
			if (c!=null && c.getOverlayID().equals(getChordID())) {
				return true;
			} else
				return false;
		} else if (keyId.equals(getChordID())) {
			return true;
		} else if (keyId.between(pre.getOverlayID(), getChordID())) {
			return true;
		}
		return false;
	}

	/* getters and setters */
	public FingerTable getFingerTable() {
		return ft;
	}

	public ChordID getChordID() {
		return (ChordID) getOverlayID();
	}

	public ChordContact getLocalContact() {
		return localContact.clone();
	}

	public void setAlive(boolean live) {
		localContact.setAlive(live);
	}

	public ChordBootstrapManager getBootstrap() {
		return bootstrap;
	}

	public void setBootstrap(ChordBootstrapManager bootstrap) {
		this.bootstrap = bootstrap;
	}

	public ChordContact getPre() {
		return pre;
	}

	public void setPre(ChordContact pre) {
		this.pre = pre;
	}

	public LinkedList<ChordContact> getSucc() {
		return succ;
	}

	public void setSucc(LinkedList<ChordContact> succ) {
		this.succ = succ;
	}

	public Map<BigInteger, DHTObject> getIndex() {
		return index;
	}

	public void registerLookup(Integer id, LookupOperation op) {
		lookups.put(id, op);
	}

	public void lookupFinished(Integer cmdID) {
		lookups.remove(cmdID);
	}

	public OperationListener getOpListener() {
		return opListener;
	}
	
	
}
