package de.tud.kom.p2psim.impl.overlay.dht.chord;

import java.math.BigInteger;
import java.util.LinkedList;
import java.util.List;

import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.overlay.KBR;
import de.tud.kom.p2psim.api.overlay.KBRListener;
import de.tud.kom.p2psim.api.overlay.OverlayContact;
import de.tud.kom.p2psim.api.overlay.OverlayID;
import de.tud.kom.p2psim.api.overlay.OverlayKey;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransLayer;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.overlay.dht.ForwardMsg;
import de.tud.kom.p2psim.impl.overlay.dht.KBRMsgHandler;
import de.tud.kom.p2psim.impl.simengine.Simulator;

/**
 * @author Yue Sheng (improved by Julius Rueckert)
 * @version 3.0, May 22, 2008
 * 
 */
public class KBRChordNode extends ChordNode implements KBR<ChordID, ChordContact>{

	private KBRListener kbrListener;
	private KBRMsgHandler kbrMsgHandler;

	/**
	 * @param transLayer
	 * @param port
	 */
	public KBRChordNode(TransLayer transLayer, short port) {
		super(transLayer, port);
	}

	@Override
	public void setKBRListener(KBRListener listener){
		this.kbrListener = listener;
		kbrMsgHandler = new KBRMsgHandler<ChordID, ChordContact>(this, this, kbrListener);
		// add KBR handler to the listeners list
		this.getTransLayer().addTransMsgListener(kbrMsgHandler, this.getPort()); 
	}

	@Override
	public void route(OverlayKey key, Message msg, ChordContact hint) {
		ChordContact nextHop=null;
		if (hint != null) {
			nextHop = hint;
		} else if (key != null) {
			nextHop = closestPre(new ChordID((BigInteger) key.getUniqueValue()));
			
			//Inform the monitors about an initiated query
			Simulator.getMonitor().queryStarted(getLocalOverlayContact(), msg);
		} else {
			System.err.println("Both key and hint are null!!");
			return;
		}
		kbrListener.forward(key, msg, nextHop);
		ForwardMsg fm = new ForwardMsg(getChordID(), nextHop.getOverlayID(), key, msg);
		getTransLayer().send(fm, nextHop.getTransInfo(), getPort(), TransProtocol.UDP);
	}

	@Override
	public List<ChordContact> local_lookup(OverlayKey key, int num) {
		LinkedList<ChordContact> result = new LinkedList<ChordContact>();
		LinkedList<ChordContact> biggestFingers = new LinkedList<ChordContact>();
		ChordID keyId = new ChordID((BigInteger) key.getUniqueValue());
		FingerTable ft = getFingerTable();
		ChordID finger_i = null, n = getChordID();
		
		// Check if there are fingers between the node and the key
		for (int i = ft.getFtSize()-1; i >= 0; i--) {
			ChordContact finger = ft.getFinger(i);
			
			if(finger != null){
				biggestFingers.add(finger.clone());
				finger_i = finger.getOverlayID();
				
				if (finger_i.between(n, keyId) || finger_i.equals(keyId)) {
					if (result.size() < num) {
						result.add(ft.getFinger(i).clone());
					} else {
						break;
					}
				} 
			}
		}
		
		if(result.size() == 0){
			//No finger is between the node and the key
			
			if(nextActiveSucc()!= null && keyId.between(getChordID(), nextActiveSucc().getOverlayID())) {
				//The successor is root of the key
				result.add(nextActiveSucc());
			} else if(biggestFingers != null){
				//No finger is between the node and the key, plus the successor is not 
				//the root of the key -> Use the biggest fingers as possible next hops
				return biggestFingers.subList(0, Math.min(num, biggestFingers.size()));
			}
		}
		return result;
	}

	@Override
	public List<ChordContact> neighborSet(int num) {
		List<ChordContact> neighbors = new LinkedList<ChordContact>();
		if (num <= 0) {
			return null;
		}
		neighbors.add(getPre());
		neighbors.addAll(getSucc());
		if (neighbors.size() > num)
			return neighbors.subList(0, num);
		else
			return neighbors;
	}

	@Override
	public OverlayID[] range(ChordContact contact, int rank) {
		// sort all neighbors in a list
		LinkedList<ChordContact> neighbors = new LinkedList<ChordContact>();
		for (int i = getSucc().size() - 1; i >= 0; i--) {
			if(getSucc().get(i).isAlive()){
				neighbors.add(getSucc().get(i));
			}
		}
		neighbors.add(getLocalContact());
		neighbors.add(getPre());
		// determine the range
		if (rank < 0 || (rank + 1) >= neighbors.size()) {
			return null; // out of range
		}
		for (int i = 0; i < neighbors.size(); i++) {
			if (contact.equals(neighbors.get(i))) {
				OverlayID[] range = new OverlayID[2];
				range[0] = neighbors.get(i + rank + 1).getOverlayID();
				range[1] = neighbors.get(i + rank).getOverlayID();
				return range;
			}
		}
		return null; // contact not in neighborhood
	}

	@Override
	public List<ChordContact> replicaSet(OverlayKey key, int maxRank) {
		if (maxRank < r) {
			return getSucc().subList(0, maxRank);
		} else {
			return getSucc();
		}
	}

	@Override
	public OverlayKey getNewOverlayKey(int rank) {
		return new ChordKey((new ChordIDGenerator()).getID(String.valueOf(rank)));
	}

	@Override
	public OverlayContact<ChordID> getLocalOverlayContact() {
		ChordID overlayID = (ChordID)getHost().getOverlay(KBR.class).getOverlayID();
		short port = getHost().getOverlay(KBR.class).getPort();
		TransInfo transInfo = getHost().getTransLayer().getLocalTransInfo(port);
		
		return new ChordContact(overlayID, transInfo); 
	}

}
