package de.tud.kom.p2psim.impl.overlay.dht.chord.operations;

import java.math.BigInteger;
import java.util.LinkedList;

import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransMessageCallback;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.common.AbstractOperation;
import de.tud.kom.p2psim.impl.overlay.AbstractOverlayNode.PeerStatus;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordContact;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordKey;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordNode;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.JoinMsg;
import de.tud.kom.p2psim.impl.simengine.Simulator;

/**
 * join a Chord ring containing node "n"
 * 
 * @author Yue Sheng
 * 
 */
public class JoinOperation extends AbstractOperation<ChordNode, Object> implements TransMessageCallback{
	
	ChordNode master = getComponent(); 
	private ChordNode n = null; //A existing node in Chord ring (a node in bootstrap)
	private ChordContact rootNode; // the successor of this new node
	private OperationCallback<Object> callback;
	/**
	 * @param node
	 *            The master node of this operation
	 * @param callback
	 *            The operation listener
	 */
	public JoinOperation(ChordNode node, OperationCallback<Object> callback){
		super(node, callback);
		this.callback=callback;
	}

	@Override
	protected void execute() {
		if(master.getPeerStatus().equals(PeerStatus.ABSENT)){
			System.err.println("peer offline, unable to join.");
			operationFinished(false);
			return;
		} else if(master.getPeerStatus().equals(PeerStatus.PRESENT)){
			System.err.println("peer already online, unable to join.");
			operationFinished(true);
			return;
		}
		master.setPeerStatus(PeerStatus.PRESENT);
		
		n = master.getBootstrap().pick();
		if (n == null) { // create a new network
			new CreateRingOperation(master,callback).scheduleImmediately();
			operationFinished(true);
			return;
		}
		master.setPre(null);
		// get root node on the ring
		ChordKey k = new ChordKey((BigInteger)master.getOverlayID().getUniqueValue());
		n.nodeLookup(k, new OperationCallback<Object>(){

			public void calledOperationFailed(Operation<Object> op) {
				System.out.println("nodeLookup failed in JoinOperation!");
			}

			public void calledOperationSucceeded(Operation<Object> op) {
				rootNode = (ChordContact) op.getResult();
				
				// get succList of root node
				sendRequest(rootNode, "succ",10); 
			}
		});
	}

	@Override
	public Object getResult() {
		return this;
	}

	public void receive(Message msg, TransInfo senderInfo, int commId) {
		if(msg instanceof GetInfoReply){
			GetInfoReply reply = (GetInfoReply) msg;
			if (reply.getCmdID()==10) {
				LinkedList succList = (LinkedList) reply.getResult();
				updateSuccList(rootNode, succList);
				
				master.getBootstrap().registerNode(master);
				master.setAlive(true);
				
				// copy keys from its successor
				JoinMsg join = new JoinMsg(master.getChordID(),rootNode.getOverlayID(),master.getLocalContact());
				master.getTransLayer().send(join, rootNode.getTransInfo(), master.getPort(), TransProtocol.UDP);
				
				master.enableBackgroundOps();
//				System.out.println(getComponent().getChordID()+" ("+getComponent().getHost().getNetLayer().getNetID()+")"+" Join Op finished~~");//test
				operationFinished(true);
			}
		}
	}
	
	public void messageTimeoutOccured(int commId) {
		System.out.println("getSucc timeout in JoinOp~~"); //test
		rootNode.setAlive(false);
		master.setPeerStatus(PeerStatus.TO_JOIN);
		operationFinished(false);
	}
	
	private int sendRequest(ChordContact target, String type, int id) {
		GetInfoRequest req = new GetInfoRequest(master.getChordID(), target.getOverlayID(), type, id);
		return master.getTransLayer().sendAndWait(req, target.getTransInfo(), master.getPort(), TransProtocol.UDP, this, 3 * Simulator.SECOND_UNIT);
	}
	
	private void updateSuccList(ChordContact newNode, LinkedList<ChordContact> succ){
		master.getSucc().clear();
		for (int i = 0; i < succ.size(); i++) {
			master.getSucc().add(succ.get(i).clone());
		}
		master.getSucc().addFirst(newNode);
		master.getSucc().removeLast();
	}
}
