package de.tud.kom.p2psim.impl.overlay.dht.chord.operations;

import java.util.LinkedList;

import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransMessageCallback;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.common.AbstractOperation;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordContact;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordID;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordNode;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.NotifyMsg;
import de.tud.kom.p2psim.impl.simengine.Simulator;

/**
 * Called periodically. Verifies this node's immediate successor, and tells the
 * successor about this node.
 * 
 * @author Yue Sheng
 */
public class StabilizeOperation extends AbstractOperation<ChordNode, Object> implements TransMessageCallback {

	private ChordNode master = getComponent();

	private ChordContact nextSucc, x = null;

	private final int succReq1 = 0, succReq2 = 1, preReq = 2;

	/**
	 * @param node
	 * @param callback
	 */
	public StabilizeOperation(ChordNode node, OperationCallback<Object> callback) {
		super(node, callback);
	}

	@Override
	protected void execute() {
		nextSucc = master.nextActiveSucc();
		
		if(nextSucc!=null){
			// if only one node in the ring, do not stabilize
			if(nextSucc.getOverlayID().equals(master.getChordID())){
				operationFinished(true);
				return;
			}
			// update succList
			sendRequest(nextSucc, "succ", succReq1);
		}
	}

	public void receive(Message msg, TransInfo senderInfo, int commId) {
		if (msg instanceof GetInfoReply) {
			GetInfoReply reply = (GetInfoReply) msg;
			if (reply.getCmdID() == succReq1) {
				LinkedList<ChordContact> succ = (LinkedList<ChordContact>) reply.getResult();
				updateSuccList(nextSucc, succ);
				// get predecessor from nextSucc
				sendRequest(nextSucc, "pre", preReq);

			} else if (reply.getCmdID() == preReq) {
				x = (ChordContact) reply.getResult();
				// check new node
				if (x != null && x.isAlive()) {
					ChordID n = master.getChordID(), successorID = nextSucc.getOverlayID();
					if (x.getOverlayID().between(n, successorID)) {
						// x is a new node between n and its old successor
							
						// Ask x for its list of successors
						sendRequest(x, "succ", succReq2);
					} else {
						// x is not between n and its old successor
							
						// Notify nextSucc, that the master of this operation 
						// is the right predecessor							
						notifyPre(nextSucc);
						operationFinished(true);
					}
					
				} else {
					// x is not alive or null
					
					// Notify nextSucc, that the master of this operation 
					// is the right predecessor
					notifyPre(nextSucc); 
					operationFinished(true);
				}
			} else if (reply.getCmdID() == succReq2) {
				LinkedList<ChordContact> succ = (LinkedList<ChordContact>) reply.getResult();
				// Add the new node to the list of successors
				updateSuccList(x, succ);

				// Notify x that the master of this operation is its new predecessor
				notifyPre(x);
				operationFinished(true);
			}
		}
	
	}

	public void messageTimeoutOccured(int commId) {
		if (commId == succReq1 || commId == preReq) {
			nextSucc.setAlive(false);
//			nextSucc = master.nextActiveSucc();
			execute();
		} else if (commId == succReq2) {
			x = null;
			notifyPre(nextSucc);
			operationFinished(true);
		}

	}

	@Override
	public Object getResult() {
		return this;
	}

	/* tell the target node that the master may be its predecessor */
	private void notifyPre(ChordContact target) {
		NotifyMsg notifyMsg = new NotifyMsg(master.getChordID(), target.getOverlayID(), master.getLocalContact());
		master.getTransLayer().send(notifyMsg, target.getTransInfo(), master.getPort(), TransProtocol.UDP);
	}

	private int sendRequest(ChordContact target, String type, int id) {
		GetInfoRequest req = new GetInfoRequest(master.getChordID(), target.getOverlayID(), type, id);
		return master.getTransLayer().sendAndWait(req, target.getTransInfo(), master.getPort(), TransProtocol.UDP, this, 3 * Simulator.SECOND_UNIT);
	}

	private void updateSuccList(ChordContact newNode, LinkedList<ChordContact> succ) {
		master.getSucc().clear();
		for (int i = 0; i < succ.size(); i++) {
			master.getSucc().add(succ.get(i).clone());
		}
		master.getSucc().addFirst(newNode);
		master.getSucc().removeLast();
		nextSucc = master.nextActiveSucc();
	}
}
