package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components;

import java.util.Collection;

import org.apache.log4j.Logger;


import de.tud.kom.p2psim.api.common.ConnectivityEvent;
import de.tud.kom.p2psim.api.common.Host;
import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.api.overlay.OverlayID;
import de.tud.kom.p2psim.api.overlay.OverlayRoutingTable;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransLayer;
import de.tud.kom.p2psim.impl.common.Operations;
import de.tud.kom.p2psim.impl.overlay.AbstractOverlayNode;
import de.tud.kom.p2psim.impl.overlay.AbstractOverlayNode.PeerStatus;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTable;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.AbstractKademliaOperation.Reason;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.KademliaMsg;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

/**
 * Abstract Kademlia overlay node that implements behaviour common to standard
 * Kademlia, Kandy (from the Chord paper) and hierarchical Kademlia. Some
 * methods have been copied from AbstractOverlayNode which has been the
 * superclass in a previous version.
 * 
 * An overlay node acts as a facade to the outside to represent a participant in
 * the overlay. To the inside, it is a container for a routing table, a local
 * database, a message handler to handle incoming, unrequested messages
 * (messages in response to an operation are handled directly in the requesting
 * operation) and an operation factory that permits to create operations
 * according to the flavour of Kademlia that is currently being used.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 * @author Sebastian Kaune
 */
public abstract class AbstractKademliaNode<T extends KademliaOverlayID>
extends AbstractOverlayNode implements Node<T> {

	protected final static Logger log = SimLogger
	.getLogger(AbstractKademliaNode.class);

	/**
	 * The host that this overlay node is part of.
	 */
	private Host host;

	/**
	 * The contact information for this node (port, network address etc.)
	 */
	private final KademliaOverlayContact<T> ownContact;

	/**
	 * A reference to the transport layer used to send messages.
	 */
	private final TransLayer transLayer;

	/**
	 * The local database (own "shared files").
	 */
	private final KademliaIndexer<T> localIndex;

	/**
	 * The handler of this node's online/offline state.
	 */
	private final StateHandler<T> stateHandler;

	/**
	 * Configuration values ("constants").
	 */
	protected final ComponentsConfig config;

	/**
	 * Constructs an abstract Kademlia node. Initialises commonly used
	 * components such as the local database. <b>The routing table, operation
	 * factory, and request handler have to be initialised (and possibly
	 * registered as listeners) in the subclasses.</b>
	 * 
	 * @param myContact
	 *                the KademliaOverlayContact of the new node.
	 * @param messageManager
	 *                the TransLayer of the new node.
	 * @param conf
	 *                a ComponentsConfig reference that permits to retrieve
	 *                configuration "constants".
	 */
	public AbstractKademliaNode(final KademliaOverlayContact<T> myContact,
			final TransLayer msgMgr, final ComponentsConfig conf) {

		super(myContact.getOverlayID(), myContact.getTransInfo().getPort());
		config = conf;
		ownContact = myContact;
		transLayer = msgMgr;
		stateHandler = new StateHandler<T>(this, conf);
		localIndex = new KademliaIndexer<T>(conf);

		/*
		 * IMPORTANT: The handler for incoming (unsolicited) messages/requests
		 * and new close neighbours has to be constructed in each subclass after
		 * the routing table has been constructed!! That handler also needs to
		 * be registered with the routing table (as proximity listener) and with
		 * the transport layer (as transport message listener).
		 */

		// connect triggered externally
		// initial routing table contacts inserted externally
	}

	/*
	 * "Own" methods
	 */

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final KademliaIndexer<T> getLocalIndex() {
		return this.localIndex;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final KademliaOverlayContact<T> getLocalContact() {
		return this.ownContact;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final TransLayer getMessageManager() {
		return transLayer;
	}

	public TransLayer getTransLayer(){
		return getMessageManager();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void addSenderToRoutingTable(final KademliaMsg<T> msg,
			final TransInfo senderAddr) {
		final KademliaOverlayContact<T> newCon = new KademliaOverlayContact<T>(
				msg.getSender(), senderAddr);
		getKademliaRoutingTable().addContact(newCon);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void addContactsToRoutingTable(
			final Collection<KademliaOverlayContact<T>> contacts) {
		for (final KademliaOverlayContact<T> contact : contacts) {
			getKademliaRoutingTable().addContact(contact);
		}
	}

	/*
	 * Methods from OverlayNode interface
	 */

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final T getTypedOverlayID() {
		return this.ownContact.getOverlayID();
	}

	public OverlayID getOverlayID() {
		return this.ownContact.getOverlayID();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final short getPort() {
		return this.ownContact.getTransInfo().getPort();
	}

	/**
	 * {@inheritDoc}
	 */
//	@Override
	public final Operation<?> createOperation(final String opName,
			final String[] params, final OperationCallback caller) {
		// TODO: remove (not used any more)
		return null;
	}

	/**
	 * Connects this Node to the Kademlia overlay network.
	 */
	public final void connect() {
		stateHandler.doConnect();
	}

	/**
	 * Disconnects this Node from the Kademlia overlay network.
	 */
	public final void disconnect() {
		stateHandler.doDisconnect();
	}

	/**
	 * Begins periodic lookups for random keys. (The lookups are executed only
	 * if this Node is already PRESENT.)
	 */
	public final void beginLookups() {
		stateHandler.enablePeriodicLookup();
	}

	/**
	 * Stores the given data in the Kademlia overlay network.
	 * 
	 * @param key
	 *                a String that will be converted into a KademliaOverlayKey.
	 * @param data
	 *                a String that represents the data item (arbitrary).
	 */
	public final void store(final String key, final String data) {
		getOperationFactory().getStoreOperation(new DHTObject() {
			@Override
			public String toString() {
				return data;
			}
		}, new KademliaOverlayKey(key, config), true, Reason.USER_INITIATED,
		Operations.EMPTY_CALLBACK).scheduleImmediately();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void routingTableBuilt() {
		stateHandler.routingTableBuilt();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void calledOperationFailed(final Operation op) {
		// TODO: remove
		// Simulator.getMonitor().operationFinished(op);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void calledOperationSucceeded(final Operation op) {
		// TODO: remove
		// Simulator.getMonitor().operationFinished(op);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void setHost(final Host host) {
		this.host = host;
		// register as listener for churn-related events
		this.host.getProperties().addConnectivityListener(stateHandler);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final Host getHost() {
		return this.host;
	}

	/*
	 * Methods transfered from BaseOverlayNode
	 */

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final PeerStatus getPeerStatus() {
		return stateHandler.getPeerStatus();
	}

	@Override
	public void connectivityChanged(ConnectivityEvent ce) {
		if (ce.isOffline()) {
			setPeerStatus(PeerStatus.ABSENT);
		} else if (ce.isOnline()) {
			setPeerStatus(PeerStatus.PRESENT);
		}
	}

	public OverlayRoutingTable getRoutingTable(){
		//TODO: Write wrapper for RoutingTable<T>
		throw new UnsupportedOperationException("This overlay does not support this call yet ");
	}

}
