package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components;

import java.math.BigInteger;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.overlay.KBR;
import de.tud.kom.p2psim.api.overlay.KBRListener;
import de.tud.kom.p2psim.api.overlay.OverlayContact;
import de.tud.kom.p2psim.api.overlay.OverlayID;
import de.tud.kom.p2psim.api.overlay.OverlayKey;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransLayer;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.overlay.dht.ForwardMsg;
import de.tud.kom.p2psim.impl.overlay.dht.KBRMsgHandler;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.setup.ConfigStub;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.setup.StaticConfig;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

public class KBRKademliaNode<T extends KademliaOverlayID> extends KademliaNode<T> implements KBR<T, KademliaOverlayContact<T>> {

	private final static Logger log = SimLogger.getLogger(KBRKademliaNode.class);
	
	private KBRListener kbrListener;
	private KBRMsgHandler kbrMsgHandler;
	
	public KBRKademliaNode(KademliaOverlayContact<T> myContact, 
						   TransLayer msgMgr, 
						   ComponentsConfig conf) {
		super(myContact, msgMgr, conf);
		
	}

	@Override
	public boolean isRootOf(OverlayKey key) {
		
		Set<KademliaOverlayContact<T>> closestAroundKey = this.getKademliaRoutingTable().localLookup(
				(KademliaOverlayKey)key, config.getBucketSize());
		
		if (closestAroundKey.contains(getLocalContact())){
			// This node is possibly the root if no other node in his routing table is "nearer"
			
			BigInteger kademliaKeyAsBigInteger = ((KademliaOverlayKey)key).getBigInt();
			
			// Compute the distance of the this node to the key
			BigInteger myDistance = 
				getLocalOverlayContact().getOverlayID().toKey().getBigInt().xor(kademliaKeyAsBigInteger);
			
			// Check if there is a node that is "nearer" than this node itself
			for (KademliaOverlayContact<T> contact : closestAroundKey) {
				BigInteger distance = contact.getOverlayID().toKey().getBigInt().xor(kademliaKeyAsBigInteger);
				if(myDistance.compareTo(distance) == 1){
					return false;
				}
			}
			
			// There was no node that is "nearer", so this node thinks it is the root
			return true;
		}
		return false;
	}

	@Override
	public List<KademliaOverlayContact<T>> local_lookup(OverlayKey key, int num) {
		
		LinkedList<KademliaOverlayContact<T>> contactList  = 
			new LinkedList<KademliaOverlayContact<T>>();
		
		Set<KademliaOverlayContact<T>> contactSet =
			getKademliaRoutingTable().localLookup((KademliaOverlayKey)key, num);
		
		contactList.addAll(contactSet);
		return contactList;
	}

	@Override
	public List<KademliaOverlayContact<T>> neighborSet(int num) {
		LinkedList<KademliaOverlayContact<T>> neighbors = 
			new LinkedList<KademliaOverlayContact<T>>();
		
		// Extract neighbors from routing table by performing a local_lockup 
		// for the key generated from the OverlayID of this node.
		Set<KademliaOverlayContact<T>> neighborSet = 
			getKademliaRoutingTable().localLookup(getTypedOverlayID().toKey(), num);
		
		neighbors.addAll(neighborSet);
		return neighbors;
	}

	@Override
	public OverlayID[] range(KademliaOverlayContact<T> contact, int rank) {
		
		// For the following computation we need the node itself, rank nodes inside the
		// interval and the two exclusive borders of the interval
		int numberOfNeededNodes = rank+3;
		
		Set<KademliaOverlayContact<T>> neighborSet = 
			getKademliaRoutingTable().localLookup(getTypedOverlayID().toKey(), numberOfNeededNodes);
		
		BigInteger xorDistanceToUpper = BigInteger.ZERO;
		BigInteger xorDistanceToLower = BigInteger.ZERO;
		
		BigInteger localKeyAsBigInteger = getTypedOverlayID().toKey().getBigInt();
 		
		// Find the distances to the nodes at the border of the interval
		
		for (KademliaOverlayContact<T> contactInSet : neighborSet) {
			BigInteger contactKeyAsBigInteger = contactInSet.getOverlayID().toKey().getBigInt();
			
			BigInteger distance = localKeyAsBigInteger.xor(contactKeyAsBigInteger);
			
			if(contactKeyAsBigInteger.compareTo(localKeyAsBigInteger) == -1){
				// The contacts key is smaller than the local key
				
				if(distance.compareTo(xorDistanceToLower) == 1){
					// Current distance is the new greatest distance to a lower key
					xorDistanceToLower = distance;
				}
			} else if (contactKeyAsBigInteger.compareTo(localKeyAsBigInteger) == 1){
				// The contacts key is greater than the local key
				
				if(distance.compareTo(xorDistanceToUpper) == 1){
					// Current distance is the new greatest distance to a upper key
					xorDistanceToUpper = distance;
				}
			}
		}
		
		
		/* Compute the range for return
		 * 
		 * Lower bound of the range = 
		 * 				(key of local node) - [(distance to lowest node in set of neighbors)/2]
		 * 
		 * Upper bound of the range = 
		 * 				(key of local node) + [(distance to greatest node in set of neighbors)/2]
		 */
		
		KademliaOverlayID[] range = new KademliaOverlayID[2];
		
		BigInteger lowerBound = 
			localKeyAsBigInteger.subtract(xorDistanceToLower.divide(BigInteger.valueOf(new Long(2))));
		
		BigInteger upperBound = 
			localKeyAsBigInteger.add(xorDistanceToUpper.divide(BigInteger.valueOf(new Long(2))));
		
		range[0] = new KademliaOverlayID(lowerBound, new ConfigStub());
		range[1] = new KademliaOverlayID(upperBound, new ConfigStub());
		
		return range;
	}

	@Override
	public List<KademliaOverlayContact<T>> replicaSet(OverlayKey key, int maxRank) {
		return local_lookup(key, maxRank);
	}

	@Override
	public void route(OverlayKey key, Message msg, KademliaOverlayContact<T> hint) {
		OverlayContact nextHop = null;
		if (hint != null) {
			nextHop = hint;
		} else if (key != null) {
			nextHop = local_lookup(key, 1).get(0);
		} else {
			log.error("Both key and hint are null!!");
			return;
		}
		if(kbrListener != null)
			kbrListener.forward(key, msg, nextHop);
		else
			log.error("There is no KBRListener to notify. Please register your Application (use method: setKBRListener).");
		
		ForwardMsg fm = new ForwardMsg(getOverlayID(), nextHop.getOverlayID(), key, msg);
		getTransLayer().send(fm, nextHop.getTransInfo(), getPort(), TransProtocol.UDP);

	}

	@Override
	public void setKBRListener(KBRListener listener) {
		this.kbrListener = listener;
		kbrMsgHandler = new KBRMsgHandler<T, KademliaOverlayContact<T>>(this, this, kbrListener);
		// add KBR handler to the listeners list
		this.getTransLayer().addTransMsgListener(this.kbrMsgHandler, this.getPort());
	}

	@Override
	public OverlayKey getNewOverlayKey(int rank) {
		return new KademliaOverlayKey(rank, new StaticConfig());
	}

	@Override
	public KademliaOverlayContact<T> getLocalOverlayContact() {
		short port = getHost().getOverlay(KBR.class).getPort();
		TransInfo transInfo = getHost().getTransLayer().getLocalTransInfo(port);
		
		return new KademliaOverlayContact<T>(getTypedOverlayID(), transInfo);
	}

}
