package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components;

import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.impl.simengine.Simulator;

/**
 * A KademliaIndexEntry groups a DHTObject together with the point in time at
 * which it has last been republished (or received) at the node that owns the
 * KademliaIndexer in which this entry is saved.
 * 
 * @author Sebastian Kaune
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
final class KademliaIndexEntry {

    /** The DHTObject (data item) to be saved. */
    private final DHTObject value;

    /**
     * The point in time at which the item associated with this
     * KademliaIndexEntry has last been republished (or received) at the node
     * that owns the index. A negative value (smaller than zero) means that it
     * has never been published.
     */
    private long lastRepublish = -1;

    /**
     * Configuration values ("constants").
     */
    protected final ComponentsConfig config;

    /**
     * Constructs a new KademliaIndexEntry that groups a DHTObject together with
     * the point in time at which it has last been republished (or received) at
     * the node that owns the KademliaIndexer in which this entry is saved. The
     * last republish time is initially set to "not republished (-1)" and has to
     * be set manually.
     * 
     * @param value
     *                the DHTObject (data item) to be saved.
     * @param conf
     *                a ComponentsConfig reference that permits to retrieve
     *                configuration "constants".
     */
    public KademliaIndexEntry(final DHTObject value, final ComponentsConfig conf) {
	this.config = conf;
	this.value = value;
    }

    /**
     * Returns the value of this index entry.
     * 
     * @return the DHTObject that is saved in this entry.
     */
    public final DHTObject getValue() {
	return this.value;
    }

    /**
     * Returns the last republish time of this entry. This method is for test
     * purposes only!
     * 
     * @return the point in time at which the item associated with this
     *         KademliaIndexEntry has last been republished (or received) at the
     *         node that owns the index (in simulation time units). A value
     *         smaller than zero means that the data item has never been
     *         published or received from another peer.
     */
    protected final long getLastRepublish() {
	return lastRepublish;
    }

    /**
     * Updates the last republish time by setting it to the current simulation
     * time. A republish takes place if either this or another peer sends the
     * data item to its neighbours.
     */
    public final void updateLastRepublish() {
	lastRepublish = Simulator.getCurrentTime();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String toString() {
	return "index entry (value=" + this.value + ", lastRepublish="
		+ this.lastRepublish + ")";
    }

    /**
     * @return whether this index entry has expired, that is, whether the last
     *         republish time is more than or exactly
     *         {@link OperationsConfig#getDataExpirationTime()} ago. Items that
     *         have never been published do not expire (it is expected that they
     *         are published at some point in time from which on they expire
     *         normally).
     */
    public final boolean hasExpired() {
	if (lastRepublish < 0) {
	    return false;
	} else if (lastRepublish + config.getDataExpirationTime() <= Simulator
		.getCurrentTime()) {
	    return true;
	}
	return false;
    }

    /**
     * @return whether this index entry needs to be republished because the last
     *         republish time is more than or exactly
     *         {@link OperationsConfig#getRepublishInterval()} ago or has never
     *         been (re)published at all.
     */
    public final boolean needsRepublish() {
	if (lastRepublish < 0
		|| lastRepublish + config.getRepublishInterval() <= Simulator
			.getCurrentTime()) {
	    return true;
	}
	return false;
    }
}
