package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components;

import java.util.Map;
import java.util.Set;


import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.api.transport.TransInfo;
import de.tud.kom.p2psim.api.transport.TransLayer;
import de.tud.kom.p2psim.api.transport.TransMessageListener;
import de.tud.kom.p2psim.api.transport.TransMsgEvent;
import de.tud.kom.p2psim.api.transport.TransProtocol;
import de.tud.kom.p2psim.impl.overlay.AbstractOverlayNode.PeerStatus;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTable.ProximityListener;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.AbstractKademliaOperation.Reason;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.DataLookupMsg;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.DataMsg;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.KClosestNodesLookupMsg;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.KademliaMsg;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.NodeListMsg;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

/**
 * A handler for incoming (unsolicited) messages/requests for a Kademlia node.
 * 
 * @author Sebastian Kaune
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
class RequestHandler<T extends KademliaOverlayID> implements
TransMessageListener, ProximityListener<T> {

	/**
	 * A logger for this class.
	 */
	final static Logger log = SimLogger.getLogger(RequestHandler.class);

	/**
	 * The Node that owns this message handler.
	 */
	private final Node<T> myNode;

	/**
	 * Reference to the transport layer used to send messages.
	 */
	protected final TransLayer transLayer;

	/**
	 * Configuration values ("constants").
	 */
	protected final ComponentsConfig config;

	/**
	 * Constructs a new standard message handler (non-hierarchical). It has to
	 * be manually registered as ProximityListener of <code>myNode</code>'s
	 * routing table and as a TransMessageListener of <code>manager</code>.
	 * 
	 * @param manager
	 *                the TransLayer used to reply to messages.
	 * @param myNode
	 *                the node that owns this message handler.
	 * @param conf
	 *                an ComponentsConfig reference that permits to retrieve
	 *                configuration "constants".
	 */
	public RequestHandler(final TransLayer manager, final Node<T> myNode,
			final ComponentsConfig conf) {
		this.config = conf;
		this.transLayer = manager;
		this.myNode = myNode;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void messageArrived(final TransMsgEvent tme) {
		final Message msg = tme.getPayload();

		// ignore messages if peer is offline
		if (myNode.getPeerStatus() == PeerStatus.ABSENT) {
			log.debug("Ignoring msg=" + msg + " at node=" + myNode);
		} else {
//			log.debug("Received msg=" + msg + " at node=" + myNode);
			processMsg(tme);
		}
	}

	/**
	 * Processes the given TransMsgEvent. As a precondition, it can be assumed
	 * that this peer is not offline (that is, the event is actually to be
	 * processed).
	 * 
	 * @param tme
	 *                the TransMsgEvent that corresponds to the message that has
	 *                been received.
	 */
	@SuppressWarnings("unchecked")
	protected void processMsg(final TransMsgEvent tme) {
		final Message msg = tme.getPayload();
		final TransInfo senderAddr = tme.getSenderTransInfo();

		if(msg instanceof KademliaMsg){
			myNode.addSenderToRoutingTable((KademliaMsg<T>) msg, senderAddr);
	
			if (msg instanceof KClosestNodesLookupMsg) {
				processNodeLookup((KClosestNodesLookupMsg<T>) msg, tme);
			} else if (msg instanceof DataLookupMsg) {
				processDataLookup((DataLookupMsg<T>) msg, tme);
			} else if (msg instanceof DataMsg) {
				processDataStore((DataMsg<T>) msg);
			} else {
				throw new IllegalStateException("[" + this.myNode.getTypedOverlayID()
						+ "] Unknown message received: " + msg);
			}
		}
	}

	/**
	 * Handles reception of a (non-hierarchical) node lookup message.
	 * 
	 * @param msg
	 *                the KClosestNodesLookupMsg that has been received. (This
	 *                is a copy of the message contained in
	 *                <code>receiveEvent</code>.)
	 * @param receiveEvent
	 *                the TransMsgEvent that occurred at reception of the
	 *                message that is to be handled here.
	 */
	private final void processNodeLookup(final KClosestNodesLookupMsg<T> msg,
			final TransMsgEvent receiveEvent) {
		final KademliaOverlayKey key = msg.getNodeKey();
		final Set<KademliaOverlayContact<T>> neighbours = nodeLookupForNode(key);
		final NodeListMsg<T> reply = new NodeListMsg<T>(this.myNode
				.getTypedOverlayID(), msg.getSender(), neighbours, msg.getReason(),
				config);
//		log.debug("Replying to node lookup (key=" + key + "; result="
//		+ neighbours + "; node=" + myNode + ")");
		transLayer.sendReply(reply, receiveEvent, myNode.getPort(),
				TransProtocol.UDP);
	}

	/**
	 * Looks up {@link ComponentsConfig#getBucketSize()} contacts to be returned
	 * if a node lookup message comes in.
	 * 
	 * @param key
	 *                the KademliaOverlayKey to be looked up.
	 * @return a Set with K contacts from the local routing table.
	 */
	protected Set<KademliaOverlayContact<T>> nodeLookupForNode(
			final KademliaOverlayKey key) {
		return myNode.getKademliaRoutingTable()
		.localLookup(key, config.getBucketSize());
	}

	/**
	 * Handles reception of a (non-hierarchical) data lookup message.
	 * 
	 * @param msg
	 *                the DataLookupMsg that has been received. (This is a copy
	 *                of the message contained in <code>receiveEvent</code>.)
	 * @param receiveEvent
	 *                the TransMsgEvent that occurred at reception of the
	 *                message that is to be handled here.
	 */
	private final void processDataLookup(final DataLookupMsg<T> msg,
			final TransMsgEvent receiveEvent) {
		final KademliaOverlayKey key = msg.getDataKey();
		final KademliaMsg<T> reply;

		// Checks if TimestampedValue is located in local index or not
		final DHTObject value = myNode.getLocalIndex().get(key);
		if (value != null) {
//			log.debug("Replying to data lookup with data (key=" + key
//			+ "; result=" + value + "; node=" + myNode + ")");
			reply = new DataMsg<T>(myNode.getTypedOverlayID(), msg.getSender(), key,
					value, msg.getReason(), config);
		} else {
			// Return message with list of k closest overlay contacts
			final Set<KademliaOverlayContact<T>> neighbours = nodeLookupForData(key);
//			log.debug("Replying to data lookup with node list (key=" + key
//			+ "; result=" + neighbours + "; node=" + myNode + ")");
			reply = new NodeListMsg<T>(myNode.getTypedOverlayID(), msg.getSender(),
					neighbours, msg.getReason(), config);
		}
		transLayer.sendReply(reply, receiveEvent, myNode.getPort(),
				TransProtocol.UDP);
	}

	/**
	 * Looks up {@link ComponentsConfig#getBucketSize()} contacts to be returned
	 * if a data lookup message comes in.
	 * 
	 * @param key
	 *                the KademliaOverlayKey to be looked up.
	 * @return a Set with K contacts from the local routing table.
	 */
	protected Set<KademliaOverlayContact<T>> nodeLookupForData(
			final KademliaOverlayKey key) {
		return myNode.getKademliaRoutingTable()
		.localLookup(key, config.getBucketSize());
	}

	/**
	 * Handles reception of an unsolicited DataMsg (used to store data on this
	 * node).
	 * 
	 * @param msg
	 *                the DataMsg that has been received.
	 */
	private final void processDataStore(final DataMsg<T> msg) {
		final KademliaOverlayKey key = msg.getKey();
		final DHTObject receivedValue = msg.getValue();

		// log.debug("Storing data (key=" + key + "; data=" + receivedValue
		// + "; node=" + myNode + ")");
		myNode.getLocalIndex().put(key, receivedValue);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void newCloseContactArrived(
			final KademliaOverlayContact<T> newContact) {
		final Map<KademliaOverlayKey, DHTObject> allItems;
		Set<KademliaOverlayContact<T>> closestAroundKey;
		DataMsg<T> message;

		/*
		 * A new close neighbour newContact has been seen - it has to be
		 * informed about the data items that it has to store. However, this
		 * method is only executed if this node (myNode) is PRESENT, that is, if
		 * its routing table is in a relatively stable state. (Else, in state
		 * TO_JOIN for instance, we cannot be sure that newContact is actually a
		 * close neighbour. There could as well be closer neighbours that just
		 * are not already known.)
		 */
		if (!myNode.getPeerStatus().equals(PeerStatus.PRESENT)) {
			return;
		}

		// get all entries from local database
		allItems = myNode.getLocalIndex().getEntries();

		/*
		 * Send all items that the new node has to store: entries that both this
		 * and the new node are responsible for. "Responsible for" is defined as
		 * being part of the K closest nodes around the key of the data item.
		 * This is determined with the knowledge of the local routing table. (We
		 * have to obtain an unfiltered view for the case of Kandy.)
		 */
		for (final Map.Entry<KademliaOverlayKey, DHTObject> item : allItems
				.entrySet()) {
			closestAroundKey = myNode.getKademliaRoutingTable().localLookup(
					item.getKey(), config.getBucketSize());

			// continue with next item if one of own/new node not responsible
			// for data. assumption: no visibility restriction
			if (!closestAroundKey.contains(myNode.getLocalContact())
					|| !closestAroundKey.contains(newContact)) {
				continue;
			}

			// else send data
			message = new DataMsg<T>(myNode.getTypedOverlayID(), newContact
					.getOverlayID(), item.getKey(), item.getValue(),
					Reason.MAINTENANCE, config);
			myNode.getMessageManager().send(message, newContact.getTransInfo(),
					myNode.getPort(), TransProtocol.UDP);
		}
	}

}
