package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable;

import java.util.Collections;
import java.util.Comparator;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;


/**
 * Visitor to mark unresponsive contacts.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
final class MarkUnresponsiveNodeVisitor<T extends KademliaOverlayID> extends
	AbstractNodeVisitor<T> {

    /**
     * Single, final GenericLookupNodeVisitor instance (non-reentrant).
     */
    private static final MarkUnresponsiveNodeVisitor singleton = new MarkUnresponsiveNodeVisitor();

    /**
     * The unresponsive contact's ID that is to be marked.
     */
    private T unresponsive;

    /**
     * A comparator that selects the node from the cache that is to replace an
     * evicted one (the maximum with respect to the comparator will be
     * selected).
     */
    private Comparator<? super RoutingTableEntry<T>> replacementComparator;

    /**
     * The proximity handler has to be notified whenever a contact is added to
     * or deleted from a kBucket (not replacement cache).
     */
    private ProximityHandler<T> proximityHandler;

    /**
     * Configuration values ("constants").
     */
    private RoutingTableConfig config;

    /**
     * Marks a contact as unresponsive and removes it from the bucket or
     * replacement cache if the threshold of STALE_COUNTER={@link RoutingTableConfig#getStaleCounter()}
     * marks has been reached.
     * <p>
     * A contact from the bucket with at least <code>STALE_COUNTER</code>
     * marks will be replaced with a contact from the replacement cache (the
     * maximum with respect to <code>replacementComparator</code> will be
     * chosen). However, if the replacement cache is empty, the contact will not
     * be removed from the bucket.
     * <p>
     * A contact from the replacement cache is immediately dropped if the
     * replacement cache is full. Else, it is removed after having received
     * <code>STALE_COUNTER</code> marks.
     * 
     * @param unresponsive
     *                the <code>KademliaOverlayID</code> of the node that is
     *                to be marked as unresponsive.
     * @param replacementComparator
     *                a comparator that selects the node from the cache that is
     *                to replace an evicted node from the bucket (the <i>maximum</i>
     *                with respect to the comparator will be selected). If this
     *                argument is <code>null</code>, the most recently seen
     *                contact from the cache will be chosen.
     * @param proxHandler
     *                a ProximityHandler that will be notified if a contact is
     *                added to or removed from a LeafNode's kBucket (not
     *                replacement cache). If <code>proxHandler</code> is
     *                <code>null</code>, it will be ignored.
     * @param conf
     *                a RoutingTableConfig reference that permits to retrieve
     *                configuration "constants".
     * @return a MarkUnresponsiveNodeVisitor instance. Note that this instance
     *         is statically shared among all clients of this class. That is, at
     *         runtime only one MarkUnresponsiveNodeVisitor instance exists.
     *         Thus, it is non-reentrant and should not be saved by clients
     *         (should used immediately).
     */
    protected static final <T extends KademliaOverlayID> MarkUnresponsiveNodeVisitor<T> getMarkUnresponsiveNodeVisitor(
	    final T unresponsive,
	    final Comparator<? super RoutingTableEntry<T>> replacementComparator,
	    final ProximityHandler<T> proxHandler, final RoutingTableConfig conf) {
	singleton.config = conf;
	singleton.unresponsive = unresponsive;
	if (replacementComparator == null) {
	    singleton.replacementComparator = RoutingTableComparators
		    .getLastSeenComp();
	} else {
	    singleton.replacementComparator = replacementComparator;
	}
	singleton.proximityHandler = proxHandler;
	return singleton;
    }

    private MarkUnresponsiveNodeVisitor() {
	// should not be called externally
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void visit(final BranchNode<T> node) {
	node.getResponsibleChild(unresponsive.getBigInt()).accept(this);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void visit(final LeafNode<T> node) {
	RoutingTableEntry<T> contact, replacement;

	if ((contact = node.kBucket.get(unresponsive)) != null
		&& contact.increaseStaleCounter() >= config.getStaleCounter()
		&& node.replacementCache.size() > 0) {
	    node.kBucket.remove(unresponsive);
	    replacement = Collections.max(node.replacementCache.values(),
		    this.replacementComparator);
	    node.kBucket.put(replacement.getContact().getOverlayID(),
		    replacement);
	    node.replacementCache.remove(replacement.getContact()
		    .getOverlayID());
	    if (proximityHandler != null) {
		proximityHandler.contactRemoved(unresponsive);
		proximityHandler.contactAdded(replacement.getContact());
	    }
	} else if ((contact = node.replacementCache.get(unresponsive)) != null
		&& (contact.increaseStaleCounter() >= config.getStaleCounter() || node.replacementCache
			.size() == config.getReplacementCacheSize())) {
	    node.replacementCache.remove(unresponsive);
	}
    }

}
