package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable;

import java.math.BigInteger;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;


/**
 * This class represents a pseudo routing tree root. Pseudo because it is not
 * the actual root of the tree, but "a node above the root". It internally keeps
 * a reference to the current root node. This class is necessary to allow the
 * actual root to replace itself with another, new root (for example when
 * splitting a bucket) in a polymorphic manner.
 * 
 * Furthermore, this class keeps information that is common to all nodes in the
 * tree, such as the owner's identifier.
 * 
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
class PseudoRootNode<T extends KademliaOverlayID> implements ParentNode<T> {

    /**
     * A reference to the root node of the routing tree. May be replaced by
     * child nodes, for example when a bucket is split and has to be replaced
     * with a branching.
     */
    protected Node<T> root;

    /**
     * The identifier of this routing tree's owner.
     */
    private final T ownID;

    /**
     * Constructs a new PseudoRootNode (and the underlying routing tree with
     * initially one bucket).
     * 
     * @param ownID
     *                the identifier of the owner of this routing table.
     * @param conf
     *                a RoutingTableConfig reference that permits to retrieve
     *                configuration "constants".
     */
    protected PseudoRootNode(final T ownID, final RoutingTableConfig conf) {
	this.ownID = ownID;
	this.root = new LeafNode<T>(this, conf);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final T getOwnID() {
	return this.ownID;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void registerNewChild(final Node<T> child) {
	this.root = child;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final int getLevel() {
	// we are one level atop of the true root
	return -1;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void accept(final NodeVisitor<T> visitor) {
	visitor.visit(this);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final BigInteger getPrefix() {
	return BigInteger.ZERO;
    }

}
