package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;

/**
 * Sets the last lookup time (lookup over the network) for a bucket specified by
 * a KademliaOverlayID that the bucket is responsible for.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
final class SetLastLookupTimeNodeVisitor<T extends KademliaOverlayID> extends
	AbstractNodeVisitor<T> {

    /**
     * Single, final SetLastLookupTimeNodeVisitor instance (non-reentrant).
     */
    private static final SetLastLookupTimeNodeVisitor singleton = new SetLastLookupTimeNodeVisitor();

    /**
     * The key that has been looked up, that is it lies in the bucket that is to
     * be updated.
     */
    private KademliaOverlayKey key;

    /**
     * The point in time of the last lookup in the specified bucket.
     */
    private long lastLookupTime;

    /**
     * Sets the last lookup time of the bucket that contains <code>key</code>
     * to <code>lastLookupTime</code>. ("Lookup" refers to a lookup that has
     * been carried out over the network.)
     * 
     * @param key
     *                a KademliaOverlayID from the bucket that is to be updated.
     * @param lastLookupTime
     *                the new "lastLookupTime" of the bucket that contains
     *                <code>key</code>.
     * @return an SetLastLookupTimeNodeVisitor instance. Note that this instance
     *         is statically shared among all clients of this class. That is, at
     *         runtime only one SetLastLookupTimeNodeVisitor instance exists.
     *         Thus, it is non-reentrant and should not be saved by clients
     *         (should used immediately).
     */
    protected static final <T extends KademliaOverlayID> SetLastLookupTimeNodeVisitor<T> getSetLastLookupTimeNodeVisitor(
	    final KademliaOverlayKey key, final long lastLookupTime) {
	singleton.key = key;
	singleton.lastLookupTime = lastLookupTime;
	return singleton;
    }

    private SetLastLookupTimeNodeVisitor() {
	// should not be called externally
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void visit(final BranchNode<T> node) {
	node.getResponsibleChild(key.getBigInt()).accept(this);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void visit(final LeafNode<T> node) {
	node.lastLookup = this.lastLookupTime;
    }

}
