package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.measurement;

import java.math.BigDecimal;
import java.math.RoundingMode;

/**
 * Permits to keep a running total as well as the number of samples and to
 * calculate the average.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
final class AvgAccumulator {

    /** the number of samples so far */
    private long count = 0;

    /** the running total (sum of all samples) */
    private BigDecimal sum = BigDecimal.ZERO;

    /** the minimum and maximum amount added in all samples. */
    private double min = 0, max = 0;

    /**
     * Whether the first sample has been added, that is, whether min and max
     * have correct values.
     */
    private boolean firstValueAdded = false;

    /**
     * Adds the sample <code>amount</code> to the current running total and
     * increases the number of samples by one.
     * 
     * @param amount
     *                the value to be added.
     */
    public final void addToTotal(final double amount) {
	sum = sum.add(BigDecimal.valueOf(amount));
	count++;
	if (!firstValueAdded) {
	    firstValueAdded = true;
	    min = amount;
	    max = amount;
	} else if (amount < min) {
	    min = amount;
	} else if (amount > max) {
	    max = amount;
	}
    }

    /**
     * Calculates the average amount per sample.
     * 
     * @return the average of all samples.
     */
    public final double getAverage() {
	if (count == 0) {
	    return 0;
	}
	return sum.divide(BigDecimal.valueOf(count), 10, RoundingMode.CEILING)
		.doubleValue();
    }

    /**
     * @return the total number of samples.
     */
    public final long getCount() {
	return count;
    }

    /**
     * @return the minimum amount that has been added in one sample.
     */
    public final double getMin() {
	return min;
    }

    /**
     * @return the maximum amount that has been added in one sample.
     */
    public final double getMax() {
	return max;
    }
}
