package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.measurement;

import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;


import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.analyzer.Analyzer.TransAnalyzer;
import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.KademliaMsg;
import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.transport.AbstractTransMessage;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

/**
 * Accumulates the average amount of bytes sent at the application layer of a
 * Kademlia overlay network. These bytes are classified into bytes sent for
 * maintenance purposes and user-initiated respectively. A measurement is made
 * for each hour of simulated realtime (this analyzer assumes that each
 * measurement is started at the first minute of the hour, and takes an integral
 * number of hours).
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class HourlyTrafficAnalyzer implements TransAnalyzer {
    private final static Logger log = SimLogger
	    .getLogger(HourlyTrafficAnalyzer.class);

    /** Whether the analysis has begun. */
    private boolean started = false;

    /**
     * A Map with the hour of measurement as keys, and the total number of bytes
     * sent as values.
     */
    private Map<Long, Long> sentUser, sentMaintenance;

    /**
     * The number of hosts in the overlay network.
     */
    private double numHosts = 1;

    /**
     * Sets the number of hosts that take part in the overlay network. If not
     * set, the default value of 1 host is assumed. Has to be set before
     * analysis is started and will be ignored else.
     * 
     * @param numberOfHosts
     *                the number of hosts that take part in the overlay network.
     */
    public final void setNumberOfHosts(final int numberOfHosts) {
	if (!started) {
	    numHosts = numberOfHosts;
	}
    }

    /**
     * Called to start the traffic analysis. Messages that are sent prior to
     * invoking this method are not taken into account for analysis. Calling
     * this method starts a <i>fresh</i> measurement. The number of the
     * participating hosts has to be set before calling this method.
     */
    public final void start() {
	if (!started) {
	    started = true;
	    sentMaintenance = new HashMap<Long, Long>();
	    sentUser = new HashMap<Long, Long>();
	}
    }

    /**
     * A Kademlia overlay message has been sent at a host.
     * 
     * @param msg
     *                the KademliaMsg that has been sent.
     */
    private void messageSent(final KademliaMsg<?> msg) {
	Map<Long, Long> bytes = null;
	Long currBytes = null;
	final Long currentHour;
	if (!started) {
	    return;
	}

	switch (msg.getReason()) {
	case MAINTENANCE:
	    bytes = sentMaintenance;
	    break;
	case USER_INITIATED:
	    bytes = sentUser;
	    break;
	}
	if (bytes == null) {
	    return;
	}
	currentHour = Simulator.getCurrentTime() / Simulator.HOUR_UNIT;
	currBytes = bytes.get(currentHour);
	if (currBytes == null) {
	    currBytes = msg.getSize();
	} else {
	    currBytes += msg.getSize();
	}
	bytes.put(currentHour, currBytes);
    }

    /**
     * Called to stop the traffic analysis.
     * 
     * @param output
     *                a Writer on which a textual representation of the
     *                measurements will be written.
     */
    public final void stop(final Writer output) {
	if (!started) {
	    return;
	}
	started = false;

	try {
	    output.write("\n**************** Application Message Statistics "
		    + "****************\n");
	    output.write(" Hour\tTotal Bytes Sent\tBytes Sent Per Host\n");
	    output.write("-- user-initiated --\n");
	    writeTraffic(output, sentUser);
	    output.write("\n-- maintenance --\n");
	    writeTraffic(output, sentMaintenance);
	    output.write(" (Assumed " + numHosts
		    + " hosts and measurement of entire ");
	    output.write("hours starting at minute 0.)\n");
	    output.write("************** Application Message Statistics End "
		    + "**************\n");
	    output.flush();
	} catch (IOException ex) {
	    log.error("Overlay traffic measurement could not be written to "
		    + "output.", ex);
	}
    }

    private void writeTraffic(final Writer output, final Map<Long, Long> bytes)
	    throws IOException {
	for (final Map.Entry<Long, Long> entry : bytes.entrySet()) {
	    output.write(String.format(" %1$02d \t%2$016d \t%3$09.2f\n",
		    entry.getKey(), entry.getValue(), entry.getValue()
			    / numHosts));
	}
    }

    /*
     * Methods used to "adapt" this class to the TransAnalyzer interface:
     */

    /**
     * {@inheritDoc}
     */
    @Override
    public final void transMsgReceived(final AbstractTransMessage msg) {
	// ignore
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void transMsgSent(final AbstractTransMessage msg) {
	final Message payload = msg.getPayload();
	if (payload != null && payload instanceof KademliaMsg<?>) {
	    messageSent((KademliaMsg<?>) payload);
	}
    }
}
