package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations;

import java.util.Map;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;
import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;


/**
 * An operation that issues bucket lookups for those buckets that need to be
 * refreshed.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 * @author Sebastian Kaune
 */
public class BucketRefreshOperation<T extends KademliaOverlayID> extends
	AbstractKademliaOperation<T> {

    final static Logger log = SimLogger.getLogger(BucketRefreshOperation.class);

    /**
     * Whether a refresh is forced for all buckets.
     */
    private final boolean forceRefresh;

    /**
     * The operation identifier of the currently executing bucket lookup
     * operation.
     */
    private int currentBucketLookupOperationID;

    /**
     * A Map that contains those buckets that still need to be refreshed.
     */
    private Map<KademliaOverlayKey, Integer> refreshBuckets;

    /**
     * Constructs a new bucket refresh operation. This operation is always for
     * maintenance purposes.
     * 
     * @param forceRefresh
     *                if true, all buckets have to be refreshed, no matter their
     *                last lookup time.
     * @param node
     *                the AbstractKademliaNode that initiates this lookup.
     * @param opCallback
     *                a callback that is informed when this operation
     *                terminates.
     * @param conf
     *                a OperationsConfig reference that permits to retrieve
     *                configuration "constants".
     */
    public BucketRefreshOperation(final boolean forceRefresh,
	    final Node<T> node, final OperationCallback opCallback,
	    final OperationsConfig conf) {
	super(node, opCallback, Reason.MAINTENANCE, conf);
	this.forceRefresh = forceRefresh;
	currentBucketLookupOperationID = -1;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void execute() {
	final long notLookedUpAfter;
//	log.debug("Started bucket refresh (node=" + getComponent()); // TODO uncomment

	if (forceRefresh) {
	    notLookedUpAfter = Simulator.getCurrentTime();
	} else {
	    notLookedUpAfter = Simulator.getCurrentTime()
		    - config.getRefreshInterval();
	}

	refreshBuckets = getComponent().getKademliaRoutingTable().getRefreshBuckets(
		notLookedUpAfter);
	// all bucket lookups will be executed one after another
	scheduleOperationTimeout(config.getLookupOperationTimeout()
		* refreshBuckets.size() + 1);
	issueBucketLookup();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void calledOperationDidFail(final Operation op) {
	// check whether event was caused by the operation we are waiting for
	if (currentBucketLookupOperationID == op.getOperationID()) {
	    // the previous lookup failed - we carry on with the next one
	    issueBucketLookup();
	}
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void calledOperationDidSucceed(final Operation op) {
	// check whether event was caused by the operation we are waiting for
	if (currentBucketLookupOperationID == op.getOperationID()) {
	    issueBucketLookup();
	}
    }

    /**
     * Issues a bucket lookup for a bucket from <code>refreshBuckets</code>.
     * The bucket for which the lookup has been scheduled will be removed from
     * <code>refreshBuckets</code> and the operation identifier of the lookup
     * operation will be saved in <code>currentBucketLookupOperationID</code>.
     * If there is no bucket left to look up, this refresh operation will be
     * finished successfully.
     */
    private final void issueBucketLookup() {
	// key of entry is bucket key, value is bucket depth
	final Map.Entry<KademliaOverlayKey, Integer> currentBucket;
	final KademliaOperation<T> bucketLookup;

	if (refreshBuckets.size() > 0) {
	    currentBucket = refreshBuckets.entrySet().iterator().next();
	    bucketLookup = getComponent().getOperationFactory()
		    .getBucketLookupOperation(currentBucket.getKey(),
			    currentBucket.getValue(), this);
	    currentBucketLookupOperationID = bucketLookup.getOperationID();
	    bucketLookup.scheduleImmediately();
	    refreshBuckets.remove(currentBucket.getKey());
	} else {
	    // no more buckets to refresh (this operation cannot fail)
//	    log.debug("Finished bucket refresh (node=" + getComponent()); // TODO uncomment
	    finishOperation(OperationState.SUCCESS);
	}
    }

}
