package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations;

import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.HKademlia2Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.AbstractKademliaOperation.Reason;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.ClusterPreferenceLookupCoordinator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.LookupCoordinator.NonhierarchicalLookupCoordinator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.HKademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;

/**
 * Operations as used in hierarchical Kademlia, that is mainly hierarchy-aware
 * routing. Instead of a customised bucket refresh lookup, a standard
 * (hierarchical) k closest nodes lookup is used.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class HKademlia2OperationFactory<H extends HKademliaOverlayID> extends
	AbstractOperationFactory<H> implements OperationFactory<H> {

    /**
     * The node that owns this operation factory
     */
    private final HKademlia2Node<H> myNode;

    /**
     * Constructs a new hierarchical Kademlia operation factory for the given
     * HKademliaNode.
     * 
     * @param myNode
     *                the HKademlia2Node that owns this operation factory.
     * @param conf
     *                an OperationsConfig reference that permits to retrieve
     *                configuration "constants".
     */
    public HKademlia2OperationFactory(final HKademlia2Node<H> myNode,
	    final OperationsConfig conf) {
	super(conf);
	this.myNode = myNode;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final KademliaOperation<H> getBucketLookupOperation(
	    final KademliaOverlayKey key, final int bucketDepth,
	    final OperationCallback opCallback) {
	return getKClosestNodesLookupOperation(key, Reason.MAINTENANCE,
		opCallback);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final DataLookupOperation<H> getDataLookupOperation(
	    final KademliaOverlayKey key, final OperationCallback opCallback) {
	/*
	 * Although HKademliaOverlayIDs are used, no use is made of hierarchical
	 * lookup messages (with restriction to minimum common cluster depth)
	 * --> use NonhierarchicalLookupCoordinator
	 */
	final NonhierarchicalLookupCoordinator<H> coord = new ClusterPreferenceLookupCoordinator<H>(
		key, myNode, config);
	final DataLookupOperation<H> op = new de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.DataLookupOperation<H>(
		coord, myNode, opCallback, Reason.USER_INITIATED, config);
	return op;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final NodeLookupOperation<H> getKClosestNodesLookupOperation(
	    final KademliaOverlayKey key, final Reason why,
	    final OperationCallback opCallback) {
	/*
	 * Although HKademliaOverlayIDs are used, no use is made of hierarchical
	 * lookup messages (with restriction to minimum common cluster depth)
	 * --> use NonhierarchicalLookupCoordinator
	 */
	final NonhierarchicalLookupCoordinator<H> coord = new ClusterPreferenceLookupCoordinator<H>(
		key, myNode, config);
	final NodeLookupOperation<H> op = new de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.KClosestNodesLookupOperation<H>(
		coord, myNode, opCallback, why, config);
	return op;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final KademliaOperation<H> getRepublishOperation(
	    final OperationCallback opCallback) {
	final KademliaOperation<H> repubOp = new RepublishOperation<H>(myNode,
		opCallback, config);
	return repubOp;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final KademliaOperation<H> getStoreOperation(final DHTObject data,
	    final KademliaOverlayKey key, final boolean storeForeignData,
	    final Reason why, final OperationCallback opCallback) {
	final KademliaOperation<H> storeOp = new StoreOperation<H>(key, data,
		storeForeignData, myNode, opCallback, why, config);
	return storeOp;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final KademliaOperation<H> getRefreshOperation(
	    final boolean forceRefresh, final OperationCallback opCallback) {
	return new BucketRefreshOperation<H>(forceRefresh, myNode, opCallback,
		config);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final KademliaOperation<H> getBuildRoutingTableOperation(
	    final OperationCallback opCallback) {
	return new RoutingTableBuildOperation<H>(myNode, opCallback, config);
    }

}
