package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations;

import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.AbstractKademliaOperation.Reason;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.StandardLookupCoordinator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.LookupCoordinator.NonhierarchicalLookupCoordinator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;

/**
 * Operations as used in standard Kademlia.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class KademliaOperationFactory<T extends KademliaOverlayID> extends
AbstractOperationFactory<T> implements OperationFactory<T> {

	/**
	 * The node that owns this operation factory
	 */
	private final Node<T> myNode;

	/**
	 * Constructs a new standard Kademlia operation factory for the given
	 * KademliaNode.
	 * 
	 * @param myNode
	 *                the Node that owns this operation factory.
	 * @param conf
	 *                an OperationsConfig reference that permits to retrieve
	 *                configuration "constants".
	 */
	public KademliaOperationFactory(final Node<T> myNode,
			final OperationsConfig conf) {
		super(conf);
		this.myNode = myNode;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	@SuppressWarnings("unused")
	public final KademliaOperation<T> getBucketLookupOperation(
			final KademliaOverlayKey key, final int bucketDepth,
			final OperationCallback opCallback) {
		return getKClosestNodesLookupOperation(key, Reason.MAINTENANCE,
				opCallback);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final DataLookupOperation<T> getDataLookupOperation(
			final KademliaOverlayKey key, final OperationCallback opCallback) {
		final NonhierarchicalLookupCoordinator<T> coord = new StandardLookupCoordinator<T>(
				key, myNode, config);
		final DataLookupOperation<T> op = new de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.DataLookupOperation<T>(
				coord, myNode, opCallback, Reason.USER_INITIATED, config);
		return op;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final NodeLookupOperation<T> getKClosestNodesLookupOperation(
			final KademliaOverlayKey key, final Reason why,
			final OperationCallback opCallback) {
		final NonhierarchicalLookupCoordinator<T> coord = new StandardLookupCoordinator<T>(
				key, myNode, config);
		final NodeLookupOperation<T> op = new de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.KClosestNodesLookupOperation<T>(
				coord, myNode, opCallback, why, config);
		return op;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final KademliaOperation<T> getRepublishOperation(
			final OperationCallback opCallback) {
		final KademliaOperation<T> repubOp = new RepublishOperation<T>(myNode,
				opCallback, config);
		return repubOp;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final KademliaOperation<T> getStoreOperation(final DHTObject data,
			final KademliaOverlayKey key, final boolean storeForeignData,
			final Reason why, final OperationCallback opCallback) {
		final KademliaOperation<T> storeOp = new StoreOperation<T>(key, data,
				storeForeignData, myNode, opCallback, why, config);
		return storeOp;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final KademliaOperation<T> getRefreshOperation(
			final boolean forceRefresh, final OperationCallback opCallback) {
		return new BucketRefreshOperation<T>(forceRefresh, myNode, opCallback,
				config);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final KademliaOperation<T> getBuildRoutingTableOperation(
			final OperationCallback opCallback) {
		return new RoutingTableBuildOperation<T>(myNode, opCallback, config);
	}
}
