package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations;

import java.util.Set;

import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.api.overlay.DHTObject;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.OperationFactory.NodeLookupOperation;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.DataMsg;


/**
 * An operation that stores a given data item in the overlay network.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class StoreOperation<T extends KademliaOverlayID> extends
AbstractKademliaOperation<T> {

	/**
	 * The key of the data item.
	 */
	private final KademliaOverlayKey dataKey;

	/**
	 * The data item to be published.
	 */
	private final DHTObject dataValue;

	/**
	 * Whether the data should be stored on the K closest nodes around key if
	 * this node is not part of them.
	 */
	private final boolean storeForeignData;

	/**
	 * The operation used to determine the k closest nodes to dataKey.
	 */
	private NodeLookupOperation<T> lookupOperation;

	/**
	 * Constructs an operation that looks up the {@KademliaConfig#K} closest
	 * nodes to <code>key</code> (over the network) and stores the data item
	 * <code>value</code> on them.
	 * 
	 * @param key
	 *                the KademliaOverlayKey of the data item to be published.
	 * @param value
	 *                the data item that is to be published in the overlay
	 *                network.
	 * @param storeForeignData
	 *                whether <code>value</code> should be stored on the K
	 *                closest neighbours of <code>key</code> if
	 *                <code>node</code> is <i>not</i> one of them. This is
	 *                determined having carried out the lookup for
	 *                <code>key</code>'s neighbours. (If set to
	 *                <code>false</code> and if <code>node</code> is not
	 *                part of the K closest nodes of <code>key</code>, this
	 *                StoreOperation will finish in state ERROR.)
	 * @param node
	 *                the Node that runs this operation.
	 * @param opCallback
	 *                a callback that is informed when this operation
	 *                terminates.
	 * @param reason
	 *                the reason why this operation is being run.
	 * @param conf
	 *                an OperationsConfig reference that permits to retrieve
	 *                configuration "constants".
	 */
	protected StoreOperation(final KademliaOverlayKey key,
			final DHTObject value, final boolean storeForeignData,
			final Node<T> node, final OperationCallback opCallback,
			final Reason reason, final OperationsConfig conf) {
		super(node, opCallback, reason, conf);
		dataKey = key;
		dataValue = value;
		this.storeForeignData = storeForeignData;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected final void execute() {
		scheduleOperationTimeout(config.getLookupOperationTimeout() + 1);
		lookupOperation = getComponent().getOperationFactory()
		.getKClosestNodesLookupOperation(dataKey, getReason(), this);
		lookupOperation.scheduleImmediately();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected final void calledOperationDidFail(final Operation op) {
		// check whether event was caused by the operation we are waiting for
		if (lookupOperation != null
				&& lookupOperation.getOperationID() == op.getOperationID()) {
			// the previous lookup failed - we will not store
			finishOperation(OperationState.ERROR);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected final void calledOperationDidSucceed(final Operation op) {
		// check whether event was caused by the operation we are waiting for
		if (lookupOperation == null
				|| lookupOperation.getOperationID() != op.getOperationID()) {
			return;
		}
		sendDataItem(lookupOperation.getNodes());
		finishOperation(OperationState.SUCCESS);
	}

	/**
	 * Delivers the data item to the given list of receivers. If the local node
	 * is one of the receivers, the data item will directly be put into its
	 * local index. Data messages are sent to distant nodes. If the local
	 * contact is not part of the receivers and
	 * <code>storeForeignData==false</code>, then this Operation is aborted
	 * with an error and without storing the data anywhere.
	 * 
	 * @param receivers
	 *                a Set of KademliaOverlayContacts that will receive data
	 *                messages with the data item. (A local contact will
	 *                directly receive the data item into its local index.)
	 */
	private void sendDataItem(final Set<KademliaOverlayContact<T>> receivers) {
		final T myID = getComponent().getTypedOverlayID();
		final KademliaOverlayContact<T> myContact = getComponent()
		.getLocalContact();
		DataMsg<T> message;

		if (!storeForeignData && !receivers.contains(myContact)) {
			// we may store data only if we are part of k closest nodes
			finishOperation(OperationState.ERROR);
			return;
		}

		for (final KademliaOverlayContact<T> receiver : receivers) {
			if (receiver.equals(myContact)) {
				/*
				 * This is required to update the last republish time in the
				 * index so that entries that the node is responsible for do not
				 * expire!
				 */
				getComponent().getLocalIndex().put(dataKey, dataValue);
			} else {
				message = new DataMsg<T>(myID, receiver.getOverlayID(),
						dataKey, dataValue, getReason(), config);
				send(message, receiver);
			}
		}
	}

}
