package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup;

import java.util.Set;

import de.tud.kom.p2psim.api.common.OperationCallback;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.OperationsConfig;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.OperationFactory.NodeLookupOperation;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.LookupCoordinator.NonhierarchicalLookupCoordinator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.LookupCoordinatorClient.NonhierarchicalLookupCoordinatorClient;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.setup.KademliaSetup;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.KClosestNodesLookupMsg;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages.KademliaMsg;


/**
 * An operation that looks up the {@link OperationsConfig#getBucketSize()}
 * closest nodes around a given key (the key is held by the lookup coordinator).
 * It has to be customised with a LookupCoordinator that adapts the lookup
 * process to a certain flavour of Kademlia (standard, Kandy, or hierarchical).
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class KClosestNodesLookupOperation<T extends KademliaOverlayID> extends
AbstractLookupOperation<T> implements
NonhierarchicalLookupCoordinatorClient<T>, NodeLookupOperation<T> {

	/**
	 * Constructs a new node lookup operation.
	 * 
	 * @param coordinator
	 *                the NonhierarchicalLookupCoordinator that implements the
	 *                details of the lookup process in a certains flavour of
	 *                Kademlia (for example standard Kademlia, Kandy, or
	 *                hierarchical Kademlia).
	 * @param node
	 *                the Node that initiates this lookup.
	 * @param opCallback
	 *                a callback that is informed when this operation
	 *                terminates.
	 * @param reason
	 *                why this operation is to be executed - either
	 *                user-initiated or for maintenance.
	 * @param conf
	 *                an OperationsConfig reference that permits to retrieve
	 *                configuration "constants".
	 */
	public KClosestNodesLookupOperation(
			final NonhierarchicalLookupCoordinator<T> coordinator,
			final Node<T> node, final OperationCallback opCallback,
			final Reason reason, final OperationsConfig conf) {
		super(coordinator, node, opCallback, reason, conf);
		coordinator.setClient(this);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final Set<KademliaOverlayContact<T>> getNodes() {
		return coordinator.getCurrentlyKnownKClosestNodes();
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final boolean isContinuationNecessary() {
		if (isFinished()) {
			return false; // operation has been stopped
		}
		return true; // continue lookup as long as possible
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void coordinatorFinished() {
		finishOperation(OperationState.SUCCESS); // node lookup "cannot" fail
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected final void operationJustFinished() {
//		log.debug("K closest nodes lookup complete (key="
//		+ coordinator.getKey() + "; |result|=" + getNodes().size()
//		+ "; node=" + getComponent() + "; how=" + getState()); // TODO uncomment
		if(KademliaSetup.getMonitor() != null)
			KademliaSetup.getMonitor().kClosestNodesLookupCompleted(
					coordinator.getKey(), getNodes(), this);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void sendLookupMessage(
			final KademliaOverlayContact<T> destination) {
		final KademliaMsg<T> message = new KClosestNodesLookupMsg<T>(
				getComponent().getTypedOverlayID(), destination.getOverlayID(),
				coordinator.getKey(), getReason(), config);
		sendAndWait(message, destination);
	}

}
