package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.HKademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;

/**
 * Interface with callback methods to be implemented by classes that use a
 * {@link LookupCoordinator}.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public abstract interface LookupCoordinatorClient<T extends KademliaOverlayID> {

    /**
     * Non-hierarchical version of LookupCoordinatorClient.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    public interface NonhierarchicalLookupCoordinatorClient<S extends KademliaOverlayID>
	    extends LookupCoordinatorClient<S> {

	/**
	 * Sends a lookup message with key {@link getKey()} to
	 * <code>destination</code>. The concrete type of the message sent
	 * depends on the type of the LookupCoordinatorClient (whether the
	 * method is invoked in the context of a node lookup or a data lookup).
	 * 
	 * @param destination
	 *                the KademliaOverlayContact to which the message will
	 *                be sent.
	 */
	public void sendLookupMessage(KademliaOverlayContact<S> destination);

    }

    /**
     * Hierarchical version of LookupCoordinatorClient.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    public interface HierarchicalLookupCoordinatorClient<H extends HKademliaOverlayID>
	    extends LookupCoordinatorClient<H> {

	/**
	 * Sends a lookup message with key {@link getKey()} to
	 * <code>destination</code>. The concrete type of the message sent
	 * depends on the type of the LookupCoordinatorClient (whether the
	 * method is invoked in the context of a node lookup or a data lookup).
	 * All results that may be sent as a response to this message need to
	 * have a common cluster depth of at least <code>minClusterDepth</code>
	 * with the sender of this message.
	 * 
	 * @param destination
	 *                the KademliaOverlayContact to which the message will
	 *                be sent.
	 * @param minClusterDepth
	 *                the minimal common cluster depth that all contacts
	 *                returned as a reply need to have in common with the
	 *                sender of the lookup message.
	 */
	public void sendLookupMessage(KademliaOverlayContact<H> destination,
		int minClusterDepth);

    }

    /**
     * @return the number of messages that are currently in transit (messages
     *         that have been sent, but for which no reply has been received and
     *         no timeout has occurred).
     */
    public int getTransitCount();

    /**
     * Determines whether the client needs the lookup to continue. For instance,
     * if the desired data item has been found, the lookup can stop. In
     * contrast, if the goal is to find the k closest neighbours of the key, the
     * lookup must continue until no more nodes are available to query.
     * 
     * In general, the lookup process can not be expected to halt as soon as
     * this method returns <code>true</code> as some work to clean up and
     * finish the lookup process might be necessary.
     * 
     * @return true iff the lookup has to continue from the lookup coordinator
     *         client's point of view.
     */
    public boolean isContinuationNecessary();

    /**
     * Called to notify the client that the coordinator has finished the lookup
     * process. It is up to the client to decide whether the lookup was
     * successful.
     */
    public void coordinatorFinished();
}
