package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup;

import java.util.Set;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTable;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.OperationsConfig;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.LookupCoordinator.NonhierarchicalLookupCoordinator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.lookup.LookupCoordinatorClient.NonhierarchicalLookupCoordinatorClient;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;


/**
 * Lookup coordinator for standard Kademlia lookups. The lookup coordinator
 * keeps a list with the K={@link OperationsConfig#getBucketSize()} currently
 * known closest nodes to {@link LookupCoordinatorClient#getKey()} and
 * concurrently queries the {@link OperationsConfig#getMaxConcurrentLookups()}
 * closest ones of them that have not yet been queried. As lookup replies are
 * received, the new contacts are inserted into the list and more distant
 * contacts are removed so that at most <code>K</code> contacts are saved.
 * Further lookup messages are sent with the constraint that no more than
 * <code>ALPHA</code> messages are in transit at the same time.
 * <p>
 * The lookup stops if
 * <ul>
 * <li>(a) the lookup coordinator client returns
 * <code>{@link LookupCoordinatorClient#isContinuationNecessary()}==false</code>
 * or
 * <li>(b) all currently known <code>K</code> closest nodes to the
 * <code>key</code> have been queried and all replies (or timeouts) have been
 * received (that is, there is no possibility to continue the lookup).
 * </ul>
 * 
 * This lookup coordinator does not consider clusters (it works as in standard
 * Kademlia).
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class StandardLookupCoordinator<T extends KademliaOverlayID> extends
	AbstractLookupCoordinator<T> implements
	NonhierarchicalLookupCoordinator<T> {

    /**
     * The client that uses this LookupCoordinator (used for callbacks).
     */
    private NonhierarchicalLookupCoordinatorClient<T> client;

    /**
     * The Node that has initiated this lookup.
     */
    protected final Node<T> node;

    /**
     * Constructs a new standard lookup coordinator.
     * 
     * @param lookupKey
     *                the KademliaOverlayKey that is to be looked up.
     * @param node
     *                the Node that initiates this lookup.
     * @param conf
     *                an OperationsConfig reference that permits to retrieve
     *                configuration "constants".
     */
    public StandardLookupCoordinator(final KademliaOverlayKey lookupKey,
	    final Node<T> node, final OperationsConfig conf) {
	super(lookupKey, conf);
	this.node = node;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final void setClient(
	    final NonhierarchicalLookupCoordinatorClient<T> client) {
	this.client = client;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final NonhierarchicalLookupCoordinatorClient<T> getClient() {
	return client;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final RoutingTable<T> getRoutingTable() {
	return node.getKademliaRoutingTable();
    }

    /*
     * Internal methods (lookup coordination)
     */

    /**
     * Initialises the lookup: get the k closest nodes from the routing table,
     * insert them into <code>kClosestNodes</code>, and mark all its entries
     * as TO_QUERY.
     */
    @Override
    protected final void init() {
	final Set<KademliaOverlayContact<T>> contactsFromRT;

	// find new contacts from routing table
	contactsFromRT = localLookup(getKey());

	// insert all contacts as TO_QUERY
	kClosestNodes.putAll(contactsFromRT, ContactState.TO_QUERY, true);
	/*
	 * Insert own contact as QUERIED. This makes sure that the own contact
	 * is never queried via the network. (If the own contact is too distant
	 * from the target to be of interest, it will be dropped.)
	 */
	kClosestNodes.put(node.getLocalContact(), ContactState.QUERIED);
    }

    /**
     * Looks up {@link OperationsConfig#getBucketSize()} contacts from the local
     * routing table.
     * 
     * @param key
     *                the KademliaOverlayKey to be looked up.
     * @return a Set containing K contacts.
     */
    protected Set<KademliaOverlayContact<T>> localLookup(
	    final KademliaOverlayKey key) {
	return node.getKademliaRoutingTable().localLookup(key, config.getBucketSize());
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final void proceed(final int numOfMessages) {
	KademliaOverlayContact<T> bestUnqueried;

	// send up to numOfMessages messages
	for (int i = 1; i <= numOfMessages; i++) {
	    /*
	     * Select "best" contact that has not already been queried, and if
	     * such a contact exists, send a message. Else stop.
	     */
	    bestUnqueried = getBestUnqueried();
	    if (bestUnqueried != null) {
		kClosestNodes.put(bestUnqueried, ContactState.QUERIED);
		client.sendLookupMessage(bestUnqueried);
	    } else {
		return;
	    }
	}
    }

}
