package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.setup;

import org.apache.log4j.Logger;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.measurement.KademliaMonitor;
import de.tud.kom.p2psim.impl.util.logging.SimLogger;

/**
 * Holds static references to global (singleton) components in a Kademlia
 * experiment.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class KademliaSetup {

	private final static Logger log = SimLogger.getLogger(KademliaSetup.class);

	/** Singleton instance of KademliaSetup. */
	private static KademliaSetup singleton;

	/** The currently used node factory. */
	private AbstractNodeFactory nodeFactory;

	/** The currently used workload generator. */
	private WorkloadGenerator workloadGen;

	/**
	 * The Monitor that is being used to monitor Kademlia specific events.
	 */
	private KademliaMonitor monitor;

	/**
	 * The Config that contains constants for use in Kademlia.
	 */
	private Config config;

	private KademliaSetup() {
		// should not be called externally
	}

	/**
	 * @return the singleton instance of KademliaSetup.
	 */
	public static KademliaSetup getInstance() {
		if (singleton == null) {
			singleton = new KademliaSetup();
		}
		return singleton;
	}

	/**
	 * Sets the currently used node factory.
	 * 
	 * @param factory
	 *                the AbstractNodeFactory that is currently being used.
	 */
	public void setNodeFactory(final AbstractNodeFactory factory) {
		nodeFactory = factory;
	}

	/**
	 * @return the currently used AbstractNodeFactory.
	 */
	public static AbstractNodeFactory getNodeFactory() {
		return getInstance().nodeFactory;
	}

	/**
	 * Sets the currently used workload generator.
	 * 
	 * @param generator
	 *                the WorkloadGenerator that is currently being used.
	 */
	public void setWorkload(final WorkloadGenerator generator) {
		workloadGen = generator;
	}

	/**
	 * @return the currently used WorkloadGenerator.
	 */
	public static WorkloadGenerator getWorkloadGenerator() {
		return getInstance().workloadGen;
	}

	/**
	 * Sets the KademliaMonitor to be used in this setup.
	 */
	public void setMonitor(final KademliaMonitor mon) {
		this.monitor = mon;
	}

	/**
	 * @return the KademliaMonitor to be used. The result of this method is
	 *         never null.
	 */
	public static KademliaMonitor getMonitor() {
		return getInstance().monitor;
	}

	/**
	 * Sets the global Config object that contains Kademlia-wide constants. This
	 * method has an effect only if a Config object has never been set before.
	 * 
	 * @param confPath
	 *                a String with the path to the Kademlia properties file
	 *                (absolute or relative to the programme directory).
	 */
	public void setConfigPath(final String confPath) {
		if (config == null) {
			config = new FileConfig(confPath);
			log.info("Using Kademlia configuration from '" + confPath + "'.");
		}
	}

	/**
	 * Sets the Config instance used in this simulation to <code>conf</code>.
	 * <b>This method is for test purposes only!!</b>
	 * 
	 * @param conf
	 *                the Config to be used.
	 */
	public static void setConfig(final Config conf) {
		getInstance().config = conf;
	}

	/**
	 * @return the global Config object that contains Kademlia-wide constants.
	 *         If none has been configured before, it will be attempted to read
	 *         configuration values from the default location or use the default
	 *         values.
	 *         <p>
	 *         For efficiency reasons, other components should copy the result
	 *         of this method into a final instance variable instead of
	 *         accessing it here each time needed. The result of this method can
	 *         be assumed to be constant over time.
	 */
	public static Config getConfig() {
		final KademliaSetup ins = getInstance();
		if (ins.config == null) {
			ins.config = new FileConfig("config/kademlia.properties");
		}
		return ins.config;
	}
}