package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.messages;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.operations.AbstractKademliaOperation.Reason;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.HKademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayKey;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.TypesConfig;

/**
 * Format for a message that requests the k closest known nodes around a given
 * key and restricts the contacts that may be returned to have a given minimum
 * common cluster depth with the sender of this message.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public final class HKClosestNodesLookupMsg<H extends HKademliaOverlayID> extends
	KClosestNodesLookupMsg<H> {

    /**
     * The minimum depth of the deepest common cluster between each returned
     * contact and the sender's ID.
     */
    private final int minClusterDepth;

    /**
     * Constructs a new message used to look up the k closest known neighbours
     * of <code>nodeKey</code> that have at least the cluster depth
     * <code>minClusterDepth</code> with respect to the sender's ID.
     * 
     * @param sender
     *                the KademliaOverlayID of the sender of this message.
     * @param destination
     *                the KademliaOverlayID of the destination of this message.
     * @param dataKey
     *                the KademliaOverlayKey of the node of which the k closest
     *                known neighbours are to be returned.
     * @param minClusterDepth
     *                the required minimum depth of the deepest common cluster
     *                between each returned contact and the sender's ID.
     * @param why
     *                the reason why this message will be sent.
     * @param conf
     *                a TypesConfig reference that permits to retrieve
     *                configuration "constants".
     */
    public HKClosestNodesLookupMsg(final H sender, final H destination,
	    final KademliaOverlayKey nodeKey, final int minClusterDepth,
	    final Reason why, final TypesConfig conf) {
	super(sender, destination, nodeKey, why, conf);
	this.minClusterDepth = minClusterDepth;
    }

    /**
     * @return the minimum required depth of the deepest common cluster between
     *         each returned contact and the ID of the sender of this message.
     */
    public final int getMinClusterDepth() {
	return minClusterDepth;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected final long getOtherFieldSize() {
	return super.getOtherFieldSize() + 4;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public final String toString() {
	return "[HKClosestNodesLookupMsg|from:" + getSender() + "; to:"
		+ getDestination() + "; reason:" + getReason() + "; lookup:"
		+ getNodeKey() + "; minClusterDepth=" + minClusterDepth + "]";
    }

}
