package de.tud.kom.p2psim.impl.util.helpers;

import java.math.BigInteger;
import java.util.Comparator;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTableComparators;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTableComparators.XORMaxComparator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;


/**
 * Defines often used comparators.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 * @author Sebastian Kaune
 */
public class Comparators {

    /**
     * Comparator that orders KademliaOverlayContacts according to the XOR value
     * of their KademliaOverlayID with a preset reference value. IDs with a
     * higher XOR are <i>"bigger"</i> than IDs with a lower XOR.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    public static class KademliaOverlayContactXORMaxComparator<T extends KademliaOverlayID>
	    extends XORMaxComparator<T> implements
	    Comparator<KademliaOverlayContact<T>> {

	/**
	 * Constructs a new KademliaOverlayContactXORMaxComparator that compares
	 * KademliaOverlayContacts by the XOR value of their KademliaOverlayID
	 * with <code>reference</code>.
	 * 
	 * @param ref
	 *                the BigInteger used to calculate the XOR which the
	 *                KademliaOverlayContacts to be ordered are compared
	 *                against.
	 */
	public KademliaOverlayContactXORMaxComparator(final BigInteger ref) {
	    super(ref);
	}

	// implementation of compare inherited from XORMaxComparator
    }

    /**
     * Comparator that orders Nodes according to the XOR value of their
     * KademliaOverlayID with a preset reference value. IDs with a higher XOR
     * are <i>"bigger"</i> than IDs with a lower XOR.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    public static class NodeXORMaxComparator<T extends KademliaOverlayID>
	    extends XORMaxComparator<T> implements Comparator<Node<T>> {

	/**
	 * Constructs a new KademliaOverlayContactXORMaxComparator that compares
	 * KademliaOverlayContacts by the XOR value of their KademliaOverlayID
	 * with <code>reference</code>.
	 * 
	 * @param ref
	 *                the BigInteger used to calculate the XOR which the
	 *                KademliaOverlayContacts to be ordered are compared
	 *                against.
	 */
	public NodeXORMaxComparator(final BigInteger ref) {
	    super(ref);
	}

	/**
	 * Compares the BigInteger values of the KademliaOverlayIDs contained in
	 * the KademliaOverlayContacts of two Nodes (from the
	 * kademlia2.components package!) according to their XOR value with the
	 * preset reference value. A BigInteger is considered greater if it has
	 * a higher XOR value with the reference. The remaining contract of this
	 * method conforms with {@link Comparator}.
	 */
	@Override
	public final int compare(final Node<T> o1, final Node<T> o2) {
	    return compare(o1.getLocalContact(), o2.getLocalContact());
	}
    }

    /**
     * Comparator that orders KademliaOverlayIDs according to the XOR value of
     * their BigInteger representation with a preset reference value. IDs with a
     * higher XOR are <i>"bigger"</i> than IDs with a lower XOR.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    public static class KademliaOverlayIDXORMaxComparator<T extends KademliaOverlayID>
            extends RoutingTableComparators.XORMaxComparator<T> implements Comparator<T> {
    
        /**
         * Constructs a new KademliaOverlayIDXORMaxComparator that compares
         * KademliaOverlayIDs via the XOR distance between their BigInteger IDs
         * and <code>ref</code>.
         * 
         * @param reference
         *                the BigInteger used to calculate the XOR which the
         *                BigInteger representation of the KademliaOverlayIDs to
         *                be ordered are compared against.
         */
        public KademliaOverlayIDXORMaxComparator(final BigInteger ref) {
            super(ref);
        }
    
        // implementation of compare inherited from XORMaxComparator
    }

}
