package de.tud.kom.p2psim.impl.util.kbrpostprocessing;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.zip.GZIPInputStream;

import de.tud.kom.p2psim.impl.simengine.Simulator;
import de.tud.kom.p2psim.impl.util.kbrpostprocessing.KBROverlayLogParser.SinglePeerStatistic.QueryStatistic;

/**
 * Parser for log files generated by de.tud.kom.p2psim.impl.analyzer.KbrOverlayAnalyzer
 * 
 * @author Julius Rueckert
 *
 */
public class KBROverlayLogParser {
	
	public static class SinglePeerStatistic {

		public class QueryStatistic{
			
			private long duration;
			private int hops;
			
			public QueryStatistic(long duration, int hops){
				this.duration = duration;
				this.hops = hops;
			}

			public long getDuration() {
				return duration;
			}

			public int getHops() {
				return hops;
			}
		}
		
		private String peerID;
		
		/*
		 * Containers for overlay message statistics
		 */
		private Map<String, Integer> overlayMsgsSent = new HashMap<String, Integer>();
		private Map<String, Integer> overlayMsgsReceived = new HashMap<String, Integer>();

		/*
		 * Containers for application message statistics
		 */
		private Map<String, Integer> applicationMsgsInited = new HashMap<String, Integer>();
		private Map<String, Integer> applicationMsgsForwarded = new HashMap<String, Integer>();
		private Map<String, Integer> applicationMsgsDelivered = new HashMap<String, Integer>();
		
		/*
		 * Containers for query statistics
		 */
		private Map<String, List<QueryStatistic>> queryStats = new HashMap<String, List<QueryStatistic>>();
		private Map<String, Integer> queriesIntitated = new HashMap<String, Integer>();

		
		public SinglePeerStatistic(String peerID){
			this.peerID = peerID;
		}

		public void overlayMsgSent(String messageType){

			if(overlayMsgsSent.containsKey(messageType)){
				overlayMsgsSent.put(messageType, overlayMsgsSent.get(messageType)+1);
			} else {
				overlayMsgsSent.put(messageType, 1);
			}
		}

		public void overlayMsgReceived(String messageType){
			
			if(overlayMsgsReceived.containsKey(messageType)){
				overlayMsgsReceived.put(messageType, overlayMsgsReceived.get(messageType)+1);
			} else {
				overlayMsgsReceived.put(messageType, 1);
			}
		}
		
		public void applicationMsgInited(String messageType){
			if(applicationMsgsInited.containsKey(messageType)){
				applicationMsgsInited.put(messageType, applicationMsgsInited.get(messageType)+1);
			} else {
				applicationMsgsInited.put(messageType, 1);
			}
		}
		
		public int applicationMsgsInitedCount(){
			return applicationMsgsInited.size();
		}
		
		public void applicationMsgForwarded(String messageType){
			if(applicationMsgsForwarded.containsKey(messageType)){
				applicationMsgsForwarded.put(messageType, applicationMsgsForwarded.get(messageType)+1);
			} else {
				applicationMsgsForwarded.put(messageType, 1);
			}
		}
		
		public int applicationMsgsForwardedCount(){
			return applicationMsgsForwarded.size();
		}
		
		public void applicationMsgDelivered(String messageType){
			if(applicationMsgsDelivered.containsKey(messageType)){
				applicationMsgsDelivered.put(messageType, applicationMsgsDelivered.get(messageType)+1);
			} else {
				applicationMsgsDelivered.put(messageType, 1);
			}
		}
		
		public int applicationMsgsDeliveredCount(){
			return applicationMsgsDelivered.size();
		}
		
		public void queryInited(String initiator, String applicationMsgType){
			if(queriesIntitated.containsKey(applicationMsgType)){
				queriesIntitated.put(applicationMsgType, queriesIntitated.get(applicationMsgType)+1);
			} else {
				queriesIntitated.put(applicationMsgType, 1);
			}
		}
		
		public void querySucceeded(String applicationMsgType, long duration, int hops){
			if(!queryStats.containsKey(applicationMsgType)){
				LinkedList<QueryStatistic> list = new LinkedList<QueryStatistic>();
				list.add(new QueryStatistic(duration, hops));
				queryStats.put(applicationMsgType, list);
			} else {
				List<QueryStatistic> list = queryStats.get(applicationMsgType);
				list.add(new QueryStatistic(duration, hops));
			}
		}
		
		public Map<String, Integer> getOverlayMsgsReceived(){
			return overlayMsgsReceived;
		}
		
		public Map<String, Integer> getOverlayMsgsSent(){
			return overlayMsgsSent;
		}
		
		public Map<String, Integer> getApplicationMsgsInited(){
			return applicationMsgsInited;
		}

		public Map<String, Integer> getApplicationMsgsForwarded(){
			return applicationMsgsForwarded;
		}

		public Map<String, Integer> getApplicationMsgsDelivered(){
			return applicationMsgsDelivered;
		}
		
		public Map<String, Integer> getQueriesInitiated(){
			return queriesIntitated;
		}

		public Map<String, List<QueryStatistic>> getQueriesSucceeded(){
			return queryStats;
		}
		
		public long getAverageQueryDuration(){
			long sum = 0;
			int count = 0;
			for(Entry<String, List<QueryStatistic>> statsEntry : queryStats.entrySet()){
				for (QueryStatistic listElem : statsEntry.getValue()) {
					sum += listElem.getDuration();
					count++;
				}
			}
			if(count>0)
				return sum/count;
			else
				return 0;
		}
		
		public long getAverageQueryHops(){
			int sum = 0;
			int count = 0;
			for(Entry<String, List<QueryStatistic>> statsEntry : queryStats.entrySet()){
				for (QueryStatistic listElem : statsEntry.getValue()) {
					sum += listElem.getHops();
					count++;
				}
			}
			if(count>0)
				return sum/count;
			else
				return 0;
		}
		
	}

	/**
	 * This is a data container to ease the handling of collected data
	 * 
	 * @param <T> the type of the value
	 */
	private static class KeyValuePair<T extends Comparable<T>> implements Comparable<KeyValuePair<T>>{
		private String key;
		private T value;
		
		public KeyValuePair(String k, T v){
			key = k;
			value = v;
		}
		
		public String getKey() {
			return key;
		}
		public void setKey(String key) {
			this.key = key;
		}
		public T getValue() {
			return value;
		}
		public void setValue(T value) {
			this.value = value;
		}

		@Override
		public int compareTo(KeyValuePair<T> secondPair) {
			return secondPair.getValue().compareTo(getValue()); 

		}
	}
	
	/**
	 * This is a data container to ease the handling of collected data
	 *
	 * @param <T> the type of the two stored values
	 */
	private static class KeyValueValueTriple<T extends Comparable<T>> implements Comparable<KeyValueValueTriple<T>>{
		private String key;
		private T value1;
		private T value2;
		
		public KeyValueValueTriple(String k, T v1, T v2){
			key = k;
			value1 = v1;
			value2 = v2;
		}
		
		public String getKey() {
			return key;
		}
		public void setKey(String key) {
			this.key = key;
		}
		public T getValue1() {
			return value1;
		}
		public void setValue1(T value) {
			this.value1 = value;
		}
		public T getValue2() {
			return value2;
		}
		public void setValue2(T value) {
			this.value2 = value;
		}

		@Override
		public int compareTo(KeyValueValueTriple<T> secondPair) {
			return secondPair.getValue1().compareTo(getValue1()); 
		}
	}
	
	private static class KeyValueValueValueQuadrupel<T extends Comparable<T>> implements Comparable<KeyValueValueValueQuadrupel<T>>{
		private String key;
		private T value1;
		private T value2;
		private T value3;
		
		public KeyValueValueValueQuadrupel(String k, T v1, T v2, T v3){
			key = k;
			value1 = v1;
			value2 = v2;
			value3 = v3;
		}
		
		public String getKey() {
			return key;
		}
		public void setKey(String key) {
			this.key = key;
		}
		public T getValue1() {
			return value1;
		}
		public void setValue1(T value) {
			this.value1 = value;
		}
		public T getValue2() {
			return value2;
		}
		public void setValue2(T value) {
			this.value2 = value;
		}
		public T getValue3() {
			return value3;
		}
		public void setValue3(T value) {
			this.value3 = value;
		}
		@Override
		public int compareTo(KeyValueValueValueQuadrupel<T> secondPair) {
			return secondPair.getValue1().compareTo(getValue1()); 
		}
	}

	/**
	 * The relative path to write all gnuplot output files
	 */
	private final static String gnuplotOutputPath = "kbrAnalyzerOutput/gnuplot/";
	
	/**
	 * The TrueType font used for plots 
	 */
	private final static String plotFont = "./LiberationSans-Regular.ttf,11";
	
	/**
	 * The path to the log to parse
	 */
	private static String logPath;

	/**
	 * Reader for log file
	 */
	private static BufferedReader inputReader;

	/**
	 * Determines if the log file is compressed
	 */
	private static boolean inputIsCompressed = false;
	
	
	/**
	 * Containers for statistics 
	 */
	private static Map<String, Integer> overallTransmittedOverlayMsgs = new HashMap<String, Integer>();
	private static int overallTransmittedOverlayMsgCount = 0;
	private static Set<String> overlayMessageTypes = new HashSet<String>();

	private static Map<String, Integer> overallInitedApplicationMsgsByType = new HashMap<String, Integer>();
	private static Map<String, Integer> overallForwardedApplicationMsgsByType = new HashMap<String, Integer>();
	private static Map<String, Integer> overallDeliveredApplicationMsgsByType = new HashMap<String, Integer>();

	private static Map<String, Integer> overallInitedQueries = new HashMap<String, Integer>();
	private static Map<String, Integer> overallSucceededQueries = new HashMap<String, Integer>();
	private static int overallFailedQueryCount = 0;
	private static int overallSucceededQueryCount = 0;
	private static Map<String, Integer> failedQueryInitiator = new HashMap<String, Integer>();
	private static Set<String> applicationMessageTypes = new HashSet<String>();
	private static Set<String> queriedApplicationMessageTypes = new HashSet<String>();
	private static List<KeyValuePair<Integer>> hopsAtSucceededQuery = new LinkedList<KeyValuePair<Integer>>();
	private static List<KeyValuePair<Double>> durationAtSucceededQuery = new LinkedList<KeyValuePair<Double>>();

	private static double overallSentNetMsgs = 0;
	private static double overallLostNetMsgs = 0;
	
	private static Set<String> peerIDs = new HashSet<String>();
	private static Map<String, SinglePeerStatistic> peerStatistics = new HashMap<String, SinglePeerStatistic>();

	public static void main(String[] args) {

		/*
		 * Give info about usage on console
		 */
		if (args.length < 1) {
			System.out.println("Missing parameters: log_path [use_compression (true/false)]");
			return;
		} else {
			/*
			 * Setup properties given through parameters
			 */
			logPath = args[0];

			if (args.length > 1)
				inputIsCompressed = Boolean.parseBoolean(args[1]);

			File logFile = new File(logPath);

			try {
				if (inputIsCompressed) {
					// Set up compressed input stream
					inputReader = new BufferedReader(
							new InputStreamReader(
									new GZIPInputStream(
											new FileInputStream(logFile))));
				} else {
					// Set up non compressed input stream
					inputReader = new BufferedReader(
							new InputStreamReader(
									new FileInputStream(logFile)));
				}

				/**
				 * Process log file
				 */
				String line;
				int processedLines = 0;
				long startTime = System.currentTimeMillis();
				
				while((line = inputReader.readLine()) != null){
					processLine(line);
					processedLines++;
				}
				
				long duration = System.currentTimeMillis() - startTime;
				
				System.out.println("The Parser processed "+processedLines+" lines in "+(duration/1000)+"s. "+
						           "That means an average speed of "+((new BigDecimal(processedLines)).divide(new BigDecimal((duration/1000)), RoundingMode.DOWN))+" lines per second.");

				/**
				 * Generate gnuplot files for statistics
				 */
				generateGnuPlotFiles();

			} catch (FileNotFoundException e) {
				System.err.println("The named logfile does not exist! ("+logPath+")");
				return;
			} catch (IOException e) {
				System.err.println("Could not read the named logfile! ("+logPath+")");
				return;
			}
		}
	}

	
	/**
	 * Processes a single line
	 * 
	 * @param line the line to process
	 */
	private static void processLine(String line){
		// Split line into parts
		String[] parts = line.split(" ");

		if(line.startsWith("ol_m_transmitted")){
			String sender 		= parts[1];
			String receiver 	= parts[2];
			String messageType 	= parts[3];
			String size 		= parts[4]; //TODO: use it or loose it?
			String duration		= parts[5]; //TODO: use it or loose it?

			// Add message type to set of types
			overlayMessageTypes.add(messageType);
			
			assureStatisticMapIsReady(sender);
			assureStatisticMapIsReady(receiver);

			/*
			 *  Overall overlay statistics
			 */
			if(overallTransmittedOverlayMsgs.containsKey(messageType)){
				overallTransmittedOverlayMsgs.put(messageType, overallTransmittedOverlayMsgs.get(messageType)+1);
			} else {
				overallTransmittedOverlayMsgs.put(messageType, 1);
			}

			overallTransmittedOverlayMsgCount++;
			
			/* 
			 * Per peer statistics
			 */
			peerStatistics.get(sender).overlayMsgSent(messageType);
			peerStatistics.get(receiver).overlayMsgReceived(messageType);

		} else if(line.startsWith("app_m_init")){
			String sender 		= parts[1];
			String receiver 	= parts[2]; //TODO: use it or loose it?
			String messageType 	= parts[3];
			String size 		= parts[4]; //TODO: use it or loose it?

			assureStatisticMapIsReady(sender);
			
			applicationMessageTypes.add(messageType);

			if(overallInitedApplicationMsgsByType.containsKey(messageType)){
				overallInitedApplicationMsgsByType.put(messageType, overallInitedApplicationMsgsByType.get(messageType)+1);
			} else {
				overallInitedApplicationMsgsByType.put(messageType, 1);
			}

			peerStatistics.get(sender).applicationMsgInited(messageType);


		} else if(line.startsWith("app_m_forward")){
			String sender 		= parts[1];
			String receiver 	= parts[2]; //TODO: use it or loose it?
			String messageType 	= parts[3];
			String size 		= parts[4]; //TODO: use it or loose it?
			
			assureStatisticMapIsReady(sender);
			
			applicationMessageTypes.add(messageType);

			if(overallForwardedApplicationMsgsByType.containsKey(messageType)){
				overallForwardedApplicationMsgsByType.put(messageType, overallForwardedApplicationMsgsByType.get(messageType)+1);
			} else {
				overallForwardedApplicationMsgsByType.put(messageType, 1);
			}

			peerStatistics.get(sender).applicationMsgForwarded(messageType);

		} else if(line.startsWith("app_m_deliver")){
			String receiver 	= parts[1];
			String messageType 	= parts[2];
			String size 		= parts[3]; //TODO: use it or loose it?
			
			assureStatisticMapIsReady(receiver);
			
			applicationMessageTypes.add(messageType);

			if(overallDeliveredApplicationMsgsByType.containsKey(messageType)){
				overallDeliveredApplicationMsgsByType.put(messageType, overallDeliveredApplicationMsgsByType.get(messageType)+1);
			} else {
				overallDeliveredApplicationMsgsByType.put(messageType, 1);
			}

			peerStatistics.get(receiver).applicationMsgDelivered(messageType);

		} else if(line.startsWith("q_init")){
			String initiator 			= parts[1];
			String applicationMsgType 	= parts[2];
			
			queriedApplicationMessageTypes.add(applicationMsgType);			
			assureStatisticMapIsReady(initiator);
			
			if(overallInitedQueries.containsKey(applicationMsgType)){
				overallInitedQueries.put(applicationMsgType, overallInitedQueries.get(applicationMsgType)+1);
			} else {
				overallInitedQueries.put(applicationMsgType, 1);
			}
			
			peerStatistics.get(initiator).queryInited(initiator, applicationMsgType);

		} else if(line.startsWith("q_succeeded")){
			String receiver 				= parts[1];
			String applicationMsgType		= parts[2];
			String duration 				= parts[3];
			String hops 					= parts[4]; 
			
			assureStatisticMapIsReady(receiver);
			
			hopsAtSucceededQuery.add(new KeyValuePair<Integer>(applicationMsgType, Integer.parseInt(hops)));
			durationAtSucceededQuery.add(new KeyValuePair<Double>(applicationMsgType, (new Double(Long.parseLong(duration))/Simulator.MILLISECOND_UNIT)));
			
			if(overallSucceededQueries.containsKey(applicationMsgType)){
				overallSucceededQueries.put(applicationMsgType, overallSucceededQueries.get(applicationMsgType)+1);
			} else {
				overallSucceededQueries.put(applicationMsgType, 1);
			}
			
			peerStatistics.get(receiver).querySucceeded(applicationMsgType, Long.parseLong(duration), Integer.parseInt(hops));
			
			overallSucceededQueryCount++;
			
		} else if(line.startsWith("q_failed")){
			String failedHop = parts[1]; //TODO: use it or lose it
			String initiator = parts[2]; 
			
			if(failedQueryInitiator.containsKey(initiator)){
				failedQueryInitiator.put(initiator, failedQueryInitiator.get(initiator)+1);
			} else {
				failedQueryInitiator.put(initiator, 1);
			}
			
			overallFailedQueryCount++;
			
		} else if(line.startsWith("m_lost")){
			String reason 				= parts[1]; //TODO: use it or lose it
			String failedReceiver	 	= parts[2]; //TODO: use it or lose it
			String overlayMessageType 	= parts[3]; //TODO: use it or lose it

			overallLostNetMsgs++;
			
		} else if(line.startsWith("m_stat")){
			String netMsgCount 			= parts[1]; //TODO: use it or lose it
			
			overallSentNetMsgs = Double.parseDouble(netMsgCount);

		} else {
			System.err.println("Could not parse line: "+line);
		}
	}

	/**
	 * Insert new statistic element for peerID in peerStatistics if there is non
	 * 
	 * @param peerID
	 */
	private static void assureStatisticMapIsReady(String peerID){
		if(!peerStatistics.containsKey(peerID)){
			peerStatistics.put(peerID, new SinglePeerStatistic(peerID));
		}
		peerIDs.add(peerID);
	}
	
	private static void generateGnuPlotFiles(){
		ensureFoldersExist(gnuplotOutputPath, false);
		
		generatePerMsgTypeAppMsgGnuplotFiles();
		generatePerMsgTypeOlMsgGnuplotFiles();
		generatePerMsgTypeQueryGnuplotFiles();
		generateQueryOverallSuccessRateFiles();
		generateStaleMsgRatioFiles();
	}
	
	private static List<KeyValuePair<Double>> processHopsStatistics(){
		
		HashMap<String, Integer> hopsPerAppMsgType = new HashMap<String, Integer>();
		HashMap<String, Double> countPerAppMsgType = new HashMap<String, Double>();
		
		for (KeyValuePair<Integer> element : hopsAtSucceededQuery) {
			if(!hopsPerAppMsgType.containsKey(element.getKey())){
				hopsPerAppMsgType.put(element.getKey(), element.getValue());
				countPerAppMsgType.put(element.getKey(), new Double(1));
			} else {
				hopsPerAppMsgType.put(element.getKey(), 
							          hopsPerAppMsgType.get(element.getKey()) + element.getValue()); 
				countPerAppMsgType.put(element.getKey(), 
									   countPerAppMsgType.get(element.getKey()) + 1);
			}
		}
		
		List<KeyValuePair<Double>> listOfHopsPerAppMsgType = new LinkedList<KeyValuePair<Double>>();
		
		for (Entry<String, Integer> entry : hopsPerAppMsgType.entrySet()) {
			listOfHopsPerAppMsgType.add(new KeyValuePair<Double>(entry.getKey(),
																  (entry.getValue()/countPerAppMsgType.get(entry.getKey()))));
		}
		
		Collections.sort(listOfHopsPerAppMsgType);
		
		return listOfHopsPerAppMsgType;
	}
	
	private static List<KeyValuePair<Double>> processDurationStatistics(){
		
		HashMap<String, Double> durationPerAppMsgType = new HashMap<String, Double>();
		HashMap<String, Double> countPerAppMsgType = new HashMap<String, Double>();
		
		for (KeyValuePair<Double> element : durationAtSucceededQuery) {
			if(!durationPerAppMsgType.containsKey(element.getKey())){
				durationPerAppMsgType.put(element.getKey(), element.getValue());
				countPerAppMsgType.put(element.getKey(), new Double(1));
			} else {
				durationPerAppMsgType.put(element.getKey(), 
							          durationPerAppMsgType.get(element.getKey()) + element.getValue()); 
				countPerAppMsgType.put(element.getKey(), 
									   countPerAppMsgType.get(element.getKey()) + 1);
			}
		}
		
		List<KeyValuePair<Double>> listOfHopsPerAppMsgType = new LinkedList<KeyValuePair<Double>>();
		
		for (Entry<String, Double> entry : durationPerAppMsgType.entrySet()) {
			listOfHopsPerAppMsgType.add(new KeyValuePair<Double>(entry.getKey(),
																  (entry.getValue()/countPerAppMsgType.get(entry.getKey()))));
		}
		
		Collections.sort(listOfHopsPerAppMsgType);
		
		return listOfHopsPerAppMsgType;
	}
	
	private static void generatePerMsgTypeAppMsgGnuplotFiles(){
		FileWriter fileWriter;

		try {

			/**
			 * Statistics for hop count
			 */
			List<KeyValuePair<Double>> hopsPerAppMsgType = processHopsStatistics();
			String dataFileName = "app_hopsAtSuccessfullMessageDelivery.dat";
			writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName, hopsPerAppMsgType);

			/**
			 * Extend the plt-file for a later use by gnuplot 
			 */
			fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), false);
			String pltString = "reset\n"+
								"set title 'Hop count at application message delivery'\n"+
								"set xlabel 'Message type'\n"+
								"set xrange [-0.5:"+(hopsPerAppMsgType.size()-0.5)+"]\n"+
								"set ylabel 'Average number of hops'\n"+
								"set style data boxes\n"+
								"set style fill solid 1.0\n"+
								"set boxwidth .2\n"+
								"set yrange [0:*]\n"+
								"set output '"+"graphics/"+"app_hopsAtSuccessfullMessageDelivery.png'\n"+
								"set terminal png giant font '"+plotFont+"'\n"+
								"set xtics nomirror rotate by -45 \n"+
								"plot '"+dataFileName+"' using 2:xtic(1) title 'Average hop count' \n"+
								"set output '"+"graphics/"+"app_hopsAtSuccessfullMessageDelivery.eps'\n"+
								"set terminal postscript eps font '"+plotFont+"'\n"+
								"replot\n\n";

			fileWriter.write(pltString);
			fileWriter.close();
			
			/**
			 * Statistics for duration
			 */
			List<KeyValuePair<Double>> durationPerAppMsgType = processDurationStatistics();
			dataFileName = "app_durationAtSuccessfullMessageDelivery.dat";
			writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName, durationPerAppMsgType);

			/**
			 * Extend the plt-file for a later use by gnuplot 
			 */
			fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), true);
			pltString = "reset\n"+
						"set title 'Duration at application message delivery'\n"+
						"set xlabel 'Message type'\n"+
						"set xrange [-0.5:"+(hopsPerAppMsgType.size()-0.5)+"]\n"+
						"set ylabel 'Average duration [ms]'\n"+
						"set style data boxes\n"+
						"set style fill solid 1.0\n"+
						"set boxwidth .2\n"+
						"set yrange [0:*]\n"+
						"set output '"+"graphics/"+"app_durationAtSuccessfullMessageDelivery.png'\n"+
						"set terminal png giant font '"+plotFont+"'\n"+
						"set xtics nomirror rotate by -45 \n"+
						"plot '"+dataFileName+"' using 2:xtic(1) title 'Average duration' \n"+
						"set output '"+"graphics/"+"app_durationAtSuccessfullMessageDelivery.eps'\n"+
						"set terminal postscript eps font '"+plotFont+"'\n"+
						"replot\n\n";

			fileWriter.write(pltString);
			fileWriter.close();
			

			/**
			 * Process application message statistics
			 */		
			for (String messageType : applicationMessageTypes) {
				/**
				 * Containers for statistics about distribution of initialized/forwarded/delivered
				 * application messages WITH differentiation between application message type. 
				 */
				List<KeyValuePair<Integer>> initedMessageCountPerHost = new LinkedList<KeyValuePair<Integer>>();
				List<KeyValuePair<Integer>> forwardedMessageCountPerHost = new LinkedList<KeyValuePair<Integer>>();
				List<KeyValuePair<Integer>> deliveredMessageCountPerHost = new LinkedList<KeyValuePair<Integer>>();

				/**
				 * Get count of specific message type for each peer
				 */
				for (Entry<String, SinglePeerStatistic> peerEntry : peerStatistics.entrySet()) {
					String peerID = peerEntry.getKey();

					Map<String, Integer> messagesInited = peerEntry.getValue().getApplicationMsgsInited();
					int countInited = (messagesInited.get(messageType)==null)?0:messagesInited.get(messageType);
					initedMessageCountPerHost.add(new KeyValuePair<Integer>(peerID, countInited));

					Map<String, Integer> messagesForwared = peerEntry.getValue().getApplicationMsgsForwarded();
					int countForwarded = (messagesForwared.get(messageType)==null)?0:messagesForwared.get(messageType);
					forwardedMessageCountPerHost.add(new KeyValuePair<Integer>(peerID, countForwarded));

					Map<String, Integer> messagesDelivered = peerEntry.getValue().getApplicationMsgsDelivered();
					int countDelivered = (messagesDelivered.get(messageType)==null)?0:messagesDelivered.get(messageType);
					deliveredMessageCountPerHost.add(new KeyValuePair<Integer>(peerID, countDelivered));
				}

				/**
				 *  Sort the lists descending by the values
				 */
				Collections.sort(initedMessageCountPerHost);
				Collections.sort(forwardedMessageCountPerHost);
				Collections.sort(deliveredMessageCountPerHost);

				/**
				 * Write files for inited application messages
				 */
				String dataFileName_init = "app_init_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName_init, initedMessageCountPerHost);

				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), true);
				pltString = "reset\n"+
							"set title 'Initialized application messages'\n"+
							"set xlabel 'Peer'\n"+
							"set ylabel 'Number of messages'\n"+
							"set log y\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"app_init_"+messageType+".png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"mean(x)=a\n"+
							"fit mean(x) '"+dataFileName_init+"' via a\n"+
							"plot '"+dataFileName_init+"' using 2 title '"+messageType+"', \\\n"+
							"     mean(x) title 'Total average'\n"+
							"set output '"+"graphics/"+"app_init_"+messageType+".eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();

				/**
				 * Write files for forwarded application messages
				 */
				String dataFileName_forw = "app_forward_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName_forw, forwardedMessageCountPerHost);

				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), true);
				pltString = "reset\n"+
							"set title 'Forwarded application messages'\n"+
							"set xlabel 'Peer'\n"+
							"set ylabel 'Number of messages'\n"+
							"set log y\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"app_forward_"+messageType+".png'\n"+
							"#set terminal postscript eps \"Helvetica\" 24\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"mean(x)=a\n"+
							"fit mean(x) '"+dataFileName_forw+"' via a\n"+
							"plot '"+dataFileName_forw+"' using 2 title '"+messageType+"', \\\n"+
							"     mean(x) title 'Total average'\n"+
							"set output '"+"graphics/"+"app_forward_"+messageType+".eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();

				/**
				 * Write files for delivered application messages
				 */
				String dataFileName_deliv = "app_delivered_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName_deliv, deliveredMessageCountPerHost);

				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), true);
				pltString = "reset\n"+
							"set title 'Delivered application messages'\n"+
							"set xlabel 'Peer'\n"+
							"set ylabel 'Number of messages'\n"+
							"set log y\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"app_delivered_"+messageType+".png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"mean(x)=a\n"+
							"fit mean(x) '"+dataFileName_deliv+"' via a\n"+
							"plot '"+dataFileName_deliv+"' using 2 title '"+messageType+"', \\\n"+
							"     mean(x) title 'Total average'\n"+
							"set output '"+"graphics/"+"app_delivered_"+messageType+".eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();
				
				/**
				 * Plot for all three graphs in one plot
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), true);
				pltString = "reset\n"+
							"set title '"+messageType+"'\n"+
							"set xlabel 'Peer'\n"+
							"set ylabel 'Number of messages'\n"+
							"set log y\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"app_delivered_init_forward_"+messageType+".png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"plot '"+dataFileName_forw+"' using :2 title 'forwarded', \\\n"+
							"'"+dataFileName_deliv+"' using :2 title 'delivered', \\\n"+
							"'"+dataFileName_init+"' using :2 title 'initialized'\n"+
							"set output '"+"graphics/"+"app_delivered_init_forward_"+messageType+".eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();
				
			}

		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+gnuplotOutputPath);
			e.printStackTrace();
		}
	}
	
	private static void generatePerMsgTypeOlMsgGnuplotFiles(){
		FileWriter fileWriter;

		List<KeyValueValueValueQuadrupel<Double>> overallSentMsgTypeDistribution = new LinkedList<KeyValueValueValueQuadrupel<Double>>();
		List<KeyValueValueValueQuadrupel<Double>> overallRecMsgTypeDistribution = new LinkedList<KeyValueValueValueQuadrupel<Double>>();

		List<String> sentOlMsgFunctionPlotStrings = new LinkedList<String>();
		List<String> recOlMsgFunctionPlotStrings = new LinkedList<String>();

		/*
		 * Process overlay message statistics
		 */
		try {
			for (String messageType : overlayMessageTypes) {
				List<KeyValuePair<Integer>> sentMessageCountPerHost = new LinkedList<KeyValuePair<Integer>>();
				List<KeyValuePair<Integer>> receivedMessageCountPerHost = new LinkedList<KeyValuePair<Integer>>();

				double sumSent = 0;
				double sumRec = 0;

				/**
				 * Get count of specific message type for each peer
				 */
				for (String peerID : peerIDs) {

					SinglePeerStatistic peerStat = peerStatistics.get(peerID);

					Map<String, Integer> messagesSent = peerStat.getOverlayMsgsSent(); 
					int countSent = (messagesSent.get(messageType)==null)?0:messagesSent.get(messageType);
					sentMessageCountPerHost.add(new KeyValuePair<Integer>(peerID, countSent));

					// Computations for standard deviation
					sumSent += countSent;

					Map<String, Integer> messagesReceived = peerStat.getOverlayMsgsReceived();
					int countReceived = (messagesReceived.get(messageType)==null)?0:messagesReceived.get(messageType);
					receivedMessageCountPerHost.add(new KeyValuePair<Integer>(peerID, countReceived));

					// Computations for standard deviation
					sumRec += countReceived;
				}

				/*
				 *  Computations for standard deviation
				 */
				double averageSent = sumSent/peerIDs.size();
				double averageRec = sumRec/peerIDs.size();

				double sumSentLow = 0;
				double sumOfSquaresSentLow = 0;
				double sumSentHigh = 0;
				double sumOfSquaresSentHigh = 0;

				double sumRecLow = 0;
				double sumOfSquaresRecLow = 0;
				double sumRecHigh = 0;
				double sumOfSquaresRecHigh = 0;

				double sumSentOfDiffSquareLow = 0;
				double sumSentOfDiffSquareHigh = 0;
				double sumRecOfDiffSquareLow = 0;
				double sumRecOfDiffSquareHigh = 0;

				int countSentLow = 0;
				int countSentHigh = 0;
				int countRecLow = 0;
				int countRecHigh = 0;

				for (String peerID : peerIDs) {

					SinglePeerStatistic peerStat = peerStatistics.get(peerID);

					Map<String, Integer> messagesSent = peerStat.getOverlayMsgsSent(); 
					int countSent = (messagesSent.get(messageType)==null)?0:messagesSent.get(messageType);

					// Computations for standard deviation
					if(countSent < averageSent) {
						countSentLow++;
						sumSentLow += countSent;
						sumOfSquaresSentLow += Math.pow(countSent, 2);

						sumSentOfDiffSquareLow += Math.pow(countSent-averageSent, 2);
					} else {
						countSentHigh++;
						sumSentHigh += countSent;
						sumOfSquaresSentHigh += Math.pow(countSent, 2);

						sumSentOfDiffSquareHigh += Math.pow(countSent-averageSent, 2);
					}

					Map<String, Integer> messagesReceived = peerStat.getOverlayMsgsReceived();
					int countReceived = (messagesReceived.get(messageType)==null)?0:messagesReceived.get(messageType);

					// Computations for standard deviation
					if(countReceived < averageRec) {
						countRecLow++;
						sumRecLow += countReceived;
						sumOfSquaresRecLow += Math.pow(countReceived, 2);

						sumRecOfDiffSquareLow += Math.pow(countReceived-averageRec, 2);
					} else {
						countRecHigh++;
						sumRecHigh += countReceived;
						sumOfSquaresRecHigh += Math.pow(countReceived, 2);

						sumRecOfDiffSquareHigh += Math.pow(countReceived-averageRec, 2);
					}
				}

				double standardDeviationSentLow = Math.sqrt(sumSentOfDiffSquareLow/countSentLow);
				double standardDeviationSentHigh = Math.sqrt(sumSentOfDiffSquareHigh/countSentHigh);
				double standardDeviationRecLow = Math.sqrt(sumRecOfDiffSquareLow/countRecLow);
				double standardDeviationRecHigh = Math.sqrt(sumRecOfDiffSquareLow/countRecHigh);


				overallSentMsgTypeDistribution.add(new KeyValueValueValueQuadrupel<Double>(messageType, 
						averageSent, 
						averageSent-standardDeviationSentLow,
						averageSent+standardDeviationSentHigh));

				overallRecMsgTypeDistribution.add(new KeyValueValueValueQuadrupel<Double>(messageType,
						averageRec,
						averageRec-standardDeviationRecLow,
						averageRec+standardDeviationRecHigh));

				/**
				 *  Sort lists descending by the values for later write out
				 */
				Collections.sort(sentMessageCountPerHost);
				Collections.sort(receivedMessageCountPerHost);
				Collections.sort(overallSentMsgTypeDistribution);
				Collections.sort(overallRecMsgTypeDistribution);


				/**
				 * Write the gnuplot files for overlay messages
				 */
				/**
				 * Write files for sent overlay messages per peer
				 */
				String dataFileName = "ol_sent_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName, sentMessageCountPerHost);

				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				String functionPltString = "'"+dataFileName+"' using 2 title '"+messageType+"'";
				sentOlMsgFunctionPlotStrings.add(functionPltString);

				fileWriter = new FileWriter(new File(gnuplotOutputPath+"ol_msgs.plt"), false);
				String pltString = "reset\n"+
								"set title 'Sent overlay messages'\n"+
								"set xlabel 'Peer'\n"+
								"set ylabel 'Number of messages'\n"+
								"set style data lines\n"+
								"set output '"+"graphics/"+"ol_sent_"+messageType+".png'\n"+
								"set terminal png giant font '"+plotFont+"'\n"+
								"mean(x)=a\n"+
								"fit mean(x) '"+dataFileName+"' via a\n"+
								"plot "+functionPltString+", \\\n"+
								"     mean(x) title 'Total average'\n"+
								"set output '"+"graphics/"+"ol_sent_"+messageType+".eps'\n"+
								"set terminal postscript eps font '"+plotFont+"'\n"+
								"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();

				/**
				 * Write files for received overlay messages per peer
				 */
				dataFileName = "ol_rec_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName, receivedMessageCountPerHost);


				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				functionPltString = "'"+dataFileName+"' using 2 title '"+messageType+"'";
				recOlMsgFunctionPlotStrings.add(functionPltString);

				fileWriter = new FileWriter(new File(gnuplotOutputPath+"ol_msgs.plt"), true);
				pltString = "reset\n"+
							"set title 'Received overlay messages'\n"+
							"set xlabel 'Peer'\n"+
							"set ylabel 'Number of messages'\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"ol_rec_"+messageType+".png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"mean(x)=a\n"+
							"fit mean(x) '"+dataFileName+"' via a\n"+
							"plot '"+dataFileName+"' using 2 title '"+messageType+"', \\\n"+
							"     mean(x) title 'Total average'\n"+
							"set output '"+"graphics/"+"ol_rec_"+messageType+".eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();

				/**
				 * Write files for overall sent overlay messages
				 */
				dataFileName = "ol_overallSentMsgTypeDistribution.dat";
				writeDatFileForKeyValueValueValueQuadrupel(gnuplotOutputPath+dataFileName, overallSentMsgTypeDistribution);

				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"ol_msgs.plt"), true);
				pltString = "reset\n"+
							"set title 'Average number of overall sent messages by type'\n"+
							"set xlabel 'Message type'\n"+
							"set xrange [-0.5:"+(overlayMessageTypes.size()-0.5)+"]\n"+
							"set ylabel 'Message count'\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"ol_overallSentMsgTypeDistribution.png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"set xtics nomirror rotate by -45\n"+
							"plot 'ol_overallSentMsgTypeDistribution.dat' using 0:2:3:4:xtic(1) title 'standard deviation' with yerrorlines\n"+
							"set output '"+"graphics/"+"ol_overallSentMsgTypeDistribution.eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();

				/**
				 * Write files for overall received overlay messages
				 */
				dataFileName = "ol_overallRecMsgTypeDistribution.dat";
				writeDatFileForKeyValueValueValueQuadrupel(gnuplotOutputPath+dataFileName, overallRecMsgTypeDistribution);

				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"ol_msgs.plt"), true);
				pltString = "reset\n"+
							"set title 'Average number of overall received messages by type'\n"+
							"set xlabel 'Message type'\n"+
							"set xrange [-0.5:"+(overlayMessageTypes.size()-0.5)+"]\n"+
							"set ylabel 'Message count'\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"ol_overallRecMsgTypeDistribution.png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"set xtics nomirror rotate by -45\n"+
							"plot 'ol_overallRecMsgTypeDistribution.dat' using 0:2:3:4:xtic(1) title 'standard deviation' with yerrorlines\n"+
							"set output '"+"graphics/"+"ol_overallRecMsgTypeDistribution.eps'\n"+
							"set terminal postscript eps font '"+plotFont+"'\n"+
							"replot\n\n";

				fileWriter.write(pltString);
				fileWriter.close();
			}


			fileWriter = new FileWriter(new File(gnuplotOutputPath+"ol_msgs.plt"), true);


			String overallFunctionPltString = "";
			for(int i=0; i<sentOlMsgFunctionPlotStrings.size(); i++) {
				overallFunctionPltString += sentOlMsgFunctionPlotStrings.get(i);
				if(i<sentOlMsgFunctionPlotStrings.size()-1){
					overallFunctionPltString += ", ";
				} 
			}

			String pltString = "reset\n"+
								"set title 'Received overlay messages'\n"+
								"set xlabel 'Peer'\n"+
								"set ylabel 'Number of messages'\n"+
								"set log y\n"+
								"set style data lines\n"+
								"set output '"+"graphics/"+"ol_sent_overall.png'\n"+
								"set terminal png giant font '"+plotFont+"'\n"+
								"mean(x)=a\n"+
								"plot "+overallFunctionPltString+"\n"+
								"set output '"+"graphics/"+"ol_sent_overall.eps'\n"+
								"set terminal postscript eps font '"+plotFont+"'\n"+
								"replot\n\n";

			fileWriter.write(pltString);

			overallFunctionPltString = "";
			for(int i=0; i<recOlMsgFunctionPlotStrings.size(); i++) {
				overallFunctionPltString += recOlMsgFunctionPlotStrings.get(i);
				if(i<recOlMsgFunctionPlotStrings.size()-1){
					overallFunctionPltString += ", ";
				} 
			}

			pltString = "reset\n"+
						"set title 'Received overlay messages'\n"+
						"set xlabel 'Peer'\n"+
						"set ylabel 'Number of messages'\n"+
						"set log y\n"+
						"set style data lines\n"+
						"set output '"+"graphics/"+"ol_rec_overall.png'\n"+
						"set terminal png giant font '"+plotFont+"'\n"+
						"mean(x)=a\n"+
						"plot "+overallFunctionPltString+"\n"+
						"set output '"+"graphics/"+"ol_rec_overall.eps'\n"+
						"set terminal postscript eps font '"+plotFont+"'\n"+
						"replot\n\n";

			fileWriter.write(pltString);
			fileWriter.close();

		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+gnuplotOutputPath);
			e.printStackTrace();
		}
	}
	
	private static void generatePerMsgTypeQueryGnuplotFiles(){
		FileWriter fileWriter;
		
		/*
		 * Process query statistics
		 */
		for (String messageType : queriedApplicationMessageTypes) {
			List<KeyValuePair<Integer>> initQueryCountPerHost = new LinkedList<KeyValuePair<Integer>>();
			List<KeyValuePair<Integer>> succeededQueryCountPerHost = new LinkedList<KeyValuePair<Integer>>();
			
			/**
			 * Get count of specific message type for each peer
			 */
			for (Entry<String, SinglePeerStatistic> peerEntry : peerStatistics.entrySet()) {
				String peerID = peerEntry.getKey();
				
				Map<String, Integer> queriesInit = peerEntry.getValue().getQueriesInitiated(); 
				int countInit = (queriesInit.get(messageType)==null)?0:queriesInit.get(messageType);
				initQueryCountPerHost.add(new KeyValuePair<Integer>(peerID, countInit));
				
				Map<String, List<QueryStatistic>> queriesSucceeded = peerEntry.getValue().getQueriesSucceeded();
				int countSucceeded = 0;
				if(queriesSucceeded.get(messageType)!=null && queriesSucceeded.get(messageType).size()!=0)
					countSucceeded = queriesSucceeded.get(messageType).size();
					//TODO: Also use hops
				succeededQueryCountPerHost.add(new KeyValuePair<Integer>(peerID, countSucceeded));
			}
		
			/**
			 *  Sort the lists descending by the values
			 */
			Collections.sort(initQueryCountPerHost);
			Collections.sort(succeededQueryCountPerHost);
			
			
			/**
			 * Write the gnuplot files for overlay messages
			 */
			try {
				/**
				 * Write files for sent overlay messages
				 */
				String dataFileName = "query_init_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName, initQueryCountPerHost);
				
				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"query_stats.plt"), false);
				String pltString = "reset\n"+
								   "set title 'Initialized querys'\n"+
								   "set xlabel 'Peer'\n"+
								   "set ylabel 'Number of messages'\n"+
								   "set style data lines\n"+
								   "set output '"+"graphics/"+"query_init_"+messageType+".png'\n"+
								   "set terminal png giant font '"+plotFont+"'\n"+
								   "mean(x)=a\n"+
								   "fit mean(x) '"+dataFileName+"' via a\n"+
								   "plot '"+dataFileName+"' using 2 title '"+messageType+"', \\\n"+
								   "     mean(x) title 'Total average'\n"+
								   "set output '"+"graphics/"+"query_init_"+messageType+".eps'\n"+
								   "set terminal postscript eps font '"+plotFont+"'\n"+
								   "replot\n\n";
				
				fileWriter.write(pltString);
				fileWriter.close();
				
				/**
				 * Write files for received overlay messages
				 */
				dataFileName = "query_succeeded_"+messageType+".dat";
				writeDatFileForKeyValuePairs(gnuplotOutputPath+dataFileName, succeededQueryCountPerHost);

				
				/**
				 * Extend the plt-file for a later use by gnuplot 
				 */
				fileWriter = new FileWriter(new File(gnuplotOutputPath+"query_stats.plt"), true);
				pltString = "reset\n"+
							"set title 'Succeeded querys'\n"+
							"set xlabel 'Peer'\n"+
							"set ylabel 'Number of messages'\n"+
							"set style data lines\n"+
							"set output '"+"graphics/"+"query_succeeded_"+messageType+".png'\n"+
							"set terminal png giant font '"+plotFont+"'\n"+
							"mean(x)=a\n"+
							"fit mean(x) '"+dataFileName+"' via a\n"+
							"plot '"+dataFileName+"' using 2 title '"+messageType+"', \\\n"+
							"     mean(x) title 'Total average'\n"+
								   "set output '"+"graphics/"+"query_succeeded_"+messageType+".eps'\n"+
								   "set terminal postscript eps font '"+plotFont+"'\n"+
								   "replot\n\n";
				
				fileWriter.write(pltString);
				fileWriter.close();
				
			} catch (IOException e) {
				System.err.println("Could not write gnuplot files to "+gnuplotOutputPath);
				e.printStackTrace();
			}
		}
	}
	
	private static void generateQueryOverallSuccessRateFiles(){
		FileWriter fileWriter;
		try {
			/**
			 * Write files for overall success rate
			 */
			String dataFileName = "query_overallSuccessData.dat";
			//Append data to existing overall statistics
			fileWriter = new FileWriter(gnuplotOutputPath+dataFileName, true);
			
			double overAllSuccessRate = (double)overallSucceededQueryCount/(overallSucceededQueryCount+overallFailedQueryCount);
			
			/**
			 * Calculate standard deviation of failed queries from a certain initiator 
			 */
			double sum = 0;
			double sumOfSquares = 0;
			int count = 0;
			for (String peerID : peerIDs) {
				count++;
				if(failedQueryInitiator.containsKey(peerID)){
					sum += failedQueryInitiator.get(peerID);
					sumOfSquares += Math.pow(failedQueryInitiator.get(peerID), 2);
				}
			}
			double averageOfSum = (count>0)?(sum/count):0;
			double averageOfSumOfSquares = (count>0)?(sumOfSquares/count):0;
			double standardDeviation = Math.sqrt(averageOfSumOfSquares - Math.pow(averageOfSum,2));
			fileWriter.write(overAllSuccessRate+" "+standardDeviation+"\n");
			fileWriter.close();
			
			/**
			 * Generate plt-file for a later use by gnuplot 
			 */
			fileWriter = new FileWriter(new File(gnuplotOutputPath+"query_overAllStats.plt"), false);
			String pltString = "reset\n"+
							   "set title 'Overall success rate of lookup'\n"+
							   "set auto x\n"+
							   "set xlabel 'Scenario'\n"+
							   "set ylabel 'Success rate'\n"+
							   "set style data boxes\n"+
							   "set style fill solid 1.0\n"+
						       "set boxwidth .2\n"+
							   "set yrange [0:*]\n"+
							   "set output '"+"graphics/"+"query_overAllSuccessRate.png'\n"+
							   "set terminal png giant font '"+plotFont+"'\n"+
							   "mean(x)=a\n"+
							   "fit mean(x) '"+dataFileName+"' using :1 via a\n"+
							   "plot '"+dataFileName+"' using 1 title 'Query success rate', \\\n"+
							   "     mean(x) title 'Total average'\n"+
							   "set output '"+"graphics/"+"query_overAllSuccessRate.eps'\n"+
							   "set terminal postscript eps font '"+plotFont+"'\n"+
							   "replot\n\n";
			
			fileWriter.write(pltString);
			
			pltString = "reset\n"+
			   "set title 'Failed query count at hosts - standard deviation'\n"+
			   "set xlabel 'Scenario'\n"+
			   "set ylabel 'Standard deviation'\n"+
			   "set style data boxes\n"+
			   "set style fill solid 1.0\n"+
		       "set boxwidth .2\n"+
			   "set yrange [0:*]\n"+
			   "set output '"+"graphics/"+"query_overAllQueryFailDeviation.png'\n"+
			   "set terminal png giant font '"+plotFont+"'\n"+
			   "mean(x)=a\n"+
			   "fit mean(x) '"+dataFileName+"' using :2 via a\n"+
			   "plot '"+dataFileName+"' using 2 title 'Standard deviation of failed query count', \\\n"+
			   "     mean(x) title 'Total average'\n"+
			   "set output '"+"graphics/"+"query_overAllQueryFailDeviation.eps'\n"+
			   "set terminal postscript eps font '"+plotFont+"'\n"+
			   "replot\n\n";

			fileWriter.write(pltString);
			
			fileWriter.close();

		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+gnuplotOutputPath);
			e.printStackTrace();
		}
	}
	
	private static void generateStaleMsgRatioFiles(){
		FileWriter fileWriter;
		try {
			/**
			 * Write files for overall stale message ratio
			 */
			String dataFileName = "nl_overallStaleMsgRatio.dat";
			//Append data to existing overall statistics
			fileWriter = new FileWriter(gnuplotOutputPath+dataFileName, true);
			
			double overallStaleMsgRatio = overallLostNetMsgs/overallSentNetMsgs;
			
			fileWriter.write(Double.toString(overallStaleMsgRatio)+"\n");
			fileWriter.close();
			
			/**
			 * Generate plt-file for a later use by gnuplot 
			 */
			fileWriter = new FileWriter(new File(gnuplotOutputPath+"nl_overAllMsgRatio.plt"), false);
			String pltString = "reset\n"+
							   "set title 'Overall stale message ratio'\n"+
							   "set xlabel 'Scenario'\n"+
							   "set ylabel 'Stale message ratio'\n"+
							   "set style data boxes\n"+
							   "set style fill solid 1.0\n"+
						       "set boxwidth .2\n"+
							   "set yrange [0:*]\n"+
							   "set output '"+"graphics/"+"nl_overAllMsgRatio.png'\n"+
							   "set terminal png giant font '"+plotFont+"'\n"+
							   "mean(x)=a\n"+
							   "fit mean(x) '"+dataFileName+"' using :1 via a\n"+
							   "plot '"+dataFileName+"' using :1 title 'Stale message ratio', \\\n"+
							   "     mean(x) title 'Total average'\n"+
							   "set output '"+"graphics/"+"nl_overAllMsgRatio.eps'\n"+
							   "set terminal postscript eps font '"+plotFont+"'\n"+
							   "replot\n\n";
			
			fileWriter.write(pltString);
			fileWriter.close();

		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+gnuplotOutputPath);
			e.printStackTrace();
		}
	}
	
	private static void generatePerPeerAppMsgGnuplotFiles(){
		
		/**
		 * Write gnuplot files for KBR message distribution statistics
		 */
		String dataPath = gnuplotOutputPath+"app_kbrDistributionPerPeer.dat";
		
		try {
			/**
			 * Write data file
			 */
			FileWriter fileWriter = new FileWriter(new File(dataPath));

			// Write head for data - used in plot
			fileWriter.write("PeerID initialized forwarded received\n");
			
			for (String peerID : peerIDs.toArray(new String[0])) {
				SinglePeerStatistic peerStat = peerStatistics.get(peerID);
				
				// Write data line
				if(peerStat!=null){
					fileWriter.write(peerID+" "
								     +peerStat.applicationMsgsInitedCount()+" "
							         +peerStat.applicationMsgsForwardedCount()+" "
							         +peerStat.applicationMsgsDeliveredCount()+"\n");
				}
			}
			fileWriter.close();
			
			
			/**
			 * Extend the plt-file for a later use by gnuplot 
			 */
			fileWriter = new FileWriter(new File(gnuplotOutputPath+"app_msgs.plt"), true);
			String pltString = "reset\n"+
							   "set title \"KBR message statistics per peer\"\n"+
							   "set auto x\n"+
							   "set xlabel 'Message count'\n"+
							   "set ylabel 'Peer ID'\n"+
							   "set style data histogram\n"+
							   "set style histogram cluster gap 1\n"+
							   "set style fill solid border -1\n"+
							   "set boxwidth 0.5\n"+
							   "set xtic rotate by -45 \n"+
							   "set output '"+"graphics/"+"app_kbrDistributionPerPeer.png'\n"+
							   "set terminal png giant font '"+plotFont+"'\n"+
							   "plot 'app_kbrDistributionPerPeer.dat' u 2:xtic(1) ti col, '' u 3 ti col, '' u 4 ti col\n"+
							   "set output '"+"graphics/"+"app_kbrDistributionPerPeer.eps'\n"+
							   "set terminal postscript eps font '"+plotFont+"'\n"+
							   "replot\n\n";

			fileWriter.write(pltString);
			fileWriter.close();
		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+dataPath);
			e.printStackTrace();
		}
		

	}
	
	private static <T extends Comparable<T>> void writeDatFileForKeyValuePairs(String filePath, List<KeyValuePair<T>> dataList){
		ensureFoldersExist(filePath, true);
		try {
			FileWriter fileWriter = new FileWriter(new File(filePath));
			for (KeyValuePair keyValuePair : dataList) {
				fileWriter.write(keyValuePair.getKey()+" "+keyValuePair.getValue()+"\n");
			}
			fileWriter.close();
		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+filePath);
			e.printStackTrace();
		}
	}
	
	private static <T extends Comparable<T>> void writeDatFileForKeyValueValueTriples(String filePath, List<KeyValueValueTriple<T>> dataList){
		ensureFoldersExist(filePath, true);
		try {
			FileWriter fileWriter = new FileWriter(new File(filePath));
			for (KeyValueValueTriple<T> keyValuePair : dataList) {
				fileWriter.write(keyValuePair.getKey()+" "+keyValuePair.getValue1()+" "+keyValuePair.getValue2()+"\n");
			}
			fileWriter.close();
		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+filePath);
			e.printStackTrace();
		}
	}
	
	private static <T extends Comparable<T>> void writeDatFileForKeyValueValueValueQuadrupel(String filePath, List<KeyValueValueValueQuadrupel<T>> dataList){
		ensureFoldersExist(filePath, true);
		try {
			FileWriter fileWriter = new FileWriter(new File(filePath));
			for (KeyValueValueValueQuadrupel<T> keyValuePair : dataList) {
				fileWriter.write(keyValuePair.getKey()+" "+keyValuePair.getValue1()+" "+keyValuePair.getValue2()+" "+keyValuePair.getValue3()+"\n");
			}
			fileWriter.close();
		} catch (IOException e) {
			System.err.println("Could not write gnuplot files to "+filePath);
			e.printStackTrace();
		}
	}
	
	/**
	 * Creates parent directories if not existing
	 * 
	 * @param PathToFile
	 */
	private static void ensureFoldersExist(String PathToFile, boolean isFile){
		File f = new File(PathToFile);
		if(!f.exists())
			if(isFile)
				f.getParentFile().mkdirs();
			else
				f.mkdirs();
	}
}
