package de.tud.kom.p2psim.impl.util.vis.analyzer.ol.chord;

import java.awt.Color;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.tud.kom.p2psim.api.common.Host;
import de.tud.kom.p2psim.api.common.Message;
import de.tud.kom.p2psim.api.common.Operation;
import de.tud.kom.p2psim.api.network.NetID;
import de.tud.kom.p2psim.api.overlay.OverlayNode;
import de.tud.kom.p2psim.impl.overlay.BootstrapManager;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordContact;
import de.tud.kom.p2psim.impl.overlay.dht.chord.ChordNode;
import de.tud.kom.p2psim.impl.overlay.dht.chord.FingerTable;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.GetInfoRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.LookupReply;
import de.tud.kom.p2psim.impl.overlay.dht.chord.messages.LookupRequest;
import de.tud.kom.p2psim.impl.overlay.dht.chord.operations.LookupOperation;
import de.tud.kom.p2psim.impl.util.vis.analyzer.OverlayAdapter;
import de.tud.kom.p2psim.impl.util.vis.analyzer.Translator.EdgeHandle;
import de.tud.kom.p2psim.impl.util.vis.analyzer.ol.gnutella.DocCountM;
import de.tud.kom.p2psim.impl.util.vis.analyzer.positioners.SchematicPositioner;

/**
 * Visualisiert das Overlay Chord spezifisch
 * 
 * @author Leo Nobach
 * @version 3.0, 30.10.2008
 *
 */
public class ChordAdapter extends OverlayAdapter{
	
	protected Map<NetID, ChordNode> overlayImpls = new HashMap<NetID, ChordNode>();
	
	/**
	 * Liste mit bisherigen Vorgängern/Nachfolgern von Knoten.
	 */
	protected Map<Host, EdgeHandle> predecessors = new HashMap <Host, EdgeHandle>();
	
	protected Map<NetID, Set<NetID>> fingertables = new HashMap<NetID, Set<NetID>>();
	protected Map<ChordNode, Integer> docCount = new HashMap<ChordNode, Integer>();
	
	protected Set<Integer> lookupRequestsOpIDs = new HashSet<Integer>();
	
	public ChordAdapter() {
		addOverlayImpl(ChordNode.class);
		addOverlayImpl(LookupRequest.class);
		addOverlayImpl(LookupReply.class);
		addOverlayImpl(GetInfoRequest.class);
		addOverlayImpl(GetInfoReply.class);

		addOverlayNodeMetric(HostsInFT.class);
		addOverlayNodeMetric(ChordIDM.class);
		addOverlayNodeMetric(DocCountM.class);
	}
	
	@Override
	public String getOverlayName() {
		return "Chord";
	}

	@Override
	public void handleLeavingHost(Host host) {
		// Nichts zu tun
	}

	@Override
	public void handleNewHost(Map<String, Serializable> attributes, Host host, OverlayNode overlayNode) {
		
		ChordNode chordNode = (ChordNode)overlayNode;
		
		attributes.put("ChordID", chordNode.getChordID());
		
		//Initialisiere Dokumenten-Count
		docCount.put(chordNode, chordNode.getIndex().size());
		attributes.put("doc_count", chordNode.getIndex().size());
		
		overlayImpls.put(host.getNetLayer().getNetID(), chordNode);
		
		
	}

	@Override
	public void handleNewHostAfter(Host host, OverlayNode overlayNode) {
		//Nichts zu tun.
	}

	@Override
	public void handleOverlayMsg(Message omsg, Host from, NetID fromID, Host to, NetID toID) {
		
		if (from != null) refreshFingertable(from);
		if (from != null) updatePredecessorFor(from);
		if (to != null) updatePredecessorFor(to);
		
		updateDocCount(fromID);
		updateDocCount(toID);
		
		//Mit dem  != null wird überprüft, ob ein entsprechender Host auch existiert.
		//Manchmal versenden Hosts Nachrichten, ohne dass sie zum Szenario dazugekommen
		//sind. Komisch.
		
		if (omsg instanceof LookupReply) {
			LookupReply lookupReply = (LookupReply) omsg;
			if (lookupRequestsOpIDs.remove(lookupReply.getOpID())) {
				this.flashEdge(fromID, toID, Color.BLUE, "LookupReply ForValue", omsg.getClass());
			}
		}
		if (omsg instanceof LookupRequest) {
			LookupRequest lookupRequest = (LookupRequest) omsg;
			if (lookupRequest.forValue()) {
				lookupRequestsOpIDs.add(lookupRequest.getOpID());
				this.flashEdge(fromID, toID, Color.RED, "LookupRequest ForValue", omsg.getClass());
			}
		}
		if (omsg instanceof GetInfoRequest) {
			GetInfoRequest getInfoRequest = (GetInfoRequest) omsg;
			if (getInfoRequest.getCmdID() == -2) {
				//getInfoRequestMap.put(currentOpID, getInfoRequest);
				this.flashEdge(fromID, toID, Color.BLACK, "GetInfoRequest", omsg.getClass());
			}
		}
		if (omsg instanceof GetInfoReply) {
			GetInfoReply getInfoReply = (GetInfoReply) omsg;
			if (getInfoReply.getCmdID() == -2) {
				//getInfoReplyMap.put(currentOpID, getInfoReply);
				this.flashEdge(fromID, toID, Color.YELLOW, "GetInfoReply", omsg.getClass());
			}
		}
		
	}
	
	private void updateDocCount(NetID nodeID) {
		ChordNode node = overlayImpls.get(nodeID);
		if (node != null) {
			int newDocCount = node.getIndex().size();
			int oldDocCount = docCount.get(node);
			
			if (newDocCount != oldDocCount) {
				getTranslator().nodeAttributeChanged(nodeID, "doc_count", newDocCount);
				docCount.put(node, newDocCount);
				System.out.println("DocCount gewechselt!");
			}
		}
	}

	private void updatePredecessorFor(Host host){
		NetID pre = getOverlayPredecessor(host);
		EdgeHandle old_pre = predecessors.get(host);
		//System.out.println("PREDECESSORS: " + pre + "|" + ((old_pre != null)?old_pre.getTo():"null"));
		
		if (pre == null) {
			if (old_pre != null) old_pre.remove();
		} else if (old_pre == null || !pre.equals(old_pre.getTo())) {
			if (old_pre != null) old_pre.remove();
			EdgeHandle newEdge = this.addEdge(netID(host), pre, Color.GREEN, "succ/pre");
			predecessors.put(host, newEdge);		
		}
	}
	
	private NetID netID(Host host) {
		return host.getNetLayer().getNetID();
	}
	
	private NetID getOverlayPredecessor(Host fromNode) {
		ChordContact pre = ((ChordNode)fromNode.getOverlay(ChordNode.class)).getPre();
			if (pre != null) return pre.getTransInfo().getNetId();
			else return null;
	}
	
	private void refreshFingertable(Host fromNode) {
		FingerTable ft = ((ChordNode)fromNode.getOverlay(ChordNode.class)).getFingerTable();
		HashSet<ChordContact> ft_set = new HashSet<ChordContact>(ft.allContacts());
		if (ft_set.contains(null)) ft_set.remove(null);
		
		//System.out.println(fromNode + " Größe FT:" + ft_set.size());
		//System.out.println(fromNode + " FT:" + ft_set);
		
			if (ft != null) {
				
				Set<NetID> vis_ft = fingertables.get(netID(fromNode));
				if (vis_ft == null) {
					vis_ft = new HashSet<NetID>();
					fingertables.put(netID(fromNode), vis_ft);
				}
				boolean ft_changed = false;
				for (ChordContact con : ft_set) {
					NetID newNetID = con.getTransInfo().getNetId();
					if (con != null && !vis_ft.contains(newNetID)) {
						this.addEdge(netID(fromNode), newNetID, Color.BLUE, "finger");
						vis_ft.add(newNetID);
						ft_changed = true;
						
					}
				}
				if (ft_changed) this.getTranslator().nodeAttributeChanged(fromNode.getNetLayer().getNetID(), "ft_hosts", ft_set.size());
			}
	}

	@Override
	public SchematicPositioner getNewPositioner() {
		return new ChordRingPositioner();
	}

	@Override
	public BootstrapManager getBootstrapManagerFor(OverlayNode node) {
		if (node instanceof ChordNode) {
			return ((ChordNode)node).getBootstrap();
		} else {
			return null;
		}
	}

	@Override
	public void handleOperation(Host host, Operation<?> op, boolean finished) {
		if (finished) {
			if (op instanceof LookupOperation) {
				LookupOperation lookupOperation = (LookupOperation) op;
				//if (lookupOperation.forValue())
					//lookupOperationMap.put(lookupOperation.getOperationID(), lookupOperation);
			}
		}
	}

}
