package de.tud.kom.p2psim.impl.util.vis.util.visualgraph;

import java.util.Collection;
import java.util.Vector;
import java.io.Serializable;

import de.tud.kom.p2psim.impl.util.vis.model.MetricObject;
import de.tud.kom.p2psim.impl.util.vis.model.ModelIterator;

/**
 * <b>"Hypergenerischer" Graph.</b>
 * 
 * Die übergebenen Typen müssen die abstrakten Klassen Node und Edge erweitern.
 * Der Vorteil bei der Verwendung dieses Graphen ist, dass eigentlich nie gecastet werden muss.
 * @author Leo
 *
 * @param <TNode>
 * @param <TEdge> 
 * 
 */

public class VisualGraph<TNode extends Node<TNode, TEdge>, TEdge extends Edge<TNode, TEdge>> implements Serializable{
	

	/**
	 * 
	 */
	private static final long serialVersionUID = -2796546862703465617L;
	
	public Vector<Node<TNode, TEdge>> nodes = new Vector<Node<TNode, TEdge>>();
	public Vector<Edge<TNode, TEdge>> edges = new Vector<Edge<TNode, TEdge>>();
	
	public Vector<VisRectangle> rectangles = new Vector<VisRectangle>();
	
	public void addNode(Node<TNode, TEdge> node) {
		nodes.add(node);
		node.setGraph(this);
	}
	
	public void addEdge(Edge<TNode, TEdge> edge) {
		synchronized(edges) {
			edges.add(edge);
			edge.setGraph(this);
		}
	}
	
	public void removeNode(Node<TNode, TEdge> node) {
		
		synchronized(nodes) {
			synchronized(edges){
				//alle Kanten entfernen, die mit dem Knoten zusammenhängen
				Collection<Edge<TNode,TEdge>> tmpEdges = new Vector<Edge<TNode,TEdge>>();
				tmpEdges.addAll(node.edges);
				
				for (Edge<TNode,TEdge> e: tmpEdges) {
					removeEdge(e);
				}
				nodes.remove(node);
				node.unsetGraph();
			}
		}
	}
	
	public void removeEdge(Edge<TNode, TEdge> edge) {
		synchronized(edges) {
			edges.remove(edge);
			edge.unsetGraph();
		}
	}
	
	public void addRectangle(VisRectangle rect) {
		rectangles.add(rect);
	}
	
	public void removeRectangle(VisRectangle rect) {
		rectangles.remove(rect);
	}
	
	/**
	 * Iteriert über die Elemente des Graphen
	 * Erst Nodes, dann Edges.
	 * @param it
	 */
	
	public void iterate(ModelIterator it, MetricObject selectedObject) {
		this.iterateRectangles(it);
		if (!it.shallStop()) this.iterateNodes(it);
		if (!it.shallStop()) this.iterateEdges(it);
	}
	
	/**
	 * Iteriert über die Elemente des Graphen
	 * Erst Edges, dann Nodes.
	 * @param it
	 */
	
	public void iterateBottomTop(ModelIterator it) {
		this.iterateRectangles(it);
		if (!it.shallStop()) this.iterateEdges(it);
		if (!it.shallStop()) this.iterateNodes(it);
	}
	
	/**
	 * Iteriert über alle Edges
	 * @param it
	 */
	public void iterateEdges(ModelIterator it) {		
		
		synchronized(edges) {
			for (Edge<TNode, TEdge> e : edges){
				e.iterate(it);
				if (it.shallStop()) break;
			}
		}

	}
	
	/**
	 * Iteriert über alle Nodes
	 * @param it
	 */
	public synchronized void iterateNodes(ModelIterator it) {
		synchronized(nodes){
			for (Node<TNode, TEdge> n : nodes){
				n.iterate(it);
				if (it.shallStop()) break;
			}
		}
	}
	
	/**
	 * Iteriert über alle Rectangles
	 * @param it
	 */
	public synchronized void iterateRectangles(ModelIterator it) {
		for (VisRectangle rect : rectangles){
			rect.iterate(it);
			if (it.shallStop()) break;
		}
	}
	
	/**
	 * 	Setzt den Graphen und seine Elemente zurück.
	 */
	public synchronized void reset() {
		for (Node<TNode, TEdge> n : nodes) {
			n.edges.clear();
		}
		nodes.clear();
		edges.clear();
	}
	
}
