package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable;

import static org.junit.Assert.*;


import java.math.BigInteger;
import java.util.Comparator;

import org.junit.BeforeClass;
import org.junit.Test;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.AddNodeVisitor;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.BranchNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.LeafNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTableEntry;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTableComparators.RoutingTableEntryHierarchyComparator;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.setup.ConfigStub;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.HKademliaOverlayID;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayContact;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;
import de.tud.kom.p2psim.util.helpers.TestHelper;

/**
 * Tests for AddNodeVisitor (except proximity listener).
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class AddNodeVisitorTest {

    private static ConfigStub config;

    /**
     * Configure necessary simulator-wide variables
     */
    @BeforeClass
    public static void beforeClass() {
	// Create dummy scheduler
	TestHelper.initSimulator();
	config = new ConfigStub();
	config.bucketSize = 2;
	config.routingTreeOrder = 2;
	config.idLength = 6;
	config.replacementCacheSize = 2;
	config.hierarchyDepth = 2;
	config.hierarchyTreeOrder = 1;
    }

    /**
     * Test insertion into nonsplittable leaf node and check reset of stale
     * counter and last seen time, too.
     */
    @Test
    public void testAddNonsplittable() throws Exception {
	ParentNodeStub<KademliaOverlayID> parent = new ParentNodeStub<KademliaOverlayID>(
		0, new KademliaOverlayID(0, config), null, BigInteger.ZERO);
	// not splittable:
	LeafNode<KademliaOverlayID> empty = new LeafNode<KademliaOverlayID>(
		BigInteger.ONE, parent, config);

	KademliaOverlayContact<KademliaOverlayID> C010000 = TestHelper
		.createContact(new KademliaOverlayID(16, config), 16);
	KademliaOverlayContact<KademliaOverlayID> C010011 = TestHelper
		.createContact(new KademliaOverlayID(19, config), 19);
	KademliaOverlayContact<KademliaOverlayID> C011100 = TestHelper
		.createContact(new KademliaOverlayID(28, config), 28);
	KademliaOverlayContact<KademliaOverlayID> C011101 = TestHelper
		.createContact(new KademliaOverlayID(29, config), 29);
	KademliaOverlayContact<KademliaOverlayID> C011111 = TestHelper
		.createContact(new KademliaOverlayID(31, config), 31);

	assertEquals("Bucket is initially empty", 0, empty.kBucket.size());
	assertEquals("Cache is initially empty", 0, empty.replacementCache
		.size());
	AddNodeVisitor<KademliaOverlayID> add010000 = AddNodeVisitor
		.getAddNodeVisitor(C010000, null, null, config);
	empty.accept(add010000);
	assertTrue(
		"Bucket now contains 010000",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010000)));

	// check reset of stale counter & last seen time (only once)
	KademliaOverlayID i010011 = new KademliaOverlayID("010011", config);
	AddNodeVisitor<KademliaOverlayID> add010011 = AddNodeVisitor
		.getAddNodeVisitor(C010011, null, null, config);
	empty.accept(add010011);
	empty.kBucket.get(i010011).increaseStaleCounter();
	assertEquals("Stale counter of 010011 is now 1", 1, empty.kBucket.get(
		i010011).getStaleCounter());
	TestHelper.setSimulationTime(239l);
	add010011 = AddNodeVisitor.getAddNodeVisitor(C010011, null, null,
		config);
	empty.accept(add010011); // add again
	assertTrue(
		"Bucket (still) contains 010011",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010011)));
	assertTrue(
		"Bucket still contains 010000",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010000)));
	assertEquals("Stale counter of 010011 is now zero", 0, empty.kBucket
		.get(i010011).getStaleCounter());
	assertEquals("Last seen time of 010011 is 239", 239l, empty.kBucket
		.get(i010011).getLastSeen());

	// add contact (in cache)
	AddNodeVisitor<KademliaOverlayID> add011100 = AddNodeVisitor
		.getAddNodeVisitor(C011100, null, null, config);
	empty.accept(add011100);
	assertTrue(
		"Cache now contains 011100",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C011100)));
	assertTrue(
		"Bucket still contains 010011",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010011)));
	assertTrue(
		"Bucket still contains 010000",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010000)));
	AddNodeVisitor<KademliaOverlayID> add011101 = AddNodeVisitor
		.getAddNodeVisitor(C011101, null, null, config);
	empty.accept(add011101);
	assertTrue(
		"Cache now contains 011101",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C011101)));
	assertTrue(
		"Cache still contains 011100",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C011100)));
	assertTrue(
		"Bucket still contains 010011",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010011)));
	assertTrue(
		"Bucket still contains 010000",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010000)));

	AddNodeVisitor<KademliaOverlayID> add011111 = AddNodeVisitor
		.getAddNodeVisitor(C011111, null, null, config);
	empty.accept(add011111); // should be dropped
	assertTrue(
		"Cache still contains 011101",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C011101)));
	assertTrue(
		"Cache still contains 011100",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C011100)));
	assertTrue(
		"Bucket still contains 010011",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010011)));
	assertTrue(
		"Bucket still contains 010000",
		empty.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010000)));
	assertEquals("Bucket size is two", 2, empty.kBucket.size());
	assertEquals("Cache size is two", 2, empty.replacementCache.size());
    }

    /**
     * Test prioritised insertion into nonsplittable leaf node.
     */
    @Test
    public void testAddNonsplittableAndPrioritised() throws Exception {
	HKademliaOverlayID i001010 = new HKademliaOverlayID(10, config); // own
	// ID
	Comparator<RoutingTableEntry<HKademliaOverlayID>> comp = new RoutingTableEntryHierarchyComparator<HKademliaOverlayID>(
		i001010);

	ParentNodeStub<HKademliaOverlayID> parent = new ParentNodeStub<HKademliaOverlayID>(
		0, i001010, null, BigInteger.ZERO);
	// not splittable (prefix 01):
	LeafNode<HKademliaOverlayID> empty = new LeafNode<HKademliaOverlayID>(
		BigInteger.ONE, parent, config);

	KademliaOverlayContact<HKademliaOverlayID> C010000 = TestHelper
		.createContact(new HKademliaOverlayID(16, config), 16);
	KademliaOverlayContact<HKademliaOverlayID> C010011 = TestHelper
		.createContact(new HKademliaOverlayID(19, config), 19);
	KademliaOverlayContact<HKademliaOverlayID> C011010 = TestHelper
		.createContact(new HKademliaOverlayID(26, config), 26);
	KademliaOverlayContact<HKademliaOverlayID> C011100 = TestHelper
		.createContact(new HKademliaOverlayID(28, config), 28);
	KademliaOverlayContact<HKademliaOverlayID> C011101 = TestHelper
		.createContact(new HKademliaOverlayID(29, config), 29);
	KademliaOverlayContact<HKademliaOverlayID> C011110 = TestHelper
		.createContact(new HKademliaOverlayID(30, config), 30);

	AddNodeVisitor<HKademliaOverlayID> add010000 = AddNodeVisitor
		.getAddNodeVisitor(C010000, comp, null, config);
	empty.accept(add010000);
	assertTrue("Bucket now contains 010000 (priority 0)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010000)));

	AddNodeVisitor<HKademliaOverlayID> add010011 = AddNodeVisitor
		.getAddNodeVisitor(C010011, comp, null, config);
	empty.accept(add010011);
	assertTrue("Bucket now contains 010011 (priority 1)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010011)));
	assertTrue("Bucket still contains 010000 (priority 0)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010000)));

	AddNodeVisitor<HKademliaOverlayID> add011100 = AddNodeVisitor
		.getAddNodeVisitor(C011100, comp, null, config);
	empty.accept(add011100);
	assertTrue(
		"Cache now contains 011100 (priority 0)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C011100)));
	assertTrue("Bucket still contains 010011 (priority 1)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010011)));
	assertTrue("Bucket still contains 010000 (priority 0)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010000)));

	AddNodeVisitor<HKademliaOverlayID> add011110 = AddNodeVisitor
		.getAddNodeVisitor(C011110, comp, null, config);
	empty.accept(add011110);
	assertTrue("Bucket now contains 011110 (priority 2)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C011110)));
	assertTrue(
		"Cache now contains 010000 (priority 0 - moved)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C010000)));
	assertFalse(
		"Bucket no longer contains 010000 (priority 0 - moved)",
		empty.kBucket
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C010000)));
	assertTrue(
		"Cache still contains 011100 (priority 0)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C011100)));
	assertTrue("Bucket still contains 010011 (priority 1)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010011)));

	AddNodeVisitor<HKademliaOverlayID> add011101 = AddNodeVisitor
		.getAddNodeVisitor(C011101, comp, null, config);
	empty.accept(add011101);
	assertFalse(
		"Bucket does not contain 011101 (priority 0 - dropped)",
		empty.kBucket
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C011101)));
	assertFalse(
		"Cache does not contain 011101 (priority 0 - dropped)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C011101)));
	assertTrue("Bucket still contains 011110 (priority 2)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C011110)));
	assertTrue(
		"Cache still contains 010000 (priority 0)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C010000)));
	assertTrue(
		"Cache still contains 011100 (priority 0)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C011100)));
	assertTrue("Bucket still contains 010011 (priority 1)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C010011)));

	AddNodeVisitor<HKademliaOverlayID> add011010 = AddNodeVisitor
		.getAddNodeVisitor(C011010, comp, null, config);
	empty.accept(add011010);
	assertTrue("Bucket now contains 011010 (priority 2)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C011010)));
	assertFalse(
		"Bucket no longer contains 010011 (priority 1 - moved",
		empty.kBucket
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C010011)));
	assertTrue(
		"Cache now contains 010011 (priority 1 - moved",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C010011)));
	assertTrue("Bucket still contains 011110 (priority 2)", empty.kBucket
		.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
			C011110)));
	assertTrue(
		"Cache still contains 010000 or 011100 (priority 0)",
		empty.replacementCache
			.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
				C010000))
			|| empty.replacementCache
				.containsValue(new RoutingTableEntry<HKademliaOverlayID>(
					C011100)));

	assertEquals("Bucket size is two", 2, empty.kBucket.size());
	assertEquals("Cache size is two", 2, empty.replacementCache.size());
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.AddNodeVisitor#isSplittable(hkademlia.components.routingtable.LeafNode)}.
     */
    @Test
    public void testIsSplittable() {
	AddNodeVisitor<KademliaOverlayID> dummy = AddNodeVisitor
		.getAddNodeVisitor(TestHelper.createStdContact("0", config),
			null, null, config);
	ParentNodeStub<KademliaOverlayID> parent = new ParentNodeStub<KademliaOverlayID>(
		-1, new KademliaOverlayID(0, config), null, BigInteger.ZERO);
	LeafNode<KademliaOverlayID> leaf = new LeafNode<KademliaOverlayID>(
		BigInteger.ZERO, parent, config);

	// Prefixes: empty has 0
	parent.level = -1;
	assertTrue("Bucket with prefix 0 at level 0 is splittable", dummy
		.isSplittable(leaf));
	parent.level = (config.idLength / config.routingTreeOrder) - 2;
	assertTrue(
		"Bucket with prefix 0 at level ID_LENGTH/B - 1 is splittable",
		dummy.isSplittable(leaf));
	parent.level = (config.idLength / config.routingTreeOrder) - 1;
	assertFalse(
		"Bucket with prefix 0 at level ID_LENGTH/B is not splittable",
		dummy.isSplittable(leaf));

	parent.level = 0;
	leaf = new LeafNode<KademliaOverlayID>(BigInteger.ONE, parent, config);
	assertFalse(
		"Bucket with prefix 01 at level 1 is not splittable for owner ID 000..00",
		dummy.isSplittable(leaf));
	parent.id = new KademliaOverlayID(29, config);
	assertTrue(
		"Bucket with prefix 01 at level 1 is splittable for owner ID 011101",
		dummy.isSplittable(leaf));
	parent.level = 1;
	leaf = new LeafNode<KademliaOverlayID>(BigInteger.TEN, parent, config);
	assertFalse(
		"Bucket with prefix 1010 at level 2 is not splittable for owner ID 011101",
		dummy.isSplittable(leaf));
	parent.id = new KademliaOverlayID(40, config);
	assertTrue(
		"Bucket with prefix 1010 at level 2 is splittable for owner ID 101000",
		dummy.isSplittable(leaf));
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.AddNodeVisitor#buildLeafNode(hkademlia.components.routingtable.LeafNode, int, hkademlia.components.routingtable.ParentNode)}.
     * 
     * Add four contacts into a splittable bucket, verify that the last lookup
     * time is correctly copied, the contacts correctly distributed and the new
     * prefixes correctly computed.
     */
    @Test
    public void testAddAndSplit() {
	AddNodeVisitor<KademliaOverlayID> visitor;
	ParentNodeStub<KademliaOverlayID> parent = new ParentNodeStub<KademliaOverlayID>(
		-1, new KademliaOverlayID(29, config), null, BigInteger.ZERO);
	LeafNode<KademliaOverlayID> root = new LeafNode<KademliaOverlayID>(
		parent, config);
	root.lastLookup = 123;
	KademliaOverlayContact<KademliaOverlayID> C000000 = TestHelper
		.createContact(new KademliaOverlayID(0, config), 0);
	KademliaOverlayContact<KademliaOverlayID> C010000 = TestHelper
		.createContact(new KademliaOverlayID(16, config), 16);
	KademliaOverlayContact<KademliaOverlayID> C100000 = TestHelper
		.createContact(new KademliaOverlayID(32, config), 32);
	KademliaOverlayContact<KademliaOverlayID> C110000 = TestHelper
		.createContact(new KademliaOverlayID(48, config), 48);

	visitor = AddNodeVisitor.getAddNodeVisitor(C000000, null, null, config);
	root.accept(visitor);
	visitor = AddNodeVisitor.getAddNodeVisitor(C010000, null, null, config);
	root.accept(visitor);
	// now split
	visitor = AddNodeVisitor.getAddNodeVisitor(C100000, null, null, config);
	root.accept(visitor);

	assertNotSame("Child of parent should have been replaced",
		parent.child, root);
	assertTrue("Child of parent should be BranchNode",
		parent.child instanceof BranchNode);
	BranchNode<KademliaOverlayID> branch = (BranchNode<KademliaOverlayID>) parent.child;
	assertEquals("Branch node has same prefix as old root node", root
		.getPrefix(), branch.getPrefix());
	assertEquals("Branch node has same parent as old root node", root
		.getParent(), branch.getParent());
	assertEquals("Branch node has exactly 2^BTREE=4 children", 4,
		branch.children.size());

	LeafNode<KademliaOverlayID> leaf = (LeafNode<KademliaOverlayID>) branch.children
		.get(BigInteger.ZERO);
	assertTrue(
		"New child 00 contains 000000",
		leaf.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C000000)));
	assertEquals("New child 00 contains only one contact", 1, leaf.kBucket
		.size());
	assertEquals("New child 00 has empty cache", 0, leaf.replacementCache
		.size());
	assertEquals("New child 00 has lastLookup=123", 123l, leaf.lastLookup);
	assertEquals("New child 00 has prefix 00", BigInteger.ZERO, leaf
		.getPrefix());

	leaf = (LeafNode<KademliaOverlayID>) branch.children
		.get(BigInteger.ONE);
	assertTrue(
		"New child 01 contains 010000",
		leaf.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C010000)));
	assertEquals("New child 01 contains only one contact", 1, leaf.kBucket
		.size());
	assertEquals("New child 01 has empty cache", 0, leaf.replacementCache
		.size());
	assertEquals("New child 01 has lastLookup=123", 123l, leaf.lastLookup);
	assertEquals("New child 01 has prefix 01", BigInteger.ONE, leaf
		.getPrefix());

	leaf = (LeafNode<KademliaOverlayID>) branch.children.get(BigInteger
		.valueOf(2));
	assertTrue(
		"New child 10 contains 100000",
		leaf.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C100000)));
	assertEquals("New child 10 contains only one contact", 1, leaf.kBucket
		.size());
	assertEquals("New child 10 has empty cache", 0, leaf.replacementCache
		.size());
	assertEquals("New child 10 has lastLookup=123", 123l, leaf.lastLookup);
	assertEquals("New child 10 has prefix 10", BigInteger.valueOf(2), leaf
		.getPrefix());

	// add another contact (also tests add rule for BranchNodes)
	visitor = AddNodeVisitor.getAddNodeVisitor(C110000, null, null, config);
	branch.accept(visitor);

	leaf = (LeafNode<KademliaOverlayID>) branch.children.get(BigInteger
		.valueOf(3));
	assertTrue(
		"New child 11 contains 110000",
		leaf.kBucket
			.containsValue(new RoutingTableEntry<KademliaOverlayID>(
				C110000)));
	assertEquals("New child 11 contains only one contact", 1, leaf.kBucket
		.size());
	assertEquals("New child 11 has empty cache", 0, leaf.replacementCache
		.size());
	assertEquals("New child 11 has lastLookup=123", 123l, leaf.lastLookup);
	assertEquals("New child 11 has prefix 11", BigInteger.valueOf(3), leaf
		.getPrefix());
    }

}
