package de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable;

import static org.junit.Assert.*;


import java.math.BigInteger;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.AbstractNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.BranchNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.LeafNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.Node;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.NodeVisitor;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.ParentNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.PseudoRootNode;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.components.routingtable.RoutingTableConfig;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.setup.ConfigStub;
import de.tud.kom.p2psim.impl.overlay.dht.kademlia2.types.KademliaOverlayID;

/**
 * Test case for PseudoRootNode.
 * 
 * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
 */
public class PseudoRootNodeTest {

    private static ConfigStub config;

    private KademliaOverlayID id;

    private PseudoRootNode<KademliaOverlayID> testObj;

    /**
     * Initialise environment.
     */
    @BeforeClass
    public static void setUpBeforeClass() {
	config = new ConfigStub();
	config.idLength = 6;
    }

    /**
     * Setup
     */
    @Before
    public void setUp() {
	id = new KademliaOverlayID(43, config);
	testObj = new PseudoRootNode<KademliaOverlayID>(id, config);
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.PseudoRootNode#getOwnID()}.
     */
    @Test
    public void testGetOwnID() {
	assertEquals("Own ID is 43", id, testObj.getOwnID());
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.PseudoRootNode#registerNewChild(hkademlia.components.routingtable.Node)}.
     */
    @Test
    public void testRegisterNewChild() {
	// automatically registers itself as child to parent==testObj
	Node<KademliaOverlayID> child = new NodeStub(BigInteger.ZERO, testObj,
		config);
	assertEquals("Child should be registered as root", child, testObj.root);
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.PseudoRootNode#getLevel()}.
     */
    @Test
    public void testGetLevel() {
	assertEquals("Level of pseudo root is always -1", -1, testObj
		.getLevel());
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.PseudoRootNode#accept(hkademlia.components.routingtable.NodeVisitor)}.
     */
    @Test
    public void testAccept() {
	VisitorStub visitor = new VisitorStub();
	testObj.accept(visitor);
	assertTrue(
		"Correct method, i.e. PseudoRootNode-accept method should be called",
		visitor.called);
    }

    /**
     * Test method for
     * {@link hkademlia.components.routingtable.PseudoRootNode#getPrefix()}.
     */
    @Test
    public void testGetPrefix() {
	assertEquals("Prefix of pseudo root node is always zero",
		BigInteger.ZERO, testObj.getPrefix());
    }

    /**
     * Node stub.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    private static class NodeStub extends AbstractNode<KademliaOverlayID> {

	public NodeStub(BigInteger prefix,
		ParentNode<KademliaOverlayID> parent, RoutingTableConfig conf) {
	    super(prefix, parent, conf);
	}

	@Override
	public void accept(NodeVisitor<KademliaOverlayID> visitor) {
	    // not used
	}
    }

    /**
     * Visitor stub that allows to verify that the correct method (the one for
     * PseudoRootNode) has been called.
     * 
     * @author Tobias Lauinger <tl1003@rbg.informatik.tu-darmstadt.de>
     */
    private static class VisitorStub implements NodeVisitor<KademliaOverlayID> {

	public boolean called = false;

	@Override
	public void visit(Node<KademliaOverlayID> node) {
	    // not used
	}

	@Override
	public void visit(PseudoRootNode<KademliaOverlayID> node) {
	    this.called = true;
	}

	@Override
	public void visit(BranchNode<KademliaOverlayID> node) {
	    // not used
	}

	@Override
	public void visit(LeafNode<KademliaOverlayID> node) {
	    // not used
	}
    }
}
